# $Id$
#
# vim: set cindent expandtab ts=4 sw=4:
# 
# Encode::PPUniDetector - pure perl unicode (utf8 only) language
# detector module. Please refer to Unicode 4.x/5.0 standard for
# completely information.
#
# Copyright (c) 1998-2007 He zhiqiang. All right reserved.
# License: GPL v2
#
# contact: He zhiqiang <hzqbbc@hzqbbc.com>

package Encode::PPUniDetector;
use strict;
use Encode qw(from_to);
use Exporter;
use vars qw($VERSION @EXPORT @EXPORT_OK);

$VERSION = '1.0';
@EXPORT = qw();
@EXPORT_OK = qw(detect detect1 trylocal trylocal1 trylocal2);

use vars qw($UTF8_TCH $UTF8_SCH $DETECT_MAPS);

# use sort method to fast detect!
use constant VAR_HAN => 1;
use constant VAR_SCH => 2;
use constant VAR_TCH => 3;
use constant VAR_JP  => 4;
use constant VAR_KR  => 5;
use constant VAR_7BIT => 6;

# HAN SCH TCH JP KR

$DETECT_MAPS = {
    'HAN+SCH+TCH+JP+KR'     => 'UTF-8',
    'HAN+SCH+TCH+KR'        => ['euc-kr', 'iso-2022-kr', 'UTF-8'],
    'HAN+SCH+TCH+JP'        => ['shift-jis', 'euc-jp', 'iso-2022-jp', 'UTF-8'],
    'HAN+SCH+JP+KR'         => 'UTF-8',
    'HAN+TCH+JP+KR'         => 'UTF-8',
    'SCH+TCH+JP+KR'         => 'UTF-8',
    'HAN+SCH+TCH'           => 'GB2312', # GBK
    'HAN+SCH+JP'            => ['euc-jp', 'shift-jis', 'UTF-8'],
    'HAN+SCH+KR'            => ['euc-kr', 'iso-2022-kr', 'UTF-8'],
    'HAN+TCH+JP'            => ['euc-jp', 'shift-jis', 'UTF-8'],
    'HAN+TCH+KR'            => ['euc-kr', 'iso-2022-kr', 'UTF-8'],
    'HAN+JP+KR'             => 'UTF-8',
    'SCH+TCH+JP'            => 'GB2312', # GBK
    'SCH+TCH+KR'            => 'UTF-8',
    'SCH+JP+KR'             => 'UTF-8',
    'TCH+JP+KR'             => 'UTF-8',
    'HAN+SCH'               => 'GB2312', # GBK
    'HAN+TCH'               => 'BIG5',
    'HAN+JP'                => ['euc-jp', 'iso-2022-jp'],
    'HAN+KR'                => ['euc-kr', 'iso-2022-kr'],
    'SCH+TCH'               => 'GB2312', # GBK
    'SCH+JP'                => 'UTF-8',
    'SCH+KR'                => 'UTF-8',
    'TCH+JP'                => 'euc-jp',
    'TCH+KR'                => 'euc-kr',
    'JP+KR'                 => 'UTF-8',
    'HAN'                   => ['GB2312', 'BIG5', 'EUC-JP', 'EUC-KR', 'UTF-8'],
    'SCH'                   => 'GB2312', # GBK
    'TCH'                   => 'BIG5',
    'JP'                    => ['shift-jis', 'euc-jp'],
    'KR'                    => ['euc-kr', 'iso-2022-kr'],
    '7BIT'                  => ['ISO-8859-1', 'UTF-8'],
};

use vars qw(@HAN @JP @KR);

# Unihan code page range maps
@HAN = (
    '\x{4E00}-\x{9FBF}',    # Unihan cjk part
    '\x{3400}-\x{4DBF}',    # Ideography extA
    '\x{F900}-\x{FAFF}',    # Compatibility Ideo*
    '\x{3190}-\x{319F}',    # Kan bun
    '\x{FF00}-\x{FFEF}',    # * Halfwidth Katakana
                            # * Halfwidth Jamo
                            # share by JP/KR
);

# japanese code page range
@JP = (
    '[\x{3040}-\x{309F}]',  # Hiragana
    '[\x{30A0}-\x{30FF}]',  # Katakana
    '[\x{31F0}-\x{31FF}]'   # Katakana Phonetic Ext
);

# korea code page range
@KR = (
    '[\x{AC00}-\x{D7AF}]',  # Hangul Syllables
    '[\x{1100}-\x{11FF}]',  # Hangul Jamo
    '[\x{3130}-\x{318F}]',  # Hangul Compatibility Jamo
);

# traditional chinese characteristic code page
$UTF8_TCH = {
    '2013' => 8211,'2014' => 8212,
    '2025' => 8229,'2027' => 8231,
    '2035' => 8245,'2105' => 8453,
    '2109' => 8457,'2121' => 8481,
    '2160' => 8544,'2161' => 8545,
    '2162' => 8546,'2163' => 8547,
    '2164' => 8548,'2165' => 8549,
    '2166' => 8550,'2167' => 8551,
    '2170' => 8560,'2171' => 8561,
    '2172' => 8562,'2173' => 8563,
    '2174' => 8564,'2175' => 8565,
    '2176' => 8566,'2177' => 8567,
    '2178' => 8568,'2179' => 8569,
    '2196' => 8598,'2197' => 8599,
    '2198' => 8600,'2199' => 8601,
    '21B8' => 8632,'21B9' => 8633,
    '21E7' => 8679,'2215' => 8725,
    '221F' => 8735,'2223' => 8739,
    '2252' => 8786,'2266' => 8806,
    '2267' => 8807,'2295' => 8853,
    '22BF' => 8895,'2460' => 9312,
    '2461' => 9313,'2462' => 9314,
    '2463' => 9315,'2464' => 9316,
    '2465' => 9317,'2466' => 9318,
    '2467' => 9319,'2468' => 9320,
    '2469' => 9321,'2474' => 9332,
    '2475' => 9333,'2476' => 9334,
    '2477' => 9335,'2478' => 9336,
    '2479' => 9337,'247A' => 9338,
    '247B' => 9339,'247C' => 9340,
    '247D' => 9341,'2500' => 9472,
    '2502' => 9474,'2550' => 9552,
    '2551' => 9553,'2552' => 9554,
    '2553' => 9555,'2554' => 9556,
    '2555' => 9557,'2556' => 9558,
    '2557' => 9559,'2558' => 9560,
    '2559' => 9561,'255A' => 9562,
    '255B' => 9563,'255C' => 9564,
    '255D' => 9565,'255E' => 9566,
    '255F' => 9567,'2560' => 9568,
    '2561' => 9569,'2562' => 9570,
    '2563' => 9571,'2564' => 9572,
    '2565' => 9573,'2566' => 9574,
    '2567' => 9575,'2568' => 9576,
    '2569' => 9577,'256A' => 9578,
    '256B' => 9579,'256C' => 9580,
    '256D' => 9581,'256E' => 9582,
    '256F' => 9583,'2570' => 9584,
    '2571' => 9585,'2572' => 9586,
    '2573' => 9587,'2574' => 9588,
    '2581' => 9601,'2582' => 9602,
    '2583' => 9603,'2584' => 9604,
    '2585' => 9605,'2586' => 9606,
    '2587' => 9607,'2588' => 9608,
    '2589' => 9609,'258A' => 9610,
    '258B' => 9611,'258C' => 9612,
    '258D' => 9613,'258E' => 9614,
    '258F' => 9615,'2593' => 9619,
    '2594' => 9620,'2595' => 9621,
    '25BC' => 9660,'25BD' => 9661,
    '25E2' => 9698,'25E3' => 9699,
    '25E4' => 9700,'25E5' => 9701,
    '273D' => 10045,'2E80' => 11904,
    '2E84' => 11908,'2E86' => 11910,
    '2E87' => 11911,'2E88' => 11912,
    '2E8A' => 11914,'2E8C' => 11916,
    '2E8D' => 11917,'2E95' => 11925,
    '2E9C' => 11932,'2E9D' => 11933,
    '2EA5' => 11941,'2EA7' => 11943,
    '2EAA' => 11946,'2EAC' => 11948,
    '2EAE' => 11950,'2EB6' => 11958,
    '2EBC' => 11964,'2EBE' => 11966,
    '2EC6' => 11974,'2ECA' => 11978,
    '2ECC' => 11980,'2ECD' => 11981,
    '2ECF' => 11983,'2ED6' => 11990,
    '2ED7' => 11991,'2EDE' => 11998,
    '2EE3' => 12003,'2F02' => 12034,
    '2F03' => 12035,'2F05' => 12037,
    '2F07' => 12039,'2F0C' => 12044,
    '2F0D' => 12045,'2F0E' => 12046,
    '2F13' => 12051,'2F16' => 12054,
    '2F19' => 12057,'2F1B' => 12059,
    '2F22' => 12066,'2F27' => 12071,
    '2F2E' => 12078,'2F33' => 12083,
    '2F34' => 12084,'2F35' => 12085,
    '2F39' => 12089,'2F3A' => 12090,
    '2F41' => 12097,'2F46' => 12102,
    '2F67' => 12135,'2F68' => 12136,
    '2FA1' => 12193,'2FAA' => 12202,
    '3006' => 12294,'3007' => 12295,
    '3012' => 12306,'301D' => 12317,
    '301E' => 12318,'3021' => 12321,
    '3022' => 12322,'3023' => 12323,
    '3024' => 12324,'3025' => 12325,
    '3026' => 12326,'3027' => 12327,
    '3028' => 12328,'3029' => 12329,
    '3038' => 12344,'3039' => 12345,
    '303A' => 12346,'309B' => 12443,
    '309C' => 12444,'309D' => 12445,
    '309E' => 12446,'30FC' => 12540,
    '30FD' => 12541,'30FE' => 12542,
    '3105' => 12549,'3106' => 12550,
    '3107' => 12551,'3108' => 12552,
    '3109' => 12553,'310A' => 12554,
    '310B' => 12555,'310C' => 12556,
    '310D' => 12557,'310E' => 12558,
    '310F' => 12559,'3110' => 12560,
    '3111' => 12561,'3112' => 12562,
    '3113' => 12563,'3114' => 12564,
    '3115' => 12565,'3116' => 12566,
    '3117' => 12567,'3118' => 12568,
    '3119' => 12569,'311A' => 12570,
    '311B' => 12571,'311C' => 12572,
    '311D' => 12573,'311E' => 12574,
    '311F' => 12575,'3120' => 12576,
    '3121' => 12577,'3122' => 12578,
    '3123' => 12579,'3124' => 12580,
    '3125' => 12581,'3126' => 12582,
    '3127' => 12583,'3128' => 12584,
    '3129' => 12585,'3231' => 12849,
    '32A3' => 12963,'338E' => 13198,
    '338F' => 13199,'339C' => 13212,
    '339D' => 13213,'339E' => 13214,
    '33A1' => 13217,'33C4' => 13252,
    '33CE' => 13262,'33D1' => 13265,
    '33D2' => 13266,'33D5' => 13269,
    '4E0F' => 19983,'4E1F' => 19999,
    '4E26' => 20006,'4E2E' => 20014,
    '4E31' => 20017,'4E33' => 20019,
    '4E3C' => 20028,'4E42' => 20034,
    '4E5A' => 20058,'4E7F' => 20095,
    '4E82' => 20098,'4E83' => 20099,
    '4E84' => 20100,'4E99' => 20121,
    '4E9E' => 20126,'4EB6' => 20150,
    '4EB9' => 20153,'4EC8' => 20168,
    '4EDA' => 20186,'4EDC' => 20188,
    '4EE9' => 20201,'4EF1' => 20209,
    '4EF4' => 20212,'4F00' => 20224,
    '4F02' => 20226,'4F04' => 20228,
    '4F05' => 20229,'4F08' => 20232,
    '4F0B' => 20235,'4F12' => 20242,
    '4F13' => 20243,'4F14' => 20244,
    '4F15' => 20245,'4F1D' => 20253,
    '4F2C' => 20268,'4F2D' => 20269,
    '4F33' => 20275,'4F3B' => 20283,
    '4F3E' => 20286,'4F3F' => 20287,
    '4F41' => 20289,'4F47' => 20295,
    '4F48' => 20296,'4F49' => 20297,
    '4F4C' => 20300,'4F52' => 20306,
    '4F54' => 20308,'4F56' => 20310,
    '4F61' => 20321,'4F62' => 20322,
    '4F6A' => 20330,'4F6B' => 20331,
    '4F6E' => 20334,'4F75' => 20341,
    '4F77' => 20343,'4F78' => 20344,
    '4F79' => 20345,'4F7A' => 20346,
    '4F7D' => 20349,'4F80' => 20352,
    '4F81' => 20353,'4F82' => 20354,
    '4F85' => 20357,'4F86' => 20358,
    '4F87' => 20359,'4F90' => 20368,
    '4F92' => 20370,'4F95' => 20373,
    '4F96' => 20374,'4F98' => 20376,
    '4F9A' => 20378,'4F9C' => 20380,
    '4F9E' => 20382,'4FB2' => 20402,
    '4FB3' => 20403,'4FB6' => 20406,
    '4FB7' => 20407,'4FB9' => 20409,
    '4FBA' => 20410,'4FBB' => 20411,
    '4FC0' => 20416,'4FC1' => 20417,
    '4FC2' => 20418,'4FC7' => 20423,
    '4FC9' => 20425,'4FCB' => 20427,
    '4FCD' => 20429,'4FD3' => 20435,
    '4FD4' => 20436,'4FD6' => 20438,
    '4FD9' => 20441,'4FDB' => 20443,
    '4FE0' => 20448,'4FEC' => 20460,
    '4FF4' => 20468,'4FF5' => 20469,
    '4FF6' => 20470,'4FF7' => 20471,
    '5000' => 20480,'5005' => 20485,
    '5006' => 20486,'5007' => 20487,
    '5009' => 20489,'500B' => 20491,
    '500E' => 20494,'5011' => 20497,
    '5013' => 20499,'5015' => 20501,
    '5016' => 20502,'5017' => 20503,
    '501B' => 20507,'501E' => 20510,
    '5020' => 20512,'5022' => 20514,
    '5023' => 20515,'5027' => 20519,
    '502B' => 20523,'502F' => 20527,
    '5030' => 20528,'5031' => 20529,
    '5033' => 20531,'5035' => 20533,
    '5037' => 20535,'5040' => 20544,
    '5041' => 20545,'5045' => 20549,
    '5046' => 20550,'5049' => 20553,
    '504A' => 20554,'504B' => 20555,
    '504D' => 20557,'5051' => 20561,
    '5053' => 20563,'5057' => 20567,
    '505B' => 20571,'505D' => 20573,
    '505E' => 20574,'505F' => 20575,
    '5060' => 20576,'5061' => 20577,
    '5062' => 20578,'5063' => 20579,
    '5064' => 20580,'5068' => 20584,
    '5069' => 20585,'506A' => 20586,
    '506B' => 20587,'506D' => 20589,
    '506E' => 20590,'506F' => 20591,
    '5070' => 20592,'5072' => 20594,
    '5073' => 20595,'5074' => 20596,
    '5075' => 20597,'507A' => 20602,
    '507D' => 20605,'5082' => 20610,
    '5083' => 20611,'5087' => 20615,
    '508B' => 20619,'508C' => 20620,
    '508E' => 20622,'5091' => 20625,
    '5092' => 20626,'5094' => 20628,
    '5095' => 20629,'5096' => 20630,
    '5098' => 20632,'5099' => 20633,
    '509A' => 20634,'509B' => 20635,
    '509C' => 20636,'509D' => 20637,
    '509E' => 20638,'50A2' => 20642,
    '50AD' => 20653,'50AE' => 20654,
    '50AF' => 20655,'50B0' => 20656,
    '50B1' => 20657,'50B3' => 20659,
    '50B4' => 20660,'50B5' => 20661,
    '50B6' => 20662,'50B7' => 20663,
    '50B8' => 20664,'50BD' => 20669,
    '50BE' => 20670,'50BF' => 20671,
    '50C1' => 20673,'50C2' => 20674,
    '50C4' => 20676,'50C5' => 20677,
    '50C6' => 20678,'50C7' => 20679,
    '50C8' => 20680,'50C9' => 20681,
    '50CA' => 20682,'50CB' => 20683,
    '50CE' => 20686,'50D1' => 20689,
    '50D3' => 20691,'50D4' => 20692,
    '50D5' => 20693,'50D7' => 20695,
    '50DB' => 20699,'50DD' => 20701,
    '50E0' => 20704,'50E3' => 20707,
    '50E4' => 20708,'50E5' => 20709,
    '50E8' => 20712,'50E9' => 20713,
    '50EA' => 20714,'50EF' => 20719,
    '50F0' => 20720,'50F1' => 20721,
    '50F6' => 20726,'50F8' => 20728,
    '50F9' => 20729,'50FD' => 20733,
    '50FE' => 20734,'50FF' => 20735,
    '5100' => 20736,'5102' => 20738,
    '5103' => 20739,'5104' => 20740,
    '5105' => 20741,'5108' => 20744,
    '5109' => 20745,'510A' => 20746,
    '510C' => 20748,'5110' => 20752,
    '5111' => 20753,'5113' => 20755,
    '5114' => 20756,'5115' => 20757,
    '5117' => 20759,'5118' => 20760,
    '511A' => 20762,'511C' => 20764,
    '511F' => 20767,'5120' => 20768,
    '5122' => 20770,'5124' => 20772,
    '5125' => 20773,'5126' => 20774,
    '5129' => 20777,'512A' => 20778,
    '512D' => 20781,'512E' => 20782,
    '5130' => 20784,'5131' => 20785,
    '5132' => 20786,'5133' => 20787,
    '5134' => 20788,'5135' => 20789,
    '5137' => 20791,'5138' => 20792,
    '5139' => 20793,'513A' => 20794,
    '513B' => 20795,'513C' => 20796,
    '513D' => 20797,'5147' => 20807,
    '514C' => 20812,'5152' => 20818,
    '5157' => 20823,'5159' => 20825,
    '515B' => 20827,'515D' => 20829,
    '515E' => 20830,'515F' => 20831,
    '5161' => 20833,'5163' => 20835,
    '5167' => 20839,'5169' => 20841,
    '5187' => 20871,'518A' => 20874,
    '518F' => 20879,'5191' => 20881,
    '5193' => 20883,'5194' => 20884,
    '5198' => 20888,'519E' => 20894,
    '51AA' => 20906,'51B9' => 20921,
    '51BE' => 20926,'51C5' => 20933,
    '51C8' => 20936,'51CA' => 20938,
    '51CD' => 20941,'51CE' => 20942,
    '51D0' => 20944,'51D4' => 20948,
    '51D7' => 20951,'51D8' => 20952,
    '51DC' => 20956,'51DE' => 20958,
    '51F1' => 20977,'5209' => 21001,
    '520C' => 21004,'5210' => 21008,
    '5213' => 21011,'521C' => 21020,
    '521E' => 21022,'5221' => 21025,
    '5225' => 21029,'522A' => 21034,
    '5231' => 21041,'5232' => 21042,
    '5235' => 21045,'5244' => 21060,
    '5246' => 21062,'5247' => 21063,
    '5249' => 21065,'524B' => 21067,
    '524E' => 21070,'5252' => 21074,
    '5255' => 21077,'525A' => 21082,
    '525B' => 21083,'525D' => 21085,
    '525F' => 21087,'5262' => 21090,
    '526B' => 21099,'526C' => 21100,
    '526D' => 21101,'526E' => 21102,
    '5274' => 21108,'5275' => 21109,
    '5277' => 21111,'5278' => 21112,
    '527A' => 21114,'527B' => 21115,
    '527C' => 21116,'5280' => 21120,
    '5283' => 21123,'5284' => 21124,
    '5287' => 21127,'5289' => 21129,
    '528A' => 21130,'528B' => 21131,
    '528C' => 21132,'528D' => 21133,
    '5291' => 21137,'5296' => 21142,
    '5297' => 21143,'5298' => 21144,
    '5299' => 21145,'52A6' => 21158,
    '52AE' => 21166,'52BB' => 21179,
    '52BC' => 21180,'52C0' => 21184,
    '52C1' => 21185,'52C2' => 21186,
    '52CD' => 21197,'52D3' => 21203,
    '52D5' => 21205,'52D7' => 21207,
    '52D9' => 21209,'52DB' => 21211,
    '52DD' => 21213,'52DE' => 21214,
    '52E2' => 21218,'52E3' => 21219,
    '52E6' => 21222,'52E9' => 21225,
    '52EB' => 21227,'52EF' => 21231,
    '52F1' => 21233,'52F3' => 21235,
    '52F4' => 21236,'52F5' => 21237,
    '52F7' => 21239,'52F8' => 21240,
    '52FB' => 21243,'52FC' => 21244,
    '5309' => 21257,'530A' => 21258,
    '530B' => 21259,'530E' => 21262,
    '5311' => 21265,'5312' => 21266,
    '531C' => 21276,'531F' => 21279,
    '5322' => 21282,'532D' => 21293,
    '532F' => 21295,'5330' => 21296,
    '5331' => 21297,'5334' => 21300,
    '5337' => 21303,'533C' => 21308,
    '533D' => 21309,'5340' => 21312,
    '534C' => 21324,'534D' => 21325,
    '5354' => 21332,'536C' => 21356,
    '5372' => 21362,'5379' => 21369,
    '537B' => 21371,'537C' => 21372,
    '538A' => 21386,'538E' => 21390,
    '538F' => 21391,'5392' => 21394,
    '5394' => 21396,'5396' => 21398,
    '5397' => 21399,'5399' => 21401,
    '539C' => 21404,'539E' => 21406,
    '53A4' => 21412,'53A7' => 21415,
    '53AC' => 21420,'53AD' => 21421,
    '53B2' => 21426,'53B4' => 21428,
    '53B9' => 21433,'53C3' => 21443,
    '53E1' => 21473,'53E2' => 21474,
    '5407' => 21511,'540B' => 21515,
    '5418' => 21528,'5419' => 21529,
    '541C' => 21532,'5424' => 21540,
    '5425' => 21541,'542A' => 21546,
    '5430' => 21552,'5433' => 21555,
    '5436' => 21558,'5437' => 21559,
    '543D' => 21565,'5441' => 21569,
    '5442' => 21570,'5445' => 21573,
    '5447' => 21575,'544E' => 21582,
    '544F' => 21583,'5460' => 21600,
    '5461' => 21601,'5463' => 21603,
    '5465' => 21605,'5467' => 21607,
    '546B' => 21611,'546C' => 21612,
    '546F' => 21615,'5470' => 21616,
    '5474' => 21620,'547A' => 21626,
    '547E' => 21630,'547F' => 21631,
    '5481' => 21633,'5487' => 21639,
    '5488' => 21640,'548D' => 21645,
    '5491' => 21649,'5498' => 21656,
    '54A0' => 21664,'54A1' => 21665,
    '54A2' => 21666,'54A5' => 21669,
    '54AE' => 21678,'54B0' => 21680,
    '54B6' => 21686,'54B7' => 21687,
    '54BA' => 21690,'54BC' => 21692,
    '54BE' => 21694,'54C3' => 21699,
    '54C5' => 21701,'54D6' => 21718,
    '54E0' => 21728,'54E1' => 21729,
    '54E2' => 21730,'54E4' => 21732,
    '54EB' => 21739,'54F1' => 21745,
    '54F7' => 21751,'54F8' => 21752,
    '54FB' => 21755,'5503' => 21763,
    '5504' => 21764,'5505' => 21765,
    '5508' => 21768,'550A' => 21770,
    '550B' => 21771,'550C' => 21772,
    '550E' => 21774,'5512' => 21778,
    '5517' => 21783,'551A' => 21786,
    '5526' => 21798,'552D' => 21805,
    '5532' => 21810,'5534' => 21812,
    '5535' => 21813,'5536' => 21814,
    '5538' => 21816,'5539' => 21817,
    '553B' => 21819,'5540' => 21824,
    '5545' => 21829,'5548' => 21832,
    '554B' => 21835,'554D' => 21837,
    '554E' => 21838,'554F' => 21839,
    '5551' => 21841,'5552' => 21842,
    '5557' => 21847,'555E' => 21854,
    '555F' => 21855,'5562' => 21858,
    '5563' => 21859,'557D' => 21885,
    '557F' => 21887,'558C' => 21900,
    '558D' => 21901,'558E' => 21902,
    '5592' => 21906,'5593' => 21907,
    '5595' => 21909,'559A' => 21914,
    '55A1' => 21921,'55A2' => 21922,
    '55A3' => 21923,'55A4' => 21924,
    '55A5' => 21925,'55A6' => 21926,
    '55A8' => 21928,'55AA' => 21930,
    '55AB' => 21931,'55AC' => 21932,
    '55AD' => 21933,'55AE' => 21934,
    '55B2' => 21938,'55BF' => 21951,
    '55C0' => 21952,'55C2' => 21954,
    '55C3' => 21955,'55C6' => 21958,
    '55C7' => 21959,'55C8' => 21960,
    '55CA' => 21962,'55CB' => 21963,
    '55CE' => 21966,'55CF' => 21967,
    '55D0' => 21968,'55D5' => 21973,
    '55D9' => 21977,'55DA' => 21978,
    '55DB' => 21979,'55E2' => 21986,
    '55E7' => 21991,'55E9' => 21993,
    '55F6' => 22006,'55F9' => 22009,
    '55FA' => 22010,'55FC' => 22012,
    '55FF' => 22015,'5602' => 22018,
    '5604' => 22020,'5606' => 22022,
    '560D' => 22029,'5610' => 22032,
    '5612' => 22034,'5613' => 22035,
    '5614' => 22036,'5615' => 22037,
    '5616' => 22038,'5617' => 22039,
    '561C' => 22044,'561D' => 22045,
    '5629' => 22057,'562A' => 22058,
    '562E' => 22062,'562F' => 22063,
    '5630' => 22064,'5633' => 22067,
    '5635' => 22069,'5638' => 22072,
    '563A' => 22074,'563D' => 22077,
    '563E' => 22078,'5640' => 22080,
    '5641' => 22081,'5642' => 22082,
    '5645' => 22085,'5646' => 22086,
    '5648' => 22088,'5649' => 22089,
    '564A' => 22090,'5653' => 22099,
    '565A' => 22106,'565E' => 22110,
    '5660' => 22112,'5663' => 22115,
    '5665' => 22117,'5666' => 22118,
    '566D' => 22125,'566E' => 22126,
    '566F' => 22127,'5670' => 22128,
    '5672' => 22130,'5673' => 22131,
    '5674' => 22132,'5677' => 22135,
    '5678' => 22136,'5679' => 22137,
    '567E' => 22142,'567F' => 22143,
    '5680' => 22144,'5681' => 22145,
    '5682' => 22146,'5683' => 22147,
    '5684' => 22148,'5687' => 22151,
    '568C' => 22156,'568D' => 22157,
    '5690' => 22160,'5695' => 22165,
    '5697' => 22167,'5698' => 22168,
    '5699' => 22169,'569A' => 22170,
    '569C' => 22172,'569D' => 22173,
    '56A5' => 22181,'56A6' => 22182,
    '56A7' => 22183,'56A8' => 22184,
    '56AA' => 22186,'56AB' => 22187,
    '56AC' => 22188,'56AD' => 22189,
    '56AE' => 22190,'56B2' => 22194,
    '56B3' => 22195,'56B4' => 22196,
    '56B5' => 22197,'56B6' => 22198,
    '56BD' => 22205,'56BE' => 22206,
    '56C0' => 22208,'56C1' => 22209,
    '56C2' => 22210,'56C3' => 22211,
    '56C5' => 22213,'56C6' => 22214,
    '56C8' => 22216,'56C9' => 22217,
    '56CB' => 22219,'56CC' => 22220,
    '56CD' => 22221,'56D1' => 22225,
    '56D3' => 22227,'56E5' => 22245,
    '56E7' => 22247,'56EA' => 22250,
    '56EE' => 22254,'56F7' => 22263,
    '5701' => 22273,'5702' => 22274,
    '5707' => 22279,'570B' => 22283,
    '570C' => 22284,'570D' => 22285,
    '5712' => 22290,'5713' => 22291,
    '5714' => 22292,'5716' => 22294,
    '5718' => 22296,'571A' => 22298,
    '571B' => 22299,'571E' => 22302,
    '5720' => 22304,'5722' => 22306,
    '5734' => 22324,'5741' => 22337,
    '5745' => 22341,'5749' => 22345,
    '574B' => 22347,'5752' => 22354,
    '5762' => 22370,'5770' => 22384,
    '5771' => 22385,'5772' => 22386,
    '5774' => 22388,'5775' => 22389,
    '577D' => 22397,'5780' => 22400,
    '578F' => 22415,'5794' => 22420,
    '5795' => 22421,'5797' => 22423,
    '5798' => 22424,'5799' => 22425,
    '579A' => 22426,'579D' => 22429,
    '579E' => 22430,'579F' => 22431,
    '57A5' => 22437,'57B5' => 22453,
    '57B6' => 22454,'57B9' => 22457,
    '57BA' => 22458,'57BC' => 22460,
    '57BD' => 22461,'57BF' => 22463,
    '57C1' => 22465,'57C6' => 22470,
    '57C7' => 22471,'57CC' => 22476,
    '57D0' => 22480,'57DC' => 22492,
    '57E1' => 22497,'57E2' => 22498,
    '57E3' => 22499,'57E5' => 22501,
    '57E7' => 22503,'57E9' => 22505,
    '57EC' => 22508,'57EE' => 22510,
    '57F0' => 22512,'57F1' => 22513,
    '57F2' => 22514,'57F3' => 22515,
    '57F5' => 22517,'57F6' => 22518,
    '57F7' => 22519,'57FB' => 22523,
    '57FC' => 22524,'5801' => 22529,
    '5804' => 22532,'5805' => 22533,
    '5808' => 22536,'5809' => 22537,
    '580A' => 22538,'580C' => 22540,
    '580E' => 22542,'5810' => 22544,
    '5814' => 22548,'581B' => 22555,
    '581C' => 22556,'581D' => 22557,
    '5823' => 22563,'5825' => 22565,
    '5827' => 22567,'5828' => 22568,
    '5829' => 22569,'582C' => 22572,
    '582D' => 22573,'582E' => 22574,
    '582F' => 22575,'5831' => 22577,
    '5832' => 22578,'5833' => 22579,
    '5834' => 22580,'5836' => 22582,
    '5837' => 22583,'5838' => 22584,
    '5839' => 22585,'583B' => 22587,
    '583D' => 22589,'583F' => 22591,
    '5848' => 22600,'5849' => 22601,
    '584A' => 22602,'584B' => 22603,
    '584E' => 22606,'584F' => 22607,
    '5852' => 22610,'5853' => 22611,
    '5855' => 22613,'5857' => 22615,
    '5859' => 22617,'585A' => 22618,
    '585B' => 22619,'585D' => 22621,
    '5862' => 22626,'5863' => 22627,
    '5864' => 22628,'5868' => 22632,
    '586D' => 22637,'586F' => 22639,
    '5871' => 22641,'5874' => 22644,
    '5875' => 22645,'5876' => 22646,
    '5879' => 22649,'587A' => 22650,
    '587B' => 22651,'587C' => 22652,
    '587D' => 22653,'587F' => 22655,
    '5882' => 22658,'5886' => 22662,
    '5887' => 22663,'5888' => 22664,
    '588A' => 22666,'588B' => 22667,
    '588E' => 22670,'588F' => 22671,
    '5890' => 22672,'5891' => 22673,
    '5894' => 22676,'5898' => 22680,
    '589C' => 22684,'589D' => 22685,
    '58A0' => 22688,'58A1' => 22689,
    '58A3' => 22691,'58A5' => 22693,
    '58A6' => 22694,'58AB' => 22699,
    '58AC' => 22700,'58AE' => 22702,
    '58AF' => 22703,'58B1' => 22705,
    '58B3' => 22707,'58BA' => 22714,
    '58BB' => 22715,'58BD' => 22717,
    '58BE' => 22718,'58BF' => 22719,
    '58C2' => 22722,'58C6' => 22726,
    '58C7' => 22727,'58C8' => 22728,
    '58C9' => 22729,'58CE' => 22734,
    '58CF' => 22735,'58D2' => 22738,
    '58D3' => 22739,'58D4' => 22740,
    '58D6' => 22742,'58D8' => 22744,
    '58D9' => 22745,'58DA' => 22746,
    '58DB' => 22747,'58DD' => 22749,
    '58DE' => 22750,'58DF' => 22751,
    '58E2' => 22754,'58E3' => 22755,
    '58E7' => 22759,'58E8' => 22760,
    '58E9' => 22761,'58EF' => 22767,
    '58F4' => 22772,'58FA' => 22778,
    '58FC' => 22780,'58FD' => 22781,
    '58FE' => 22782,'58FF' => 22783,
    '5903' => 22787,'5906' => 22790,
    '590C' => 22796,'590E' => 22798,
    '5912' => 22802,'5917' => 22807,
    '5920' => 22816,'5922' => 22818,
    '592C' => 22828,'593E' => 22846,
    '5940' => 22848,'5945' => 22853,
    '594A' => 22858,'5950' => 22864,
    '5953' => 22867,'595C' => 22876,
    '5961' => 22881,'5967' => 22887,
    '5969' => 22889,'596A' => 22890,
    '596B' => 22891,'596D' => 22893,
    '596E' => 22894,'5970' => 22896,
    '5971' => 22897,'5972' => 22898,
    '5977' => 22903,'597B' => 22907,
    '597C' => 22908,'597E' => 22910,
    '597F' => 22911,'5980' => 22912,
    '5985' => 22917,'598E' => 22926,
    '598F' => 22927,'5990' => 22928,
    '5998' => 22936,'599D' => 22941,
    '59A0' => 22944,'59A1' => 22945,
    '59A2' => 22946,'59A6' => 22950,
    '59A7' => 22951,'59B1' => 22961,
    '59B3' => 22963,'59B4' => 22964,
    '59B5' => 22965,'59B6' => 22966,
    '59BA' => 22970,'59BC' => 22972,
    '59BD' => 22973,'59C0' => 22976,
    '59C1' => 22977,'59C3' => 22979,
    '59C5' => 22981,'59C7' => 22983,
    '59C8' => 22984,'59CC' => 22988,
    '59CD' => 22989,'59CE' => 22990,
    '59CF' => 22991,'59D6' => 22998,
    '59DB' => 23003,'59DE' => 23006,
    '59E0' => 23008,'59E1' => 23009,
    '59E4' => 23012,'59E6' => 23014,
    '59E9' => 23017,'59EA' => 23018,
    '59ED' => 23021,'59EE' => 23022,
    '59F1' => 23025,'59F2' => 23026,
    '59F3' => 23027,'59F4' => 23028,
    '59F5' => 23029,'59F6' => 23030,
    '59F7' => 23031,'59FA' => 23034,
    '59FC' => 23036,'59FD' => 23037,
    '59FE' => 23038,'5A00' => 23040,
    '5A0A' => 23050,'5A0F' => 23055,
    '5A15' => 23061,'5A16' => 23062,
    '5A17' => 23063,'5A19' => 23065,
    '5A1B' => 23067,'5A1E' => 23070,
    '5A2D' => 23085,'5A2E' => 23086,
    '5A33' => 23091,'5A35' => 23093,
    '5A37' => 23095,'5A38' => 23096,
    '5A39' => 23097,'5A3E' => 23102,
    '5A41' => 23105,'5A42' => 23106,
    '5A43' => 23107,'5A44' => 23108,
    '5A47' => 23111,'5A48' => 23112,
    '5A4C' => 23116,'5A4D' => 23117,
    '5A50' => 23120,'5A51' => 23121,
    '5A52' => 23122,'5A53' => 23123,
    '5A56' => 23126,'5A57' => 23127,
    '5A58' => 23128,'5A5B' => 23131,
    '5A5C' => 23132,'5A5D' => 23133,
    '5A5E' => 23134,'5A5F' => 23135,
    '5A60' => 23136,'5A64' => 23140,
    '5A65' => 23141,'5A66' => 23142,
    '5A69' => 23145,'5A6C' => 23148,
    '5A6D' => 23149,'5A70' => 23152,
    '5A78' => 23160,'5A7B' => 23163,
    '5A7C' => 23164,'5A7D' => 23165,
    '5A83' => 23171,'5A84' => 23172,
    '5A8A' => 23178,'5A8B' => 23179,
    '5A8C' => 23180,'5A8E' => 23182,
    '5A8F' => 23183,'5A90' => 23184,
    '5A93' => 23187,'5A94' => 23188,
    '5A95' => 23189,'5A97' => 23191,
    '5A9C' => 23196,'5A9D' => 23197,
    '5A9E' => 23198,'5A9F' => 23199,
    '5AA2' => 23202,'5AA5' => 23205,
    '5AA6' => 23206,'5AA7' => 23207,
    '5AA9' => 23209,'5AAC' => 23212,
    '5AAE' => 23214,'5AAF' => 23215,
    '5AB0' => 23216,'5AB1' => 23217,
    '5AB4' => 23220,'5AB6' => 23222,
    '5AB7' => 23223,'5AB9' => 23225,
    '5ABA' => 23226,'5ABB' => 23227,
    '5ABC' => 23228,'5ABD' => 23229,
    '5ABF' => 23231,'5AC0' => 23232,
    '5AC4' => 23236,'5AC6' => 23238,
    '5AC7' => 23239,'5AC8' => 23240,
    '5ACA' => 23242,'5ACB' => 23243,
    '5ACD' => 23245,'5AD5' => 23253,
    '5AD7' => 23255,'5AD9' => 23257,
    '5ADA' => 23258,'5ADB' => 23259,
    '5ADD' => 23261,'5ADE' => 23262,
    '5ADF' => 23263,'5AE2' => 23266,
    '5AE5' => 23269,'5AE8' => 23272,
    '5AEA' => 23274,'5AEC' => 23276,
    '5AED' => 23277,'5AEE' => 23278,
    '5AF3' => 23283,'5AF4' => 23284,
    '5AF5' => 23285,'5AF6' => 23286,
    '5AF7' => 23287,'5AF8' => 23288,
    '5AF9' => 23289,'5AFA' => 23290,
    '5AFB' => 23291,'5AFD' => 23293,
    '5AFF' => 23295,'5B01' => 23297,
    '5B02' => 23298,'5B03' => 23299,
    '5B05' => 23301,'5B07' => 23303,
    '5B08' => 23304,'5B0B' => 23307,
    '5B0C' => 23308,'5B0F' => 23311,
    '5B10' => 23312,'5B13' => 23315,
    '5B14' => 23316,'5B19' => 23321,
    '5B1A' => 23322,'5B1B' => 23323,
    '5B1D' => 23325,'5B1E' => 23326,
    '5B20' => 23328,'5B21' => 23329,
    '5B23' => 23331,'5B24' => 23332,
    '5B25' => 23333,'5B26' => 23334,
    '5B27' => 23335,'5B28' => 23336,
    '5B2A' => 23338,'5B2C' => 23340,
    '5B2D' => 23341,'5B2E' => 23342,
    '5B2F' => 23343,'5B30' => 23344,
    '5B38' => 23352,'5B3C' => 23356,
    '5B3D' => 23357,'5B3E' => 23358,
    '5B3F' => 23359,'5B43' => 23363,
    '5B45' => 23365,'5B47' => 23367,
    '5B48' => 23368,'5B4B' => 23371,
    '5B4C' => 23372,'5B4D' => 23373,
    '5B4E' => 23374,'5B56' => 23382,
    '5B6B' => 23403,'5B6E' => 23406,
    '5B72' => 23410,'5B77' => 23415,
    '5B78' => 23416,'5B7B' => 23419,
    '5B7F' => 23423,'5B8E' => 23438,
    '5B92' => 23442,'5BA7' => 23463,
    '5BA8' => 23464,'5BAC' => 23468,
    '5BAD' => 23469,'5BAE' => 23470,
    '5BC0' => 23488,'5BC1' => 23489,
    '5BCA' => 23498,'5BCB' => 23499,
    '5BCD' => 23501,'5BCE' => 23502,
    '5BD1' => 23505,'5BD4' => 23508,
    '5BD6' => 23510,'5BD8' => 23512,
    '5BD9' => 23513,'5BE0' => 23520,
    '5BE2' => 23522,'5BE3' => 23523,
    '5BE6' => 23526,'5BE7' => 23527,
    '5BE9' => 23529,'5BEA' => 23530,
    '5BEB' => 23531,'5BEC' => 23532,
    '5BEF' => 23535,'5BF1' => 23537,
    '5BF2' => 23538,'5BF5' => 23541,
    '5BF6' => 23542,'5C03' => 23555,
    '5C07' => 23559,'5C08' => 23560,
    '5C0B' => 23563,'5C0C' => 23564,
    '5C0D' => 23565,'5C0E' => 23566,
    '5C10' => 23568,'5C12' => 23570,
    '5C1F' => 23583,'5C28' => 23592,
    '5C2A' => 23594,'5C30' => 23600,
    '5C33' => 23603,'5C37' => 23607,
    '5C44' => 23620,'5C46' => 23622,
    '5C47' => 23623,'5C4C' => 23628,
    '5C4D' => 23629,'5C54' => 23636,
    '5C56' => 23638,'5C58' => 23640,
    '5C5C' => 23644,'5C5D' => 23645,
    '5C62' => 23650,'5C64' => 23652,
    '5C67' => 23655,'5C68' => 23656,
    '5C69' => 23657,'5C6A' => 23658,
    '5C6C' => 23660,'5C6D' => 23661,
    '5C73' => 23667,'5C74' => 23668,
    '5C7B' => 23675,'5C7C' => 23676,
    '5C7E' => 23678,'5C86' => 23686,
    '5C89' => 23689,'5C8A' => 23690,
    '5C8B' => 23691,'5C8F' => 23695,
    '5C92' => 23698,'5C93' => 23699,
    '5C95' => 23701,'5C9D' => 23709,
    '5C9F' => 23711,'5CA0' => 23712,
    '5CA1' => 23713,'5CA4' => 23716,
    '5CA5' => 23717,'5CA6' => 23718,
    '5CA7' => 23719,'5CA8' => 23720,
    '5CAA' => 23722,'5CAE' => 23726,
    '5CAF' => 23727,'5CB0' => 23728,
    '5CB6' => 23734,'5CC6' => 23750,
    '5CC7' => 23751,'5CC8' => 23752,
    '5CC9' => 23753,'5CCA' => 23754,
    '5CCC' => 23756,'5CCE' => 23758,
    '5CCF' => 23759,'5CD0' => 23760,
    '5CD3' => 23763,'5CD4' => 23764,
    '5CD6' => 23766,'5CD7' => 23767,
    '5CD8' => 23768,'5CDA' => 23770,
    '5CDB' => 23771,'5CDE' => 23774,
    '5CDF' => 23775,'5CEC' => 23788,
    '5CEE' => 23790,'5CF1' => 23793,
    '5CF4' => 23796,'5CF6' => 23798,
    '5CF7' => 23799,'5CF8' => 23800,
    '5CF9' => 23801,'5CFD' => 23805,
    '5CFF' => 23807,'5D00' => 23808,
    '5D01' => 23809,'5D0B' => 23819,
    '5D0C' => 23820,'5D0D' => 23821,
    '5D0F' => 23823,'5D11' => 23825,
    '5D12' => 23826,'5D17' => 23831,
    '5D19' => 23833,'5D1A' => 23834,
    '5D1D' => 23837,'5D1F' => 23839,
    '5D20' => 23840,'5D22' => 23842,
    '5D23' => 23843,'5D25' => 23845,
    '5D28' => 23848,'5D30' => 23856,
    '5D31' => 23857,'5D32' => 23858,
    '5D33' => 23859,'5D35' => 23861,
    '5D36' => 23862,'5D37' => 23863,
    '5D38' => 23864,'5D39' => 23865,
    '5D3A' => 23866,'5D3C' => 23868,
    '5D3F' => 23871,'5D40' => 23872,
    '5D41' => 23873,'5D42' => 23874,
    '5D43' => 23875,'5D45' => 23877,
    '5D49' => 23881,'5D4E' => 23886,
    '5D50' => 23888,'5D51' => 23889,
    '5D52' => 23890,'5D55' => 23893,
    '5D59' => 23897,'5D5E' => 23902,
    '5D62' => 23906,'5D63' => 23907,
    '5D65' => 23909,'5D67' => 23911,
    '5D68' => 23912,'5D71' => 23921,
    '5D72' => 23922,'5D77' => 23927,
    '5D79' => 23929,'5D7A' => 23930,
    '5D7C' => 23932,'5D7D' => 23933,
    '5D7E' => 23934,'5D7F' => 23935,
    '5D80' => 23936,'5D81' => 23937,
    '5D84' => 23940,'5D86' => 23942,
    '5D87' => 23943,'5D88' => 23944,
    '5D89' => 23945,'5D8A' => 23946,
    '5D8D' => 23949,'5D92' => 23954,
    '5D93' => 23955,'5D94' => 23956,
    '5D95' => 23957,'5D97' => 23959,
    '5D9A' => 23962,'5D9C' => 23964,
    '5D9E' => 23966,'5D9F' => 23967,
    '5DA0' => 23968,'5DA1' => 23969,
    '5DA2' => 23970,'5DA7' => 23975,
    '5DA8' => 23976,'5DA9' => 23977,
    '5DAA' => 23978,'5DAC' => 23980,
    '5DAD' => 23981,'5DAE' => 23982,
    '5DAF' => 23983,'5DB0' => 23984,
    '5DB1' => 23985,'5DB2' => 23986,
    '5DB4' => 23988,'5DB5' => 23989,
    '5DB8' => 23992,'5DBA' => 23994,
    '5DBC' => 23996,'5DBD' => 23997,
    '5DC0' => 24000,'5DC2' => 24002,
    '5DC3' => 24003,'5DC6' => 24006,
    '5DC7' => 24007,'5DC9' => 24009,
    '5DCB' => 24011,'5DCF' => 24015,
    '5DD1' => 24017,'5DD2' => 24018,
    '5DD4' => 24020,'5DD5' => 24021,
    '5DD6' => 24022,'5DD8' => 24024,
    '5DDF' => 24031,'5DE0' => 24032,
    '5DF0' => 24048,'5DF9' => 24057,
    '5DFF' => 24063,'5E04' => 24068,
    '5E0A' => 24074,'5E0E' => 24078,
    '5E17' => 24087,'5E1F' => 24095,
    '5E20' => 24096,'5E21' => 24097,
    '5E22' => 24098,'5E23' => 24099,
    '5E24' => 24100,'5E25' => 24101,
    '5E28' => 24104,'5E29' => 24105,
    '5E2B' => 24107,'5E33' => 24115,
    '5E34' => 24116,'5E36' => 24118,
    '5E3E' => 24126,'5E40' => 24128,
    '5E41' => 24129,'5E43' => 24131,
    '5E4A' => 24138,'5E4B' => 24139,
    '5E4D' => 24141,'5E4E' => 24142,
    '5E4F' => 24143,'5E53' => 24147,
    '5E57' => 24151,'5E58' => 24152,
    '5E59' => 24153,'5E5C' => 24156,
    '5E5D' => 24157,'5E5F' => 24159,
    '5E60' => 24160,'5E63' => 24163,
    '5E66' => 24166,'5E67' => 24167,
    '5E68' => 24168,'5E69' => 24169,
    '5E6A' => 24170,'5E6B' => 24171,
    '5E6C' => 24172,'5E6D' => 24173,
    '5E6E' => 24174,'5E6F' => 24175,
    '5E70' => 24176,'5E75' => 24181,
    '5E79' => 24185,'5E7E' => 24190,
    '5E82' => 24194,'5E88' => 24200,
    '5E89' => 24201,'5E8C' => 24204,
    '5E8D' => 24205,'5E9B' => 24219,
    '5EA2' => 24226,'5EA3' => 24227,
    '5EA4' => 24228,'5EA8' => 24232,
    '5EAA' => 24234,'5EAB' => 24235,
    '5EAC' => 24236,'5EAE' => 24238,
    '5EB0' => 24240,'5EB1' => 24241,
    '5EB2' => 24242,'5EB4' => 24244,
    '5EC1' => 24257,'5EC2' => 24258,
    '5EC4' => 24260,'5EC5' => 24261,
    '5EC6' => 24262,'5EC7' => 24263,
    '5EC8' => 24264,'5ECB' => 24267,
    '5ECC' => 24268,'5ECE' => 24270,
    '5ED4' => 24276,'5ED5' => 24277,
    '5ED7' => 24279,'5ED8' => 24280,
    '5ED9' => 24281,'5EDA' => 24282,
    '5EDC' => 24284,'5EDD' => 24285,
    '5EDE' => 24286,'5EDF' => 24287,
    '5EE0' => 24288,'5EE1' => 24289,
    '5EE2' => 24290,'5EE3' => 24291,
    '5EE5' => 24293,'5EE6' => 24294,
    '5EE7' => 24295,'5EE9' => 24297,
    '5EEC' => 24300,'5EEE' => 24302,
    '5EEF' => 24303,'5EF1' => 24305,
    '5EF2' => 24306,'5EF3' => 24307,
    '5F05' => 24325,'5F07' => 24327,
    '5F12' => 24338,'5F14' => 24340,
    '5F1A' => 24346,'5F1D' => 24349,
    '5F22' => 24354,'5F23' => 24355,
    '5F24' => 24356,'5F28' => 24360,
    '5F2E' => 24366,'5F30' => 24368,
    '5F33' => 24371,'5F35' => 24373,
    '5F36' => 24374,'5F37' => 24375,
    '5F38' => 24376,'5F43' => 24387,
    '5F44' => 24388,'5F46' => 24390,
    '5F48' => 24392,'5F49' => 24393,
    '5F4A' => 24394,'5F4B' => 24395,
    '5F4C' => 24396,'5F4E' => 24398,
    '5F4F' => 24399,'5F54' => 24404,
    '5F59' => 24409,'5F5E' => 24414,
    '5F65' => 24421,'5F67' => 24423,
    '5F6B' => 24427,'5F6F' => 24431,
    '5F74' => 24436,'5F76' => 24438,
    '5F78' => 24440,'5F7D' => 24445,
    '5F7E' => 24446,'5F7F' => 24447,
    '5F86' => 24454,'5F91' => 24465,
    '5F96' => 24470,'5F9B' => 24475,
    '5F9E' => 24478,'5F9F' => 24479,
    '5FA0' => 24480,'5FA5' => 24485,
    '5FA6' => 24486,'5FA9' => 24489,
    '5FAB' => 24491,'5FAC' => 24492,
    '5FAF' => 24495,'5FB2' => 24498,
    '5FB6' => 24502,'5FB9' => 24505,
    '5FBB' => 24507,'5FBE' => 24510,
    '5FBF' => 24511,'5FC0' => 24512,
    '5FC1' => 24513,'5FD4' => 24532,
    '5FD5' => 24533,'5FDE' => 24542,
    '5FE3' => 24547,'5FE5' => 24549,
    '5FE8' => 24552,'5FEF' => 24559,
    '5FF3' => 24563,'5FF4' => 24564,
    '5FF7' => 24567,'5FFA' => 24570,
    '6009' => 24585,'600B' => 24587,
    '600C' => 24588,'6010' => 24592,
    '6011' => 24593,'6013' => 24595,
    '6017' => 24599,'601A' => 24602,
    '601E' => 24606,'6022' => 24610,
    '6024' => 24612,'602C' => 24620,
    '602D' => 24621,'602E' => 24622,
    '6032' => 24626,'6033' => 24627,
    '6034' => 24628,'6037' => 24631,
    '6039' => 24633,'6040' => 24640,
    '6044' => 24644,'6045' => 24645,
    '6046' => 24646,'6047' => 24647,
    '6049' => 24649,'604C' => 24652,
    '6053' => 24659,'6054' => 24660,
    '6058' => 24664,'605B' => 24667,
    '605E' => 24670,'605F' => 24671,
    '6065' => 24677,'6066' => 24678,
    '606E' => 24686,'6072' => 24690,
    '6080' => 24704,'6081' => 24705,
    '6085' => 24709,'6086' => 24710,
    '6087' => 24711,'6088' => 24712,
    '608A' => 24714,'608E' => 24718,
    '6090' => 24720,'6095' => 24725,
    '6097' => 24727,'609C' => 24732,
    '60A2' => 24738,'60B0' => 24752,
    '60B5' => 24757,'60B6' => 24758,
    '60B7' => 24759,'60B9' => 24761,
    '60BA' => 24762,'60BD' => 24765,
    '60BE' => 24766,'60BF' => 24767,
    '60C0' => 24768,'60C1' => 24769,
    '60C3' => 24771,'60C4' => 24772,
    '60C7' => 24775,'60C8' => 24776,
    '60C9' => 24777,'60CC' => 24780,
    '60CD' => 24781,'60CE' => 24782,
    '60CF' => 24783,'60D3' => 24787,
    '60D4' => 24788,'60D9' => 24793,
    '60DB' => 24795,'60E1' => 24801,
    '60E2' => 24802,'60E4' => 24804,
    '60F1' => 24817,'60F2' => 24818,
    '60F5' => 24821,'60F7' => 24823,
    '60F8' => 24824,'60FB' => 24827,
    '60FC' => 24828,'60FE' => 24830,
    '60FF' => 24831,'6103' => 24835,
    '6104' => 24836,'6105' => 24837,
    '610A' => 24842,'610B' => 24843,
    '6110' => 24848,'6112' => 24850,
    '6113' => 24851,'6114' => 24852,
    '6116' => 24854,'6118' => 24856,
    '611B' => 24859,'611C' => 24860,
    '611D' => 24861,'6128' => 24872,
    '6129' => 24873,'612C' => 24876,
    '612E' => 24878,'612F' => 24879,
    '6132' => 24882,'6134' => 24884,
    '6136' => 24886,'6137' => 24887,
    '613B' => 24891,'613E' => 24894,
    '6140' => 24896,'6141' => 24897,
    '6144' => 24900,'6145' => 24901,
    '6146' => 24902,'6147' => 24903,
    '6149' => 24905,'614B' => 24907,
    '614D' => 24909,'614F' => 24911,
    '6152' => 24914,'6153' => 24915,
    '6154' => 24916,'6156' => 24918,
    '6158' => 24920,'615A' => 24922,
    '615B' => 24923,'615E' => 24926,
    '615F' => 24927,'6161' => 24929,
    '6163' => 24931,'6165' => 24933,
    '6166' => 24934,'616A' => 24938,
    '616B' => 24939,'616C' => 24940,
    '616E' => 24942,'6171' => 24945,
    '6172' => 24946,'6173' => 24947,
    '6174' => 24948,'6176' => 24950,
    '6179' => 24953,'617A' => 24954,
    '617C' => 24956,'617E' => 24958,
    '6180' => 24960,'6182' => 24962,
    '6183' => 24963,'6189' => 24969,
    '618A' => 24970,'618C' => 24972,
    '618D' => 24973,'6190' => 24976,
    '6191' => 24977,'6192' => 24978,
    '6193' => 24979,'6196' => 24982,
    '619A' => 24986,'619B' => 24987,
    '619F' => 24991,'61A1' => 24993,
    '61A2' => 24994,'61A4' => 24996,
    '61AA' => 25002,'61AB' => 25003,
    '61AD' => 25005,'61AE' => 25006,
    '61AF' => 25007,'61B0' => 25008,
    '61B1' => 25009,'61B2' => 25010,
    '61B3' => 25011,'61B4' => 25012,
    '61B5' => 25013,'61B6' => 25014,
    '61B8' => 25016,'61BA' => 25018,
    '61BC' => 25020,'61BF' => 25023,
    '61C1' => 25025,'61C3' => 25027,
    '61C5' => 25029,'61C6' => 25030,
    '61C7' => 25031,'61C9' => 25033,
    '61CC' => 25036,'61CD' => 25037,
    '61D6' => 25046,'61D8' => 25048,
    '61DE' => 25054,'61DF' => 25055,
    '61E0' => 25056,'61E3' => 25059,
    '61E4' => 25060,'61E5' => 25061,
    '61E7' => 25063,'61E8' => 25064,
    '61E9' => 25065,'61EA' => 25066,
    '61EB' => 25067,'61ED' => 25069,
    '61EE' => 25070,'61F0' => 25072,
    '61F1' => 25073,'61F2' => 25074,
    '61F6' => 25078,'61F7' => 25079,
    '61F8' => 25080,'61F9' => 25081,
    '61FA' => 25082,'61FB' => 25083,
    '61FC' => 25084,'61FD' => 25085,
    '61FE' => 25086,'6200' => 25088,
    '6201' => 25089,'6203' => 25091,
    '6204' => 25092,'6207' => 25095,
    '6209' => 25097,'6214' => 25108,
    '6219' => 25113,'6220' => 25120,
    '6223' => 25123,'6227' => 25127,
    '6229' => 25129,'622B' => 25131,
    '622D' => 25133,'6230' => 25136,
    '6232' => 25138,'6236' => 25142,
    '623A' => 25146,'6242' => 25154,
    '6246' => 25158,'624A' => 25162,
    '6250' => 25168,'6259' => 25177,
    '625A' => 25178,'625C' => 25180,
    '625E' => 25182,'6260' => 25184,
    '6261' => 25185,'6262' => 25186,
    '6264' => 25188,'6265' => 25189,
    '6271' => 25201,'6272' => 25202,
    '6274' => 25204,'6277' => 25207,
    '627A' => 25210,'627B' => 25211,
    '627D' => 25213,'6281' => 25217,
    '6283' => 25219,'6286' => 25222,
    '6287' => 25223,'6288' => 25224,
    '628C' => 25228,'628E' => 25230,
    '628F' => 25231,'6294' => 25236,
    '62A9' => 25257,'62AA' => 25258,
    '62AD' => 25261,'62AE' => 25262,
    '62AF' => 25263,'62B0' => 25264,
    '62B3' => 25267,'62B4' => 25268,
    '62B6' => 25270,'62B8' => 25272,
    '62BE' => 25278,'62CB' => 25291,
    '62CF' => 25295,'62D1' => 25297,
    '62EB' => 25323,'62F0' => 25328,
    '62F2' => 25330,'62F5' => 25333,
    '62F8' => 25336,'62F9' => 25337,
    '62FA' => 25338,'62FB' => 25339,
    '6300' => 25344,'6303' => 25347,
    '630B' => 25355,'630C' => 25356,
    '630D' => 25357,'630F' => 25359,
    '6310' => 25360,'6313' => 25363,
    '6314' => 25364,'6315' => 25365,
    '6329' => 25385,'632C' => 25388,
    '632D' => 25389,'6333' => 25395,
    '6334' => 25396,'6336' => 25398,
    '6338' => 25400,'633B' => 25403,
    '633C' => 25404,'633E' => 25406,
    '6340' => 25408,'6341' => 25409,
    '6344' => 25412,'6347' => 25415,
    '6348' => 25416,'634A' => 25418,
    '6351' => 25425,'6354' => 25428,
    '6356' => 25430,'6357' => 25431,
    '6358' => 25432,'6359' => 25433,
    '635A' => 25434,'6365' => 25445,
    '6368' => 25448,'636B' => 25451,
    '636F' => 25455,'6370' => 25456,
    '6372' => 25458,'6375' => 25461,
    '6378' => 25464,'637C' => 25468,
    '637D' => 25469,'6381' => 25473,
    '6383' => 25475,'6384' => 25476,
    '6385' => 25477,'638D' => 25485,
    '6391' => 25489,'6394' => 25492,
    '6397' => 25495,'6399' => 25497,
    '639B' => 25499,'639C' => 25500,
    '639D' => 25501,'639E' => 25502,
    '639F' => 25503,'63A1' => 25505,
    '63A4' => 25508,'63AB' => 25515,
    '63AF' => 25519,'63B1' => 25521,
    '63BD' => 25533,'63C0' => 25536,
    '63C2' => 25538,'63C3' => 25539,
    '63C5' => 25541,'63C7' => 25543,
    '63C8' => 25544,'63CA' => 25546,
    '63CB' => 25547,'63CC' => 25548,
    '63D3' => 25555,'63D5' => 25557,
    '63D7' => 25559,'63D8' => 25560,
    '63D9' => 25561,'63DA' => 25562,
    '63DB' => 25563,'63DC' => 25564,
    '63DD' => 25565,'63DF' => 25567,
    '63E4' => 25572,'63E5' => 25573,
    '63E7' => 25575,'63E8' => 25576,
    '63EB' => 25579,'63EE' => 25582,
    '63EF' => 25583,'63F0' => 25584,
    '63F1' => 25585,'63F3' => 25587,
    '63F5' => 25589,'63F9' => 25593,
    '6406' => 25606,'6409' => 25609,
    '640A' => 25610,'640D' => 25613,
    '640E' => 25614,'6412' => 25618,
    '6415' => 25621,'6416' => 25622,
    '6417' => 25623,'6418' => 25624,
    '641A' => 25626,'641F' => 25631,
    '6422' => 25634,'6423' => 25635,
    '6424' => 25636,'6425' => 25637,
    '6427' => 25639,'6428' => 25640,
    '642B' => 25643,'642E' => 25646,
    '642F' => 25647,'6430' => 25648,
    '6433' => 25651,'6435' => 25653,
    '6436' => 25654,'6437' => 25655,
    '6439' => 25657,'643E' => 25662,
    '6440' => 25664,'6443' => 25667,
    '644B' => 25675,'644D' => 25677,
    '644E' => 25678,'6450' => 25680,
    '6451' => 25681,'6453' => 25683,
    '6459' => 25689,'645B' => 25691,
    '645C' => 25692,'645D' => 25693,
    '645F' => 25695,'6460' => 25696,
    '6461' => 25697,'6465' => 25701,
    '6466' => 25702,'6468' => 25704,
    '646B' => 25707,'646C' => 25708,
    '646E' => 25710,'646F' => 25711,
    '6470' => 25712,'6472' => 25714,
    '6473' => 25715,'6474' => 25716,
    '6475' => 25717,'6476' => 25718,
    '6477' => 25719,'647B' => 25723,
    '647D' => 25725,'647F' => 25727,
    '6488' => 25736,'6489' => 25737,
    '648A' => 25738,'648B' => 25739,
    '648C' => 25740,'648F' => 25743,
    '6490' => 25744,'6493' => 25747,
    '6497' => 25751,'6498' => 25752,
    '649A' => 25754,'649C' => 25756,
    '649D' => 25757,'649F' => 25759,
    '64A0' => 25760,'64A2' => 25762,
    '64A3' => 25763,'64A5' => 25765,
    '64A6' => 25766,'64AB' => 25771,
    '64B1' => 25777,'64B2' => 25778,
    '64B3' => 25779,'64BB' => 25787,
    '64BD' => 25789,'64BE' => 25790,
    '64BF' => 25791,'64C1' => 25793,
    '64C3' => 25795,'64C4' => 25796,
    '64C7' => 25799,'64C9' => 25801,
    '64CA' => 25802,'64CB' => 25803,
    '64CF' => 25807,'64D4' => 25812,
    '64D6' => 25814,'64D9' => 25817,
    '64DA' => 25818,'64DB' => 25819,
    '64E0' => 25824,'64E3' => 25827,
    '64E8' => 25832,'64E9' => 25833,
    '64EB' => 25835,'64EC' => 25836,
    '64ED' => 25837,'64EF' => 25839,
    '64F0' => 25840,'64F1' => 25841,
    '64F2' => 25842,'64F3' => 25843,
    '64F4' => 25844,'64F7' => 25847,
    '64F8' => 25848,'64FA' => 25850,
    '64FB' => 25851,'64FC' => 25852,
    '64FD' => 25853,'64FE' => 25854,
    '64FF' => 25855,'6501' => 25857,
    '6503' => 25859,'6504' => 25860,
    '6506' => 25862,'6507' => 25863,
    '650C' => 25868,'650D' => 25869,
    '650E' => 25870,'650F' => 25871,
    '6510' => 25872,'6513' => 25875,
    '6514' => 25876,'6515' => 25877,
    '6516' => 25878,'6517' => 25879,
    '6519' => 25881,'651B' => 25883,
    '651C' => 25884,'651D' => 25885,
    '6520' => 25888,'6521' => 25889,
    '6522' => 25890,'6523' => 25891,
    '6524' => 25892,'6526' => 25894,
    '6529' => 25897,'652A' => 25898,
    '652C' => 25900,'652D' => 25901,
    '6532' => 25906,'6533' => 25907,
    '6537' => 25911,'653D' => 25917,
    '6541' => 25921,'6543' => 25923,
    '6546' => 25926,'654A' => 25930,
    '6553' => 25939,'6554' => 25940,
    '6557' => 25943,'6558' => 25944,
    '655C' => 25948,'6564' => 25956,
    '6565' => 25957,'6567' => 25959,
    '6568' => 25960,'656A' => 25962,
    '656F' => 25967,'6573' => 25971,
    '6575' => 25973,'6576' => 25974,
    '6578' => 25976,'6579' => 25977,
    '657A' => 25978,'657B' => 25979,
    '657C' => 25980,'657F' => 25983,
    '6580' => 25984,'6581' => 25985,
    '6582' => 25986,'6583' => 25987,
    '6584' => 25988,'6592' => 26002,
    '6594' => 26004,'6595' => 26005,
    '6596' => 26006,'659D' => 26013,
    '659E' => 26014,'65A0' => 26016,
    '65A2' => 26018,'65A8' => 26024,
    '65AA' => 26026,'65AC' => 26028,
    '65AE' => 26030,'65B2' => 26034,
    '65B3' => 26035,'65B6' => 26038,
    '65B7' => 26039,'65B8' => 26040,
    '65BB' => 26043,'65BF' => 26047,
    '65C2' => 26050,'65CD' => 26061,
    '65D0' => 26064,'65D3' => 26067,
    '65DA' => 26074,'65DB' => 26075,
    '65DD' => 26077,'65DE' => 26078,
    '65DF' => 26079,'65E1' => 26081,
    '65F2' => 26098,'65F3' => 26099,
    '65F4' => 26100,'65F5' => 26101,
    '65FB' => 26107,'65FC' => 26108,
    '65FD' => 26109,'6604' => 26116,
    '6605' => 26117,'6607' => 26119,
    '6608' => 26120,'6609' => 26121,
    '660B' => 26123,'660D' => 26125,
    '6610' => 26128,'6611' => 26129,
    '6612' => 26130,'661C' => 26140,
    '6621' => 26145,'6622' => 26146,
    '6624' => 26148,'6626' => 26150,
    '662B' => 26155,'662E' => 26158,
    '6632' => 26162,'6633' => 26163,
    '6639' => 26169,'663A' => 26170,
    '6642' => 26178,'6645' => 26181,
    '6647' => 26183,'6649' => 26185,
    '664A' => 26186,'6651' => 26193,
    '6659' => 26201,'665B' => 26203,
    '665C' => 26204,'665D' => 26205,
    '665E' => 26206,'6662' => 26210,
    '6665' => 26213,'666A' => 26218,
    '666C' => 26220,'6671' => 26225,
    '6672' => 26226,'6678' => 26232,
    '6679' => 26233,'667B' => 26235,
    '667C' => 26236,'6680' => 26240,
    '6686' => 26246,'6688' => 26248,
    '6689' => 26249,'668A' => 26250,
    '668B' => 26251,'668D' => 26253,
    '6690' => 26256,'6694' => 26260,
    '6695' => 26261,'6698' => 26264,
    '6699' => 26265,'669F' => 26271,
    '66A0' => 26272,'66A1' => 26273,
    '66A2' => 26274,'66A9' => 26281,
    '66AA' => 26282,'66AB' => 26283,
    '66AF' => 26287,'66B0' => 26288,
    '66B1' => 26289,'66B2' => 26290,
    '66B5' => 26293,'66B7' => 26295,
    '66B8' => 26296,'66BA' => 26298,
    '66BB' => 26299,'66BD' => 26301,
    '66C0' => 26304,'66C4' => 26308,
    '66C6' => 26310,'66C7' => 26311,
    '66C8' => 26312,'66C9' => 26313,
    '66CA' => 26314,'66CB' => 26315,
    '66CC' => 26316,'66CF' => 26319,
    '66D2' => 26322,'66D6' => 26326,
    '66D8' => 26328,'66DA' => 26330,
    '66DE' => 26334,'66E0' => 26336,
    '66E3' => 26339,'66E4' => 26340,
    '66E8' => 26344,'66EB' => 26347,
    '66EC' => 26348,'66ED' => 26349,
    '66EE' => 26350,'66F6' => 26358,
    '66F8' => 26360,'6701' => 26369,
    '6703' => 26371,'6704' => 26372,
    '6705' => 26373,'670F' => 26383,
    '6712' => 26386,'6713' => 26387,
    '6718' => 26392,'6720' => 26400,
    '6721' => 26401,'6722' => 26402,
    '6723' => 26403,'6727' => 26407,
    '672E' => 26414,'6733' => 26419,
    '6738' => 26424,'6739' => 26425,
    '673B' => 26427,'673C' => 26428,
    '673E' => 26430,'673F' => 26431,
    '6745' => 26437,'6747' => 26439,
    '674B' => 26443,'674D' => 26445,
    '6755' => 26453,'6757' => 26455,
    '6759' => 26457,'675A' => 26458,
    '675D' => 26461,'676C' => 26476,
    '6771' => 26481,'6774' => 26484,
    '6776' => 26486,'6778' => 26488,
    '6779' => 26489,'677A' => 26490,
    '677B' => 26491,'677D' => 26493,
    '6783' => 26499,'6785' => 26501,
    '6786' => 26502,'678C' => 26508,
    '678D' => 26509,'678E' => 26510,
    '6791' => 26513,'6792' => 26514,
    '6793' => 26515,'6794' => 26516,
    '6799' => 26521,'679F' => 26527,
    '67AE' => 26542,'67B2' => 26546,
    '67B4' => 26548,'67B9' => 26553,
    '67BA' => 26554,'67BB' => 26555,
    '67C0' => 26560,'67C2' => 26562,
    '67C5' => 26565,'67C6' => 26566,
    '67C8' => 26568,'67C9' => 26569,
    '67CA' => 26570,'67CB' => 26571,
    '67CC' => 26572,'67CD' => 26573,
    '67CE' => 26574,'67DB' => 26587,
    '67DF' => 26591,'67E3' => 26595,
    '67E4' => 26596,'67E6' => 26598,
    '67E7' => 26599,'67EA' => 26602,
    '67EB' => 26603,'67ED' => 26605,
    '67EE' => 26606,'67F2' => 26610,
    '67F5' => 26613,'67F6' => 26614,
    '67F7' => 26615,'67F8' => 26616,
    '67FA' => 26618,'67FC' => 26620,
    '6812' => 26642,'6814' => 26644,
    '6818' => 26648,'681A' => 26650,
    '681C' => 26652,'681F' => 26655,
    '6820' => 26656,'6825' => 26661,
    '6826' => 26662,'6828' => 26664,
    '682B' => 26667,'682D' => 26669,
    '682E' => 26670,'682F' => 26671,
    '6831' => 26673,'6834' => 26676,
    '6835' => 26677,'683A' => 26682,
    '683B' => 26683,'684B' => 26699,
    '684D' => 26701,'684F' => 26703,
    '686D' => 26733,'686E' => 26734,
    '686F' => 26735,'6871' => 26737,
    '6872' => 26738,'6875' => 26741,
    '6878' => 26744,'6879' => 26745,
    '687B' => 26747,'687C' => 26748,
    '687D' => 26749,'687E' => 26750,
    '687F' => 26751,'6880' => 26752,
    '6882' => 26754,'6887' => 26759,
    '6889' => 26761,'688A' => 26762,
    '688B' => 26763,'688C' => 26764,
    '6890' => 26768,'6891' => 26769,
    '6892' => 26770,'6894' => 26772,
    '6896' => 26774,'689B' => 26779,
    '689C' => 26780,'689D' => 26781,
    '689F' => 26783,'68A0' => 26784,
    '68A1' => 26785,'68A3' => 26787,
    '68A4' => 26788,'68A9' => 26793,
    '68AA' => 26794,'68AB' => 26795,
    '68AC' => 26796,'68AE' => 26798,
    '68B1' => 26801,'68B2' => 26802,
    '68B4' => 26804,'68C4' => 26820,
    '68C6' => 26822,'68C7' => 26823,
    '68C8' => 26824,'68CC' => 26828,
    '68CE' => 26830,'68D0' => 26832,
    '68D1' => 26833,'68D3' => 26835,
    '68D4' => 26836,'68D6' => 26838,
    '68D7' => 26839,'68DC' => 26844,
    '68DD' => 26845,'68DE' => 26846,
    '68DF' => 26847,'68E1' => 26849,
    '68E4' => 26852,'68E6' => 26854,
    '68E7' => 26855,'68E8' => 26856,
    '68E9' => 26857,'68EA' => 26858,
    '68EB' => 26859,'68EC' => 26860,
    '68EF' => 26863,'68F2' => 26866,
    '68F3' => 26867,'68F4' => 26868,
    '68F6' => 26870,'68F7' => 26871,
    '68F8' => 26872,'68FB' => 26875,
    '68FD' => 26877,'6904' => 26884,
    '6906' => 26886,'6907' => 26887,
    '6908' => 26888,'690A' => 26890,
    '690C' => 26892,'690F' => 26895,
    '6911' => 26897,'6913' => 26899,
    '6914' => 26900,'6915' => 26901,
    '6917' => 26903,'6925' => 26917,
    '692A' => 26922,'692F' => 26927,
    '6932' => 26930,'6933' => 26931,
    '6935' => 26933,'6937' => 26935,
    '6938' => 26936,'693B' => 26939,
    '693C' => 26940,'6940' => 26944,
    '6941' => 26945,'6944' => 26948,
    '6945' => 26949,'6948' => 26952,
    '6949' => 26953,'694A' => 26954,
    '694B' => 26955,'694C' => 26956,
    '694E' => 26958,'694F' => 26959,
    '6951' => 26961,'6952' => 26962,
    '6953' => 26963,'6956' => 26966,
    '6958' => 26968,'6959' => 26969,
    '695B' => 26971,'695C' => 26972,
    '695F' => 26975,'6962' => 26978,
    '6965' => 26981,'6968' => 26984,
    '6969' => 26985,'696A' => 26986,
    '696C' => 26988,'696D' => 26989,
    '696F' => 26991,'6970' => 26992,
    '6974' => 26996,'6975' => 26997,
    '6976' => 26998,'697A' => 27002,
    '697B' => 27003,'6983' => 27011,
    '698E' => 27022,'6990' => 27024,
    '6991' => 27025,'6993' => 27027,
    '6996' => 27030,'6997' => 27031,
    '6999' => 27033,'699A' => 27034,
    '699E' => 27038,'69A0' => 27040,
    '69A1' => 27041,'69A3' => 27043,
    '69A4' => 27044,'69A5' => 27045,
    '69A6' => 27046,'69A9' => 27049,
    '69AA' => 27050,'69AC' => 27052,
    '69AE' => 27054,'69AF' => 27055,
    '69B0' => 27056,'69B3' => 27059,
    '69B5' => 27061,'69B6' => 27062,
    '69B9' => 27065,'69BC' => 27068,
    '69BD' => 27069,'69BE' => 27070,
    '69BF' => 27071,'69C2' => 27074,
    '69C3' => 27075,'69C4' => 27076,
    '69C6' => 27078,'69C9' => 27081,
    '69CB' => 27083,'69CD' => 27085,
    '69CF' => 27087,'69D3' => 27091,
    '69D9' => 27097,'69E2' => 27106,
    '69E4' => 27108,'69E5' => 27109,
    '69E6' => 27110,'69E7' => 27111,
    '69E8' => 27112,'69EB' => 27115,
    '69EC' => 27116,'69EE' => 27118,
    '69F1' => 27121,'69F3' => 27123,
    '69F4' => 27124,'69F6' => 27126,
    '69F7' => 27127,'69F8' => 27128,
    '69FB' => 27131,'69FC' => 27132,
    '69FE' => 27134,'6A00' => 27136,
    '6A01' => 27137,'6A02' => 27138,
    '6A04' => 27140,'6A05' => 27141,
    '6A06' => 27142,'6A07' => 27143,
    '6A08' => 27144,'6A09' => 27145,
    '6A0D' => 27149,'6A0F' => 27151,
    '6A11' => 27153,'6A13' => 27155,
    '6A14' => 27156,'6A15' => 27157,
    '6A16' => 27158,'6A19' => 27161,
    '6A1B' => 27163,'6A1D' => 27165,
    '6A1E' => 27166,'6A20' => 27168,
    '6A23' => 27171,'6A25' => 27173,
    '6A26' => 27174,'6A27' => 27175,
    '6A32' => 27186,'6A34' => 27188,
    '6A38' => 27192,'6A39' => 27193,
    '6A3A' => 27194,'6A3B' => 27195,
    '6A3C' => 27196,'6A3F' => 27199,
    '6A40' => 27200,'6A41' => 27201,
    '6A46' => 27206,'6A48' => 27208,
    '6A49' => 27209,'6A4B' => 27211,
    '6A4D' => 27213,'6A4E' => 27214,
    '6A4F' => 27215,'6A51' => 27217,
    '6A54' => 27220,'6A55' => 27221,
    '6A56' => 27222,'6A5A' => 27226,
    '6A5D' => 27229,'6A5E' => 27230,
    '6A5F' => 27231,'6A60' => 27232,
    '6A62' => 27234,'6A64' => 27236,
    '6A66' => 27238,'6A67' => 27239,
    '6A68' => 27240,'6A69' => 27241,
    '6A6A' => 27242,'6A6B' => 27243,
    '6A6D' => 27245,'6A6F' => 27247,
    '6A76' => 27254,'6A7E' => 27262,
    '6A7F' => 27263,'6A81' => 27265,
    '6A83' => 27267,'6A85' => 27269,
    '6A87' => 27271,'6A89' => 27273,
    '6A8C' => 27276,'6A8D' => 27277,
    '6A92' => 27282,'6A93' => 27283,
    '6A94' => 27284,'6A95' => 27285,
    '6A96' => 27286,'6A9A' => 27290,
    '6A9B' => 27291,'6A9C' => 27292,
    '6A9E' => 27294,'6A9F' => 27295,
    '6AA1' => 27297,'6AA2' => 27298,
    '6AA3' => 27299,'6AA4' => 27300,
    '6AA5' => 27301,'6AA6' => 27302,
    '6AA8' => 27304,'6AAD' => 27309,
    '6AAE' => 27310,'6AAF' => 27311,
    '6AB3' => 27315,'6AB4' => 27316,
    '6AB6' => 27318,'6AB7' => 27319,
    '6AB8' => 27320,'6AB9' => 27321,
    '6ABA' => 27322,'6ABB' => 27323,
    '6ABD' => 27325,'6AC2' => 27330,
    '6AC3' => 27331,'6AC5' => 27333,
    '6AC6' => 27334,'6AC7' => 27335,
    '6ACB' => 27339,'6ACC' => 27340,
    '6ACD' => 27341,'6ACF' => 27343,
    '6AD0' => 27344,'6AD1' => 27345,
    '6AD3' => 27347,'6AD9' => 27353,
    '6ADA' => 27354,'6ADB' => 27355,
    '6ADC' => 27356,'6ADD' => 27357,
    '6ADE' => 27358,'6ADF' => 27359,
    '6AE0' => 27360,'6AE1' => 27361,
    '6AE5' => 27365,'6AE7' => 27367,
    '6AE8' => 27368,'6AEA' => 27370,
    '6AEB' => 27371,'6AEC' => 27372,
    '6AEE' => 27374,'6AEF' => 27375,
    '6AF0' => 27376,'6AF1' => 27377,
    '6AF3' => 27379,'6AF8' => 27384,
    '6AF9' => 27385,'6AFA' => 27386,
    '6AFB' => 27387,'6AFC' => 27388,
    '6B00' => 27392,'6B02' => 27394,
    '6B03' => 27395,'6B04' => 27396,
    '6B08' => 27400,'6B09' => 27401,
    '6B0A' => 27402,'6B0B' => 27403,
    '6B0F' => 27407,'6B10' => 27408,
    '6B11' => 27409,'6B12' => 27410,
    '6B13' => 27411,'6B16' => 27414,
    '6B17' => 27415,'6B18' => 27416,
    '6B19' => 27417,'6B1A' => 27418,
    '6B1E' => 27422,'6B25' => 27429,
    '6B28' => 27432,'6B2C' => 27436,
    '6B2D' => 27437,'6B2F' => 27439,
    '6B31' => 27441,'6B33' => 27443,
    '6B34' => 27444,'6B36' => 27446,
    '6B38' => 27448,'6B3B' => 27451,
    '6B3C' => 27452,'6B3D' => 27453,
    '6B3F' => 27455,'6B41' => 27457,
    '6B42' => 27458,'6B45' => 27461,
    '6B48' => 27464,'6B4A' => 27466,
    '6B4B' => 27467,'6B4D' => 27469,
    '6B4E' => 27470,'6B50' => 27472,
    '6B51' => 27473,'6B54' => 27476,
    '6B55' => 27477,'6B56' => 27478,
    '6B5B' => 27483,'6B5C' => 27484,
    '6B5E' => 27486,'6B5F' => 27487,
    '6B60' => 27488,'6B61' => 27489,
    '6B6D' => 27501,'6B72' => 27506,
    '6B76' => 27510,'6B77' => 27511,
    '6B78' => 27512,'6B7E' => 27518,
    '6B7F' => 27519,'6B80' => 27520,
    '6B88' => 27528,'6B8C' => 27532,
    '6B8E' => 27534,'6B8F' => 27535,
    '6B91' => 27537,'6B94' => 27540,
    '6B95' => 27541,'6B97' => 27543,
    '6B98' => 27544,'6B99' => 27545,
    '6B9E' => 27550,'6B9F' => 27551,
    '6BA0' => 27552,'6BA2' => 27554,
    '6BA3' => 27555,'6BA4' => 27556,
    '6BA5' => 27557,'6BA6' => 27558,
    '6BA7' => 27559,'6BAB' => 27563,
    '6BAD' => 27565,'6BAE' => 27566,
    '6BAF' => 27567,'6BB0' => 27568,
    '6BB2' => 27570,'6BB6' => 27574,
    '6BBA' => 27578,'6BBC' => 27580,
    '6BBD' => 27581,'6BC0' => 27584,
    '6BC3' => 27587,'6BC4' => 27588,
    '6BC6' => 27590,'6BC7' => 27591,
    '6BC8' => 27592,'6BC9' => 27593,
    '6BCA' => 27594,'6BCC' => 27596,
    '6BD0' => 27600,'6BD8' => 27608,
    '6BDA' => 27610,'6BDE' => 27614,
    '6BE0' => 27616,'6BE2' => 27618,
    '6BE3' => 27619,'6BE4' => 27620,
    '6BE6' => 27622,'6BE7' => 27623,
    '6BE8' => 27624,'6BEC' => 27628,
    '6BF0' => 27632,'6BF2' => 27634,
    '6BF7' => 27639,'6BF8' => 27640,
    '6BFB' => 27643,'6BFC' => 27644,
    '6BFE' => 27646,'6BFF' => 27647,
    '6C00' => 27648,'6C01' => 27649,
    '6C02' => 27650,'6C03' => 27651,
    '6C04' => 27652,'6C08' => 27656,
    '6C09' => 27657,'6C0B' => 27659,
    '6C0C' => 27660,'6C1D' => 27677,
    '6C20' => 27680,'6C23' => 27683,
    '6C25' => 27685,'6C2B' => 27691,
    '6C2C' => 27692,'6C33' => 27699,
    '6C36' => 27702,'6C3B' => 27707,
    '6C3E' => 27710,'6C3F' => 27711,
    '6C43' => 27715,'6C4B' => 27723,
    '6C4C' => 27724,'6C4D' => 27725,
    '6C4E' => 27726,'6C4F' => 27727,
    '6C52' => 27730,'6C59' => 27737,
    '6C65' => 27749,'6C66' => 27750,
    '6C67' => 27751,'6C6B' => 27755,
    '6C6D' => 27757,'6C6F' => 27759,
    '6C71' => 27761,'6C73' => 27763,
    '6C78' => 27768,'6C7A' => 27770,
    '6C7B' => 27771,'6C80' => 27776,
    '6C84' => 27780,'6C87' => 27783,
    '6C8A' => 27786,'6C8B' => 27787,
    '6C8D' => 27789,'6C8E' => 27790,
    '6C92' => 27794,'6C95' => 27797,
    '6C96' => 27798,'6C98' => 27800,
    '6C9A' => 27802,'6C9C' => 27804,
    '6C9D' => 27805,'6CAC' => 27820,
    '6CB0' => 27824,'6CB4' => 27828,
    '6CB6' => 27830,'6CB7' => 27831,
    '6CBA' => 27834,'6CC0' => 27840,
    '6CC1' => 27841,'6CC2' => 27842,
    '6CC3' => 27843,'6CC6' => 27846,
    '6CC7' => 27847,'6CCD' => 27853,
    '6CCF' => 27855,'6CD1' => 27857,
    '6CD2' => 27858,'6CD9' => 27865,
    '6CDA' => 27866,'6CDC' => 27868,
    '6CDD' => 27869,'6CE7' => 27879,
    '6CE9' => 27881,'6CEC' => 27884,
    '6CED' => 27885,'6CF2' => 27890,
    '6CF9' => 27897,'6D00' => 27904,
    '6D03' => 27907,'6D08' => 27912,
    '6D09' => 27913,'6D0A' => 27914,
    '6D0D' => 27917,'6D0F' => 27919,
    '6D10' => 27920,'6D11' => 27921,
    '6D16' => 27926,'6D18' => 27928,
    '6D1D' => 27933,'6D1F' => 27935,
    '6D20' => 27936,'6D22' => 27938,
    '6D28' => 27944,'6D29' => 27945,
    '6D2C' => 27948,'6D2D' => 27949,
    '6D2F' => 27951,'6D30' => 27952,
    '6D34' => 27956,'6D36' => 27958,
    '6D37' => 27959,'6D38' => 27960,
    '6D3A' => 27962,'6D3F' => 27967,
    '6D40' => 27968,'6D42' => 27970,
    '6D58' => 27992,'6D5F' => 27999,
    '6D61' => 28001,'6D62' => 28002,
    '6D64' => 28004,'6D65' => 28005,
    '6D67' => 28007,'6D68' => 28008,
    '6D6C' => 28012,'6D6D' => 28013,
    '6D70' => 28016,'6D75' => 28021,
    '6D76' => 28022,'6D79' => 28025,
    '6D7A' => 28026,'6D7B' => 28027,
    '6D7D' => 28029,'6D7E' => 28030,
    '6D7F' => 28031,'6D80' => 28032,
    '6D83' => 28035,'6D84' => 28036,
    '6D86' => 28038,'6D87' => 28039,
    '6D8A' => 28042,'6D8B' => 28043,
    '6D8D' => 28045,'6D90' => 28048,
    '6D92' => 28050,'6D97' => 28055,
    '6D98' => 28056,'6DAC' => 28076,
    '6DB3' => 28083,'6DB4' => 28084,
    '6DB7' => 28087,'6DBA' => 28090,
    '6DBB' => 28091,'6DBC' => 28092,
    '6DBD' => 28093,'6DBE' => 28094,
    '6DC2' => 28098,'6DC8' => 28104,
    '6DC9' => 28105,'6DCA' => 28106,
    '6DCD' => 28109,'6DCF' => 28111,
    '6DD0' => 28112,'6DD2' => 28114,
    '6DD3' => 28115,'6DD4' => 28116,
    '6DD5' => 28117,'6DD7' => 28119,
    '6DDA' => 28122,'6DDB' => 28123,
    '6DDC' => 28124,'6DDF' => 28127,
    '6DE2' => 28130,'6DE3' => 28131,
    '6DE5' => 28133,'6DE8' => 28136,
    '6DE9' => 28137,'6DEA' => 28138,
    '6DED' => 28141,'6DEF' => 28143,
    '6DF0' => 28144,'6DF2' => 28146,
    '6DF4' => 28148,'6DF5' => 28149,
    '6DF6' => 28150,'6DFA' => 28154,
    '6DFD' => 28157,'6E00' => 28160,
    '6E03' => 28163,'6E19' => 28185,
    '6E1B' => 28187,'6E1C' => 28188,
    '6E1F' => 28191,'6E22' => 28194,
    '6E26' => 28198,'6E27' => 28199,
    '6E28' => 28200,'6E2C' => 28204,
    '6E2E' => 28206,'6E30' => 28208,
    '6E31' => 28209,'6E33' => 28211,
    '6E35' => 28213,'6E36' => 28214,
    '6E39' => 28217,'6E3B' => 28219,
    '6E3C' => 28220,'6E3D' => 28221,
    '6E3E' => 28222,'6E3F' => 28223,
    '6E40' => 28224,'6E41' => 28225,
    '6E45' => 28229,'6E46' => 28230,
    '6E47' => 28231,'6E49' => 28233,
    '6E4A' => 28234,'6E4B' => 28235,
    '6E51' => 28241,'6E52' => 28242,
    '6E55' => 28245,'6E5A' => 28250,
    '6E5C' => 28252,'6E5D' => 28253,
    '6E5E' => 28254,'6E60' => 28256,
    '6E61' => 28257,'6E62' => 28258,
    '6E63' => 28259,'6E64' => 28260,
    '6E65' => 28261,'6E66' => 28262,
    '6E67' => 28263,'6E68' => 28264,
    '6E69' => 28265,'6E6F' => 28271,
    '6E71' => 28273,'6E72' => 28274,
    '6E73' => 28275,'6E74' => 28276,
    '6E77' => 28279,'6E78' => 28280,
    '6E79' => 28281,'6E88' => 28296,
    '6E8D' => 28301,'6E8E' => 28302,
    '6E92' => 28306,'6E93' => 28307,
    '6E94' => 28308,'6E96' => 28310,
    '6E97' => 28311,'6E99' => 28313,
    '6E9B' => 28315,'6E9D' => 28317,
    '6E9E' => 28318,'6EA0' => 28320,
    '6EA1' => 28321,'6EA3' => 28323,
    '6EA4' => 28324,'6EA6' => 28326,
    '6EAB' => 28331,'6EAE' => 28334,
    '6EB0' => 28336,'6EB3' => 28339,
    '6EB9' => 28345,'6EBC' => 28348,
    '6EBE' => 28350,'6EBF' => 28351,
    '6EC0' => 28352,'6EC3' => 28355,
    '6EC4' => 28356,'6EC5' => 28357,
    '6EC6' => 28358,'6EC8' => 28360,
    '6EC9' => 28361,'6ECA' => 28362,
    '6ECC' => 28364,'6ECD' => 28365,
    '6ECE' => 28366,'6ED0' => 28368,
    '6ED2' => 28370,'6ED6' => 28374,
    '6ED8' => 28376,'6EDC' => 28380,
    '6EEB' => 28395,'6EEC' => 28396,
    '6EED' => 28397,'6EEE' => 28398,
    '6EEF' => 28399,'6EF1' => 28401,
    '6EF2' => 28402,'6EF5' => 28405,
    '6EF6' => 28406,'6EF7' => 28407,
    '6EF8' => 28408,'6EFB' => 28411,
    '6EFC' => 28412,'6EFD' => 28413,
    '6EFE' => 28414,'6EFF' => 28415,
    '6F00' => 28416,'6F01' => 28417,
    '6F03' => 28419,'6F05' => 28421,
    '6F07' => 28423,'6F08' => 28424,
    '6F0A' => 28426,'6F0D' => 28429,
    '6F0E' => 28430,'6F12' => 28434,
    '6F18' => 28440,'6F19' => 28441,
    '6F1A' => 28442,'6F1C' => 28444,
    '6F1E' => 28446,'6F1F' => 28447,
    '6F21' => 28449,'6F22' => 28450,
    '6F23' => 28451,'6F25' => 28453,
    '6F26' => 28454,'6F27' => 28455,
    '6F2C' => 28460,'6F2E' => 28462,
    '6F30' => 28464,'6F32' => 28466,
    '6F35' => 28469,'6F37' => 28471,
    '6F38' => 28472,'6F39' => 28473,
    '6F3A' => 28474,'6F3B' => 28475,
    '6F3C' => 28476,'6F3F' => 28479,
    '6F40' => 28480,'6F41' => 28481,
    '6F43' => 28483,'6F4E' => 28494,
    '6F4F' => 28495,'6F50' => 28496,
    '6F51' => 28497,'6F52' => 28498,
    '6F53' => 28499,'6F54' => 28500,
    '6F55' => 28501,'6F57' => 28503,
    '6F5A' => 28506,'6F5B' => 28507,
    '6F5D' => 28509,'6F5F' => 28511,
    '6F60' => 28512,'6F61' => 28513,
    '6F63' => 28515,'6F64' => 28516,
    '6F67' => 28519,'6F69' => 28521,
    '6F6A' => 28522,'6F6B' => 28523,
    '6F6C' => 28524,'6F6F' => 28527,
    '6F70' => 28528,'6F73' => 28531,
    '6F76' => 28534,'6F77' => 28535,
    '6F7B' => 28539,'6F7D' => 28541,
    '6F7E' => 28542,'6F7F' => 28543,
    '6F80' => 28544,'6F82' => 28546,
    '6F85' => 28549,'6F86' => 28550,
    '6F87' => 28551,'6F8B' => 28555,
    '6F90' => 28560,'6F92' => 28562,
    '6F93' => 28563,'6F94' => 28564,
    '6F95' => 28565,'6F96' => 28566,
    '6F97' => 28567,'6F9E' => 28574,
    '6FA0' => 28576,'6FA2' => 28578,
    '6FA3' => 28579,'6FA4' => 28580,
    '6FA5' => 28581,'6FA6' => 28582,
    '6FA8' => 28584,'6FA9' => 28585,
    '6FAA' => 28586,'6FAB' => 28587,
    '6FAC' => 28588,'6FAD' => 28589,
    '6FAE' => 28590,'6FAF' => 28591,
    '6FB0' => 28592,'6FB1' => 28593,
    '6FB2' => 28594,'6FB4' => 28596,
    '6FB8' => 28600,'6FBA' => 28602,
    '6FBC' => 28604,'6FBD' => 28605,
    '6FBF' => 28607,'6FC1' => 28609,
    '6FC3' => 28611,'6FC4' => 28612,
    '6FC6' => 28614,'6FC7' => 28615,
    '6FC8' => 28616,'6FCA' => 28618,
    '6FCB' => 28619,'6FCC' => 28620,
    '6FCD' => 28621,'6FCE' => 28622,
    '6FCF' => 28623,'6FD4' => 28628,
    '6FD5' => 28629,'6FD8' => 28632,
    '6FDB' => 28635,'6FDC' => 28636,
    '6FDD' => 28637,'6FDF' => 28639,
    '6FE2' => 28642,'6FE3' => 28643,
    '6FE4' => 28644,'6FE6' => 28646,
    '6FE7' => 28647,'6FE8' => 28648,
    '6FE9' => 28649,'6FEB' => 28651,
    '6FEC' => 28652,'6FED' => 28653,
    '6FF0' => 28656,'6FF1' => 28657,
    '6FF2' => 28658,'6FF4' => 28660,
    '6FF7' => 28663,'6FFA' => 28666,
    '6FFB' => 28667,'6FFC' => 28668,
    '6FFE' => 28670,'6FFF' => 28671,
    '7000' => 28672,'7001' => 28673,
    '7004' => 28676,'7005' => 28677,
    '7006' => 28678,'7007' => 28679,
    '7009' => 28681,'700A' => 28682,
    '700B' => 28683,'700C' => 28684,
    '700D' => 28685,'700E' => 28686,
    '700F' => 28687,'7014' => 28692,
    '7015' => 28693,'7016' => 28694,
    '7017' => 28695,'7018' => 28696,
    '7019' => 28697,'701C' => 28700,
    '701D' => 28701,'701F' => 28703,
    '7020' => 28704,'7021' => 28705,
    '7022' => 28706,'7024' => 28708,
    '7026' => 28710,'7027' => 28711,
    '7028' => 28712,'7029' => 28713,
    '702A' => 28714,'702B' => 28715,
    '702F' => 28719,'7030' => 28720,
    '7031' => 28721,'7032' => 28722,
    '7033' => 28723,'7034' => 28724,
    '7037' => 28727,'7038' => 28728,
    '703A' => 28730,'703B' => 28731,
    '703C' => 28732,'703E' => 28734,
    '703F' => 28735,'7040' => 28736,
    '7041' => 28737,'7042' => 28738,
    '7043' => 28739,'7044' => 28740,
    '7045' => 28741,'7046' => 28742,
    '7048' => 28744,'7049' => 28745,
    '704A' => 28746,'7051' => 28753,
    '7052' => 28754,'7055' => 28757,
    '7056' => 28758,'7057' => 28759,
    '7058' => 28760,'705A' => 28762,
    '705B' => 28763,'705D' => 28765,
    '705F' => 28767,'7060' => 28768,
    '7061' => 28769,'7062' => 28770,
    '7063' => 28771,'7064' => 28772,
    '7065' => 28773,'7066' => 28774,
    '7068' => 28776,'7069' => 28777,
    '706A' => 28778,'7071' => 28785,
    '7074' => 28788,'707A' => 28794,
    '707D' => 28797,'7082' => 28802,
    '7083' => 28803,'7084' => 28804,
    '7086' => 28806,'7091' => 28817,
    '7093' => 28819,'7098' => 28824,
    '709A' => 28826,'709F' => 28831,
    '70A1' => 28833,'70A4' => 28836,
    '70A9' => 28841,'70B0' => 28848,
    '70B4' => 28852,'70B5' => 28853,
    '70BA' => 28858,'70BE' => 28862,
    '70C5' => 28869,'70C6' => 28870,
    '70C7' => 28871,'70CB' => 28875,
    '70CD' => 28877,'70CE' => 28878,
    '70CF' => 28879,'70D1' => 28881,
    '70D2' => 28882,'70D3' => 28883,
    '70D4' => 28884,'70D7' => 28887,
    '70DA' => 28890,'70DC' => 28892,
    '70DD' => 28893,'70DE' => 28894,
    '70E0' => 28896,'70E1' => 28897,
    '70E2' => 28898,'70F0' => 28912,
    '70F3' => 28915,'70F4' => 28916,
    '70F6' => 28918,'70F8' => 28920,
    '70FA' => 28922,'70FB' => 28923,
    '70FC' => 28924,'70FF' => 28927,
    '7100' => 28928,'7102' => 28930,
    '7104' => 28932,'7106' => 28934,
    '710B' => 28939,'710C' => 28940,
    '710D' => 28941,'710E' => 28942,
    '7117' => 28951,'711B' => 28955,
    '711C' => 28956,'711E' => 28958,
    '711F' => 28959,'7120' => 28960,
    '7121' => 28961,'7122' => 28962,
    '7123' => 28963,'7125' => 28965,
    '7128' => 28968,'712E' => 28974,
    '7132' => 28978,'713A' => 28986,
    '7141' => 28993,'7142' => 28994,
    '7143' => 28995,'7144' => 28996,
    '7146' => 28998,'7147' => 28999,
    '7149' => 29001,'714B' => 29003,
    '714D' => 29005,'7150' => 29008,
    '7152' => 29010,'7153' => 29011,
    '7154' => 29012,'7156' => 29014,
    '7158' => 29016,'7159' => 29017,
    '715A' => 29018,'715D' => 29021,
    '715F' => 29023,'7160' => 29024,
    '7161' => 29025,'7162' => 29026,
    '7163' => 29027,'7165' => 29029,
    '7169' => 29033,'716A' => 29034,
    '716C' => 29036,'7170' => 29040,
    '717B' => 29051,'7180' => 29056,
    '7181' => 29057,'7182' => 29058,
    '7185' => 29061,'7186' => 29062,
    '7187' => 29063,'7189' => 29065,
    '7190' => 29072,'7192' => 29074,
    '7197' => 29079,'719A' => 29082,
    '719B' => 29083,'719C' => 29084,
    '719D' => 29085,'719E' => 29086,
    '71A1' => 29089,'71A4' => 29092,
    '71A5' => 29093,'71A7' => 29095,
    '71A9' => 29097,'71AA' => 29098,
    '71AF' => 29103,'71B0' => 29104,
    '71B1' => 29105,'71B2' => 29106,
    '71B8' => 29112,'71BC' => 29116,
    '71BD' => 29117,'71BE' => 29118,
    '71BF' => 29119,'71C0' => 29120,
    '71C1' => 29121,'71C2' => 29122,
    '71C4' => 29124,'71C5' => 29125,
    '71C6' => 29126,'71C7' => 29127,
    '71C8' => 29128,'71C9' => 29129,
    '71CA' => 29130,'71CB' => 29131,
    '71CF' => 29135,'71D0' => 29136,
    '71D2' => 29138,'71D6' => 29142,
    '71D8' => 29144,'71D9' => 29145,
    '71DA' => 29146,'71DB' => 29147,
    '71DC' => 29148,'71DF' => 29151,
    '71E1' => 29153,'71E2' => 29154,
    '71E4' => 29156,'71E6' => 29158,
    '71E8' => 29160,'71EC' => 29164,
    '71ED' => 29165,'71F0' => 29168,
    '71F1' => 29169,'71F2' => 29170,
    '71F4' => 29172,'71F8' => 29176,
    '71FB' => 29179,'71FC' => 29180,
    '71FD' => 29181,'71FE' => 29182,
    '71FF' => 29183,'7201' => 29185,
    '7202' => 29186,'7203' => 29187,
    '7205' => 29189,'7207' => 29191,
    '720A' => 29194,'720C' => 29196,
    '720D' => 29197,'7210' => 29200,
    '7213' => 29203,'7214' => 29204,
    '7219' => 29209,'721A' => 29210,
    '721B' => 29211,'721E' => 29214,
    '721F' => 29215,'7222' => 29218,
    '7223' => 29219,'7226' => 29222,
    '7227' => 29223,'7229' => 29225,
    '722D' => 29229,'723A' => 29242,
    '723E' => 29246,'7241' => 29249,
    '7242' => 29250,'7244' => 29252,
    '7246' => 29254,'7249' => 29257,
    '724A' => 29258,'724B' => 29259,
    '724F' => 29263,'7253' => 29267,
    '7258' => 29272,'725A' => 29274,
    '725E' => 29278,'7260' => 29280,
    '7263' => 29283,'726A' => 29290,
    '726C' => 29292,'7270' => 29296,
    '7273' => 29299,'7274' => 29300,
    '7276' => 29302,'7277' => 29303,
    '7278' => 29304,'727B' => 29307,
    '727C' => 29308,'727D' => 29309,
    '7285' => 29317,'7286' => 29318,
    '7288' => 29320,'7289' => 29321,
    '728C' => 29324,'728E' => 29326,
    '7290' => 29328,'7291' => 29329,
    '7293' => 29331,'7295' => 29333,
    '7296' => 29334,'7297' => 29335,
    '7298' => 29336,'729A' => 29338,
    '729B' => 29339,'729D' => 29341,
    '729E' => 29342,'72A1' => 29345,
    '72A2' => 29346,'72A3' => 29347,
    '72A4' => 29348,'72A5' => 29349,
    '72A6' => 29350,'72A7' => 29351,
    '72A8' => 29352,'72A9' => 29353,
    '72AA' => 29354,'72AE' => 29358,
    '72B5' => 29365,'72BA' => 29370,
    '72BD' => 29373,'72BF' => 29375,
    '72C0' => 29376,'72C5' => 29381,
    '72C6' => 29382,'72C9' => 29385,
    '72CA' => 29386,'72CB' => 29387,
    '72CC' => 29388,'72D1' => 29393,
    '72D4' => 29396,'72D6' => 29398,
    '72D8' => 29400,'72DA' => 29402,
    '72DC' => 29404,'72DF' => 29407,
    '72E3' => 29411,'72E4' => 29412,
    '72E6' => 29414,'72EA' => 29418,
    '72EB' => 29419,'72F6' => 29430,
    '72F9' => 29433,'72FD' => 29437,
    '72FE' => 29438,'72FF' => 29439,
    '7300' => 29440,'7307' => 29447,
    '7308' => 29448,'730B' => 29451,
    '730C' => 29452,'730F' => 29455,
    '7311' => 29457,'7312' => 29458,
    '7318' => 29464,'7319' => 29465,
    '7323' => 29475,'7326' => 29478,
    '7327' => 29479,'732D' => 29485,
    '7330' => 29488,'7332' => 29490,
    '7333' => 29491,'7335' => 29493,
    '7336' => 29494,'733A' => 29498,
    '733B' => 29499,'733C' => 29500,
    '7340' => 29504,'7342' => 29506,
    '7343' => 29507,'7344' => 29508,
    '7345' => 29509,'7349' => 29513,
    '734A' => 29514,'734C' => 29516,
    '734E' => 29518,'7351' => 29521,
    '7358' => 29528,'7359' => 29529,
    '735A' => 29530,'735B' => 29531,
    '735D' => 29533,'735E' => 29534,
    '735F' => 29535,'7361' => 29537,
    '7362' => 29538,'7365' => 29541,
    '7366' => 29542,'7367' => 29543,
    '7368' => 29544,'7369' => 29545,
    '736A' => 29546,'736B' => 29547,
    '736E' => 29550,'7370' => 29552,
    '7372' => 29554,'7373' => 29555,
    '7375' => 29557,'7376' => 29558,
    '7377' => 29559,'7378' => 29560,
    '737A' => 29562,'737B' => 29563,
    '737C' => 29564,'737D' => 29565,
    '737F' => 29567,'7380' => 29568,
    '7381' => 29569,'7382' => 29570,
    '7383' => 29571,'7385' => 29573,
    '7386' => 29574,'7388' => 29576,
    '738A' => 29578,'7392' => 29586,
    '7393' => 29587,'7394' => 29588,
    '7395' => 29589,'7397' => 29591,
    '739D' => 29597,'73A0' => 29600,
    '73A1' => 29601,'73A4' => 29604,
    '73A5' => 29605,'73A6' => 29606,
    '73A8' => 29608,'73AC' => 29612,
    '73AD' => 29613,'73B4' => 29620,
    '73B5' => 29621,'73B6' => 29622,
    '73B8' => 29624,'73B9' => 29625,
    '73BC' => 29628,'73BE' => 29630,
    '73BF' => 29631,'73C3' => 29635,
    '73C5' => 29637,'73C6' => 29638,
    '73C7' => 29639,'73CB' => 29643,
    '73CC' => 29644,'73D2' => 29650,
    '73D3' => 29651,'73D4' => 29652,
    '73D6' => 29654,'73D7' => 29655,
    '73D8' => 29656,'73DA' => 29658,
    '73DB' => 29659,'73DC' => 29660,
    '73DD' => 29661,'73E3' => 29667,
    '73E8' => 29672,'73EA' => 29674,
    '73EB' => 29675,'73EE' => 29678,
    '73F4' => 29684,'73F5' => 29685,
    '73F6' => 29686,'73F8' => 29688,
    '73FA' => 29690,'73FC' => 29692,
    '73FD' => 29693,'73FE' => 29694,
    '73FF' => 29695,'7400' => 29696,
    '7401' => 29697,'7404' => 29700,
    '7407' => 29703,'7408' => 29704,
    '740B' => 29707,'740C' => 29708,
    '740D' => 29709,'7416' => 29718,
    '741D' => 29725,'7420' => 29728,
    '7421' => 29729,'7423' => 29731,
    '7424' => 29732,'7429' => 29737,
    '742B' => 29739,'742D' => 29741,
    '742F' => 29743,'7431' => 29745,
    '7432' => 29746,'743A' => 29754,
    '743F' => 29759,'7440' => 29760,
    '7442' => 29762,'7444' => 29764,
    '7446' => 29766,'744A' => 29770,
    '744B' => 29771,'744D' => 29773,
    '744E' => 29774,'744F' => 29775,
    '7450' => 29776,'7451' => 29777,
    '7452' => 29778,'7454' => 29780,
    '7462' => 29794,'7463' => 29795,
    '7464' => 29796,'7467' => 29799,
    '7469' => 29801,'746A' => 29802,
    '746E' => 29806,'746F' => 29807,
    '7471' => 29809,'7472' => 29810,
    '7473' => 29811,'7475' => 29813,
    '7479' => 29817,'747C' => 29820,
    '747D' => 29821,'747F' => 29823,
    '7485' => 29829,'7486' => 29830,
    '7488' => 29832,'7489' => 29833,
    '748A' => 29834,'7492' => 29842,
    '7494' => 29844,'7495' => 29845,
    '7497' => 29847,'7498' => 29848,
    '749A' => 29850,'749F' => 29855,
    '74A0' => 29856,'74A1' => 29857,
    '74A3' => 29859,'74A5' => 29861,
    '74A6' => 29862,'74AA' => 29866,
    '74AB' => 29867,'74AD' => 29869,
    '74AF' => 29871,'74B0' => 29872,
    '74B1' => 29873,'74B2' => 29874,
    '74B5' => 29877,'74B6' => 29878,
    '74B7' => 29879,'74B8' => 29880,
    '74BB' => 29883,'74BD' => 29885,
    '74BE' => 29886,'74BF' => 29887,
    '74C0' => 29888,'74C1' => 29889,
    '74C2' => 29890,'74C3' => 29891,
    '74C5' => 29893,'74CA' => 29898,
    '74CB' => 29899,'74CF' => 29903,
    '74D4' => 29908,'74D5' => 29909,
    '74D6' => 29910,'74D7' => 29911,
    '74D8' => 29912,'74D9' => 29913,
    '74DA' => 29914,'74DB' => 29915,
    '74DD' => 29917,'74DF' => 29919,
    '74E1' => 29921,'74E5' => 29925,
    '74E8' => 29928,'74E9' => 29929,
    '74EC' => 29932,'74F5' => 29941,
    '74FB' => 29947,'74FD' => 29949,
    '74FE' => 29950,'7500' => 29952,
    '7502' => 29954,'7503' => 29955,
    '7507' => 29959,'7508' => 29960,
    '750B' => 29963,'750C' => 29964,
    '7510' => 29968,'7512' => 29970,
    '7514' => 29972,'7515' => 29973,
    '7516' => 29974,'7517' => 29975,
    '751D' => 29981,'7521' => 29985,
    '7522' => 29986,'7526' => 29990,
    '752A' => 29994,'752E' => 29998,
    '7539' => 30009,'753D' => 30013,
    '753F' => 30015,'7547' => 30023,
    '755D' => 30045,'755F' => 30047,
    '7562' => 30050,'7563' => 30051,
    '7564' => 30052,'756B' => 30059,
    '756C' => 30060,'756F' => 30063,
    '7570' => 30064,'7576' => 30070,
    '7577' => 30071,'757D' => 30077,
    '757E' => 30078,'7580' => 30080,
    '7584' => 30084,'7587' => 30087,
    '758A' => 30090,'758C' => 30092,
    '7590' => 30096,'7595' => 30101,
    '7598' => 30104,'75A2' => 30114,
    '75A7' => 30119,'75AA' => 30122,
    '75B6' => 30134,'75BA' => 30138,
    '75BB' => 30139,'75BF' => 30143,
    '75C0' => 30144,'75C1' => 30145,
    '75CB' => 30155,'75CC' => 30156,
    '75CE' => 30158,'75CF' => 30159,
    '75D0' => 30160,'75D1' => 30161,
    '75D7' => 30167,'75D9' => 30169,
    '75DA' => 30170,'75DD' => 30173,
    '75DF' => 30175,'75E0' => 30176,
    '75E1' => 30177,'75ED' => 30189,
    '75EF' => 30191,'75F2' => 30194,
    '75F3' => 30195,'75F5' => 30197,
    '75F6' => 30198,'75F7' => 30199,
    '75F8' => 30200,'75FA' => 30202,
    '75FB' => 30203,'75FD' => 30205,
    '75FE' => 30206,'7608' => 30216,
    '7609' => 30217,'760B' => 30219,
    '760D' => 30221,'760F' => 30223,
    '7611' => 30225,'7613' => 30227,
    '7614' => 30228,'7616' => 30230,
    '761A' => 30234,'761C' => 30236,
    '761D' => 30237,'761E' => 30238,
    '7621' => 30241,'7623' => 30243,
    '7627' => 30247,'7628' => 30248,
    '762F' => 30255,'7631' => 30257,
    '7632' => 30258,'763A' => 30266,
    '763D' => 30269,'7642' => 30274,
    '7646' => 30278,'7647' => 30279,
    '7648' => 30280,'7649' => 30281,
    '7650' => 30288,'7652' => 30290,
    '7653' => 30291,'7657' => 30295,
    '7658' => 30296,'7659' => 30297,
    '765A' => 30298,'765F' => 30303,
    '7660' => 30304,'7661' => 30305,
    '7662' => 30306,'7664' => 30308,
    '7665' => 30309,'7669' => 30313,
    '766A' => 30314,'766C' => 30316,
    '766D' => 30317,'766E' => 30318,
    '7670' => 30320,'7671' => 30321,
    '7672' => 30322,'7675' => 30325,
    '7679' => 30329,'767C' => 30332,
    '767F' => 30335,'7681' => 30337,
    '7689' => 30345,'768A' => 30346,
    '768F' => 30351,'7692' => 30354,
    '7695' => 30357,'769A' => 30362,
    '769B' => 30363,'769C' => 30364,
    '769D' => 30365,'769E' => 30366,
    '76A6' => 30374,'76AA' => 30378,
    '76AB' => 30379,'76AD' => 30381,
    '76AF' => 30383,'76B0' => 30384,
    '76B5' => 30389,'76B8' => 30392,
    '76BA' => 30394,'76BB' => 30395,
    '76BD' => 30397,'76BE' => 30398,
    '76C3' => 30403,'76C4' => 30404,
    '76C9' => 30409,'76D3' => 30419,
    '76DA' => 30426,'76DC' => 30428,
    '76DD' => 30429,'76DE' => 30430,
    '76E1' => 30433,'76E3' => 30435,
    '76E4' => 30436,'76E6' => 30438,
    '76E7' => 30439,'76E9' => 30441,
    '76EA' => 30442,'76EC' => 30444,
    '76ED' => 30445,'76F0' => 30448,
    '76F3' => 30451,'76F5' => 30453,
    '76F7' => 30455,'76FA' => 30458,
    '76FB' => 30459,'7703' => 30467,
    '7705' => 30469,'770A' => 30474,
    '7710' => 30480,'7711' => 30481,
    '7712' => 30482,'7713' => 30483,
    '7715' => 30485,'771B' => 30491,
    '771D' => 30493,'7723' => 30499,
    '7725' => 30501,'7727' => 30503,
    '7731' => 30513,'7732' => 30514,
    '7733' => 30515,'7734' => 30516,
    '7739' => 30521,'773B' => 30523,
    '773D' => 30525,'773E' => 30526,
    '7744' => 30532,'7745' => 30533,
    '7746' => 30534,'774A' => 30538,
    '774B' => 30539,'774C' => 30540,
    '774D' => 30541,'774E' => 30542,
    '774F' => 30543,'7752' => 30546,
    '7754' => 30548,'7755' => 30549,
    '7756' => 30550,'7759' => 30553,
    '775C' => 30556,'775E' => 30558,
    '775F' => 30559,'7760' => 30560,
    '7767' => 30567,'7769' => 30569,
    '776A' => 30570,'776D' => 30573,
    '776E' => 30574,'776F' => 30575,
    '777C' => 30588,'7781' => 30593,
    '7782' => 30594,'7783' => 30595,
    '7787' => 30599,'7788' => 30600,
    '7789' => 30601,'778B' => 30603,
    '778F' => 30607,'7795' => 30613,
    '7797' => 30615,'7799' => 30617,
    '779A' => 30618,'779B' => 30619,
    '779C' => 30620,'779D' => 30621,
    '779E' => 30622,'77A1' => 30625,
    '77A3' => 30627,'77A8' => 30632,
    '77AB' => 30635,'77AD' => 30637,
    '77B1' => 30641,'77B2' => 30642,
    '77B4' => 30644,'77B6' => 30646,
    '77B7' => 30647,'77BA' => 30650,
    '77BC' => 30652,'77C2' => 30658,
    '77C4' => 30660,'77C7' => 30663,
    '77C9' => 30665,'77CA' => 30666,
    '77CC' => 30668,'77CE' => 30670,
    '77CF' => 30671,'77D0' => 30672,
    '77D3' => 30675,'77D4' => 30676,
    '77D5' => 30677,'77D8' => 30680,
    '77D9' => 30681,'77DA' => 30682,
    '77DE' => 30686,'77E0' => 30688,
    '77E8' => 30696,'77EF' => 30703,
    '77F0' => 30704,'77F1' => 30705,
    '77F2' => 30706,'77F7' => 30711,
    '77F9' => 30713,'77FA' => 30714,
    '77FB' => 30715,'77FC' => 30716,
    '7803' => 30723,'7805' => 30725,
    '7806' => 30726,'780E' => 30734,
    '780F' => 30735,'7810' => 30736,
    '7813' => 30739,'7820' => 30752,
    '7821' => 30753,'7822' => 30754,
    '7828' => 30760,'782A' => 30762,
    '782B' => 30763,'782E' => 30766,
    '782F' => 30767,'7831' => 30769,
    '7832' => 30770,'7833' => 30771,
    '7835' => 30773,'7843' => 30787,
    '7848' => 30792,'7849' => 30793,
    '784A' => 30794,'784D' => 30797,
    '785C' => 30812,'785E' => 30814,
    '7860' => 30816,'7862' => 30818,
    '7864' => 30820,'7865' => 30821,
    '7868' => 30824,'7869' => 30825,
    '786F' => 30831,'7870' => 30832,
    '7871' => 30833,'7879' => 30841,
    '787B' => 30843,'787E' => 30846,
    '787F' => 30847,'7880' => 30848,
    '7881' => 30849,'7883' => 30851,
    '7884' => 30852,'7885' => 30853,
    '7886' => 30854,'788F' => 30863,
    '7894' => 30868,'7895' => 30869,
    '7896' => 30870,'7899' => 30873,
    '789E' => 30878,'78A0' => 30880,
    '78A2' => 30882,'78A4' => 30884,
    '78A8' => 30888,'78A9' => 30889,
    '78AA' => 30890,'78AB' => 30891,
    '78AC' => 30892,'78AD' => 30893,
    '78BA' => 30906,'78BB' => 30907,
    '78BC' => 30908,'78C3' => 30915,
    '78C4' => 30916,'78C8' => 30920,
    '78CC' => 30924,'78CD' => 30925,
    '78CE' => 30926,'78CF' => 30927,
    '78D1' => 30929,'78DA' => 30938,
    '78DB' => 30939,'78DD' => 30941,
    '78DE' => 30942,'78DF' => 30943,
    '78E0' => 30944,'78E1' => 30945,
    '78E2' => 30946,'78E3' => 30947,
    '78E5' => 30949,'78E7' => 30951,
    '78E9' => 30953,'78EA' => 30954,
    '78ED' => 30957,'78EF' => 30959,
    '78F3' => 30963,'78F9' => 30969,
    '78FB' => 30971,'78FC' => 30972,
    '78FD' => 30973,'78FE' => 30974,
    '78FF' => 30975,'7902' => 30978,
    '7904' => 30980,'7909' => 30985,
    '790C' => 30988,'790E' => 30990,
    '7910' => 30992,'7911' => 30993,
    '7912' => 30994,'7914' => 30996,
    '7917' => 30999,'7919' => 31001,
    '791B' => 31003,'791C' => 31004,
    '791D' => 31005,'7921' => 31009,
    '7923' => 31011,'7925' => 31013,
    '7926' => 31014,'7927' => 31015,
    '7928' => 31016,'7929' => 31017,
    '792A' => 31018,'792B' => 31019,
    '792C' => 31020,'792D' => 31021,
    '792F' => 31023,'7931' => 31025,
    '7935' => 31029,'7938' => 31032,
    '7939' => 31033,'793D' => 31037,
    '793F' => 31039,'7942' => 31042,
    '7944' => 31044,'7945' => 31045,
    '7947' => 31047,'794A' => 31050,
    '794B' => 31051,'794C' => 31052,
    '794F' => 31055,'7950' => 31056,
    '7951' => 31057,'7952' => 31058,
    '7954' => 31060,'7955' => 31061,
    '7961' => 31073,'7963' => 31075,
    '7964' => 31076,'7969' => 31081,
    '796A' => 31082,'796B' => 31083,
    '7970' => 31088,'7972' => 31090,
    '7973' => 31091,'7974' => 31092,
    '7979' => 31097,'797C' => 31100,
    '797D' => 31101,'797F' => 31103,
    '7982' => 31106,'7988' => 31112,
    '798B' => 31115,'798D' => 31117,
    '798E' => 31118,'7990' => 31120,
    '7992' => 31122,'7993' => 31123,
    '7994' => 31124,'7995' => 31125,
    '7996' => 31126,'7997' => 31127,
    '7998' => 31128,'799B' => 31131,
    '799C' => 31132,'79A0' => 31136,
    '79A1' => 31137,'79A2' => 31138,
    '79A4' => 31140,'79A6' => 31142,
    '79A8' => 31144,'79AA' => 31146,
    '79AB' => 31147,'79AC' => 31148,
    '79AD' => 31149,'79AE' => 31150,
    '79B0' => 31152,'79B1' => 31153,
    '79B2' => 31154,'79B4' => 31156,
    '79B6' => 31158,'79B7' => 31159,
    '79B8' => 31160,'79BF' => 31167,
    '79C5' => 31173,'79C8' => 31176,
    '79CE' => 31182,'79CF' => 31183,
    '79D6' => 31190,'79DC' => 31196,
    '79DD' => 31197,'79DE' => 31198,
    '79E0' => 31200,'79EA' => 31210,
    '79EC' => 31212,'79EE' => 31214,
    '79F6' => 31222,'79F7' => 31223,
    '79FA' => 31226,'7A04' => 31236,
    '7A05' => 31237,'7A08' => 31240,
    '7A0A' => 31242,'7A0C' => 31244,
    '7A10' => 31248,'7A11' => 31249,
    '7A12' => 31250,'7A13' => 31251,
    '7A15' => 31253,'7A18' => 31256,
    '7A19' => 31257,'7A1B' => 31259,
    '7A1C' => 31260,'7A1F' => 31263,
    '7A22' => 31266,'7A26' => 31270,
    '7A28' => 31272,'7A2B' => 31275,
    '7A2E' => 31278,'7A2F' => 31279,
    '7A30' => 31280,'7A31' => 31281,
    '7A40' => 31296,'7A44' => 31300,
    '7A47' => 31303,'7A48' => 31304,
    '7A4A' => 31306,'7A4B' => 31307,
    '7A4C' => 31308,'7A4D' => 31309,
    '7A4E' => 31310,'7A54' => 31316,
    '7A56' => 31318,'7A58' => 31320,
    '7A5A' => 31322,'7A5B' => 31323,
    '7A5C' => 31324,'7A5F' => 31327,
    '7A60' => 31328,'7A61' => 31329,
    '7A62' => 31330,'7A67' => 31335,
    '7A68' => 31336,'7A69' => 31337,
    '7A6B' => 31339,'7A6C' => 31340,
    '7A6D' => 31341,'7A6E' => 31342,
    '7A71' => 31345,'7A75' => 31349,
    '7A7B' => 31355,'7A7E' => 31358,
    '7A85' => 31365,'7A87' => 31367,
    '7A89' => 31369,'7A8A' => 31370,
    '7A8B' => 31371,'7A8C' => 31372,
    '7A8F' => 31375,'7A90' => 31376,
    '7A94' => 31380,'7A99' => 31385,
    '7A9E' => 31390,'7AA2' => 31394,
    '7AA3' => 31395,'7AA9' => 31401,
    '7AAA' => 31402,'7AAB' => 31403,
    '7AAE' => 31406,'7AAF' => 31407,
    '7AB1' => 31409,'7AB2' => 31410,
    '7AB4' => 31412,'7AB5' => 31413,
    '7AB6' => 31414,'7AB7' => 31415,
    '7AB8' => 31416,'7ABA' => 31418,
    '7ABE' => 31422,'7AC0' => 31424,
    '7AC1' => 31425,'7AC4' => 31428,
    '7AC5' => 31429,'7AC7' => 31431,
    '7ACA' => 31434,'7AD1' => 31441,
    '7AD8' => 31448,'7AE4' => 31460,
    '7AEB' => 31467,'7AEE' => 31470,
    '7AF6' => 31478,'7AF7' => 31479,
    '7AFB' => 31483,'7B00' => 31488,
    '7B01' => 31489,'7B05' => 31493,
    '7B09' => 31497,'7B0E' => 31502,
    '7B10' => 31504,'7B12' => 31506,
    '7B13' => 31507,'7B18' => 31512,
    '7B1A' => 31514,'7B1D' => 31517,
    '7B22' => 31522,'7B23' => 31523,
    '7B2D' => 31533,'7B2F' => 31535,
    '7B30' => 31536,'7B32' => 31538,
    '7B34' => 31540,'7B35' => 31541,
    '7B3B' => 31547,'7B40' => 31552,
    '7B44' => 31556,'7B46' => 31558,
    '7B48' => 31560,'7B4A' => 31562,
    '7B4D' => 31565,'7B4E' => 31566,
    '7B61' => 31585,'7B63' => 31587,
    '7B64' => 31588,'7B65' => 31589,
    '7B66' => 31590,'7B67' => 31591,
    '7B69' => 31593,'7B6D' => 31597,
    '7B70' => 31600,'7B73' => 31603,
    '7B74' => 31604,'7B76' => 31606,
    '7B78' => 31608,'7B82' => 31618,
    '7B84' => 31620,'7B87' => 31623,
    '7B88' => 31624,'7B8A' => 31626,
    '7B8B' => 31627,'7B8C' => 31628,
    '7B8E' => 31630,'7B8F' => 31631,
    '7B91' => 31633,'7B96' => 31638,
    '7B98' => 31640,'7B99' => 31641,
    '7B9B' => 31643,'7BA0' => 31648,
    '7BA4' => 31652,'7BAF' => 31663,
    '7BB5' => 31669,'7BB7' => 31671,
    '7BB9' => 31673,'7BBE' => 31678,
    '7BC0' => 31680,'7BC4' => 31684,
    '7BC9' => 31689,'7BCA' => 31690,
    '7BCB' => 31691,'7BCE' => 31694,
    '7BD4' => 31700,'7BD5' => 31701,
    '7BD8' => 31704,'7BDB' => 31707,
    '7BDC' => 31708,'7BDE' => 31710,
    '7BDF' => 31711,'7BE0' => 31712,
    '7BE2' => 31714,'7BE3' => 31715,
    '7BE4' => 31716,'7BE7' => 31719,
    '7BE8' => 31720,'7BE9' => 31721,
    '7BEB' => 31723,'7BF0' => 31728,
    '7BF2' => 31730,'7BF3' => 31731,
    '7BF4' => 31732,'7BF8' => 31736,
    '7BF9' => 31737,'7BFB' => 31739,
    '7BFD' => 31741,'7BFF' => 31743,
    '7C00' => 31744,'7C01' => 31745,
    '7C02' => 31746,'7C03' => 31747,
    '7C05' => 31749,'7C06' => 31750,
    '7C09' => 31753,'7C0A' => 31754,
    '7C0D' => 31757,'7C0E' => 31758,
    '7C10' => 31760,'7C11' => 31761,
    '7C19' => 31769,'7C1C' => 31772,
    '7C1D' => 31773,'7C1E' => 31774,
    '7C20' => 31776,'7C21' => 31777,
    '7C22' => 31778,'7C23' => 31779,
    '7C25' => 31781,'7C28' => 31784,
    '7C29' => 31785,'7C2B' => 31787,
    '7C2C' => 31788,'7C2D' => 31789,
    '7C30' => 31792,'7C33' => 31795,
    '7C37' => 31799,'7C39' => 31801,
    '7C3B' => 31803,'7C3C' => 31804,
    '7C3D' => 31805,'7C3E' => 31806,
    '7C43' => 31811,'7C45' => 31813,
    '7C47' => 31815,'7C48' => 31816,
    '7C49' => 31817,'7C4A' => 31818,
    '7C4C' => 31820,'7C50' => 31824,
    '7C53' => 31827,'7C54' => 31828,
    '7C57' => 31831,'7C59' => 31833,
    '7C5A' => 31834,'7C5B' => 31835,
    '7C5C' => 31836,'7C5F' => 31839,
    '7C60' => 31840,'7C63' => 31843,
    '7C64' => 31844,'7C65' => 31845,
    '7C66' => 31846,'7C67' => 31847,
    '7C69' => 31849,'7C6A' => 31850,
    '7C6B' => 31851,'7C6C' => 31852,
    '7C6E' => 31854,'7C6F' => 31855,
    '7C72' => 31858,'7C75' => 31861,
    '7C78' => 31864,'7C79' => 31865,
    '7C7A' => 31866,'7C7F' => 31871,
    '7C80' => 31872,'7C81' => 31873,
    '7C84' => 31876,'7C85' => 31877,
    '7C88' => 31880,'7C8A' => 31882,
    '7C8C' => 31884,'7C8D' => 31885,
    '7C94' => 31892,'7C96' => 31894,
    '7CA1' => 31905,'7CA3' => 31907,
    '7CA7' => 31911,'7CA8' => 31912,
    '7CAF' => 31919,'7CB4' => 31924,
    '7CB5' => 31925,'7CBA' => 31930,
    '7CBB' => 31931,'7CBF' => 31935,
    '7CCB' => 31947,'7CCE' => 31950,
    '7CD0' => 31952,'7CD1' => 31953,
    '7CD2' => 31954,'7CD4' => 31956,
    '7CDD' => 31965,'7CDE' => 31966,
    '7CE2' => 31970,'7CE7' => 31975,
    '7CEA' => 31978,'7CEC' => 31980,
    '7CEE' => 31982,'7CF0' => 31984,
    '7CF1' => 31985,'7CF2' => 31986,
    '7CF4' => 31988,'7CF6' => 31990,
    '7CF7' => 31991,'7CFD' => 31997,
    '7CFE' => 31998,'7D00' => 32000,
    '7D01' => 32001,'7D02' => 32002,
    '7D03' => 32003,'7D04' => 32004,
    '7D05' => 32005,'7D06' => 32006,
    '7D07' => 32007,'7D08' => 32008,
    '7D09' => 32009,'7D0B' => 32011,
    '7D0C' => 32012,'7D0D' => 32013,
    '7D0E' => 32014,'7D0F' => 32015,
    '7D10' => 32016,'7D11' => 32017,
    '7D12' => 32018,'7D13' => 32019,
    '7D14' => 32020,'7D15' => 32021,
    '7D16' => 32022,'7D17' => 32023,
    '7D18' => 32024,'7D19' => 32025,
    '7D1A' => 32026,'7D1B' => 32027,
    '7D1C' => 32028,'7D1D' => 32029,
    '7D1E' => 32030,'7D1F' => 32031,
    '7D21' => 32033,'7D28' => 32040,
    '7D29' => 32041,'7D2C' => 32044,
    '7D2E' => 32046,'7D30' => 32048,
    '7D31' => 32049,'7D32' => 32050,
    '7D33' => 32051,'7D35' => 32053,
    '7D36' => 32054,'7D38' => 32056,
    '7D39' => 32057,'7D3A' => 32058,
    '7D3B' => 32059,'7D3C' => 32060,
    '7D3D' => 32061,'7D3E' => 32062,
    '7D3F' => 32063,'7D40' => 32064,
    '7D41' => 32065,'7D42' => 32066,
    '7D43' => 32067,'7D44' => 32068,
    '7D45' => 32069,'7D46' => 32070,
    '7D47' => 32071,'7D4A' => 32074,
    '7D4E' => 32078,'7D4F' => 32079,
    '7D50' => 32080,'7D51' => 32081,
    '7D52' => 32082,'7D53' => 32083,
    '7D54' => 32084,'7D55' => 32085,
    '7D56' => 32086,'7D58' => 32088,
    '7D5B' => 32091,'7D5C' => 32092,
    '7D5E' => 32094,'7D5F' => 32095,
    '7D61' => 32097,'7D62' => 32098,
    '7D63' => 32099,'7D66' => 32102,
    '7D67' => 32103,'7D68' => 32104,
    '7D69' => 32105,'7D6A' => 32106,
    '7D6B' => 32107,'7D6D' => 32109,
    '7D6F' => 32111,'7D70' => 32112,
    '7D71' => 32113,'7D72' => 32114,
    '7D73' => 32115,'7D79' => 32121,
    '7D7A' => 32122,'7D7B' => 32123,
    '7D7C' => 32124,'7D7D' => 32125,
    '7D7F' => 32127,'7D80' => 32128,
    '7D81' => 32129,'7D83' => 32131,
    '7D84' => 32132,'7D85' => 32133,
    '7D86' => 32134,'7D88' => 32136,
    '7D8C' => 32140,'7D8D' => 32141,
    '7D8E' => 32142,'7D8F' => 32143,
    '7D91' => 32145,'7D92' => 32146,
    '7D93' => 32147,'7D94' => 32148,
    '7D96' => 32150,'7D9C' => 32156,
    '7D9D' => 32157,'7D9E' => 32158,
    '7D9F' => 32159,'7DA0' => 32160,
    '7DA1' => 32161,'7DA2' => 32162,
    '7DA3' => 32163,'7DA7' => 32167,
    '7DA9' => 32169,'7DAA' => 32170,
    '7DAC' => 32172,'7DAD' => 32173,
    '7DAF' => 32175,'7DB0' => 32176,
    '7DB1' => 32177,'7DB2' => 32178,
    '7DB4' => 32180,'7DB5' => 32181,
    '7DB7' => 32183,'7DB8' => 32184,
    '7DB9' => 32185,'7DBA' => 32186,
    '7DBB' => 32187,'7DBC' => 32188,
    '7DBD' => 32189,'7DBE' => 32190,
    '7DBF' => 32191,'7DC0' => 32192,
    '7DC1' => 32193,'7DC2' => 32194,
    '7DC4' => 32196,'7DC5' => 32197,
    '7DC6' => 32198,'7DC7' => 32199,
    '7DC9' => 32201,'7DCA' => 32202,
    '7DCB' => 32203,'7DCC' => 32204,
    '7DCE' => 32206,'7DD2' => 32210,
    '7DD7' => 32215,'7DD8' => 32216,
    '7DD9' => 32217,'7DDA' => 32218,
    '7DDB' => 32219,'7DDD' => 32221,
    '7DDE' => 32222,'7DDF' => 32223,
    '7DE0' => 32224,'7DE1' => 32225,
    '7DE3' => 32227,'7DE6' => 32230,
    '7DE7' => 32231,'7DE8' => 32232,
    '7DE9' => 32233,'7DEA' => 32234,
    '7DEC' => 32236,'7DEE' => 32238,
    '7DEF' => 32239,'7DF0' => 32240,
    '7DF1' => 32241,'7DF2' => 32242,
    '7DF3' => 32243,'7DF4' => 32244,
    '7DF6' => 32246,'7DF7' => 32247,
    '7DF9' => 32249,'7DFA' => 32250,
    '7DFB' => 32251,'7E03' => 32259,
    '7E08' => 32264,'7E09' => 32265,
    '7E0A' => 32266,'7E0B' => 32267,
    '7E0C' => 32268,'7E0D' => 32269,
    '7E0E' => 32270,'7E0F' => 32271,
    '7E10' => 32272,'7E11' => 32273,
    '7E12' => 32274,'7E13' => 32275,
    '7E14' => 32276,'7E15' => 32277,
    '7E16' => 32278,'7E17' => 32279,
    '7E1A' => 32282,'7E1B' => 32283,
    '7E1C' => 32284,'7E1D' => 32285,
    '7E1E' => 32286,'7E1F' => 32287,
    '7E20' => 32288,'7E21' => 32289,
    '7E22' => 32290,'7E23' => 32291,
    '7E24' => 32292,'7E25' => 32293,
    '7E29' => 32297,'7E2A' => 32298,
    '7E2B' => 32299,'7E2D' => 32301,
    '7E2E' => 32302,'7E2F' => 32303,
    '7E30' => 32304,'7E31' => 32305,
    '7E32' => 32306,'7E33' => 32307,
    '7E34' => 32308,'7E35' => 32309,
    '7E36' => 32310,'7E37' => 32311,
    '7E38' => 32312,'7E39' => 32313,
    '7E3A' => 32314,'7E3C' => 32316,
    '7E3D' => 32317,'7E3E' => 32318,
    '7E3F' => 32319,'7E40' => 32320,
    '7E42' => 32322,'7E43' => 32323,
    '7E44' => 32324,'7E45' => 32325,
    '7E46' => 32326,'7E48' => 32328,
    '7E49' => 32329,'7E4C' => 32332,
    '7E50' => 32336,'7E51' => 32337,
    '7E52' => 32338,'7E53' => 32339,
    '7E54' => 32340,'7E55' => 32341,
    '7E56' => 32342,'7E57' => 32343,
    '7E58' => 32344,'7E59' => 32345,
    '7E5A' => 32346,'7E5C' => 32348,
    '7E5E' => 32350,'7E5F' => 32351,
    '7E60' => 32352,'7E61' => 32353,
    '7E62' => 32354,'7E63' => 32355,
    '7E68' => 32360,'7E69' => 32361,
    '7E6A' => 32362,'7E6B' => 32363,
    '7E6D' => 32365,'7E6F' => 32367,
    '7E70' => 32368,'7E72' => 32370,
    '7E73' => 32371,'7E74' => 32372,
    '7E75' => 32373,'7E76' => 32374,
    '7E77' => 32375,'7E78' => 32376,
    '7E79' => 32377,'7E7A' => 32378,
    '7E7B' => 32379,'7E7C' => 32380,
    '7E7D' => 32381,'7E7E' => 32382,
    '7E80' => 32384,'7E81' => 32385,
    '7E86' => 32390,'7E87' => 32391,
    '7E88' => 32392,'7E8A' => 32394,
    '7E8B' => 32395,'7E8C' => 32396,
    '7E8D' => 32397,'7E8F' => 32399,
    '7E91' => 32401,'7E93' => 32403,
    '7E94' => 32404,'7E95' => 32405,
    '7E96' => 32406,'7E97' => 32407,
    '7E98' => 32408,'7E99' => 32409,
    '7E9A' => 32410,'7E9C' => 32412,
    '7F39' => 32569,'7F3D' => 32573,
    '7F3E' => 32574,'7F3F' => 32575,
    '7F43' => 32579,'7F48' => 32584,
    '7F4A' => 32586,'7F4B' => 32587,
    '7F4C' => 32588,'7F4D' => 32589,
    '7F4F' => 32591,'7F5B' => 32603,
    '7F5C' => 32604,'7F5D' => 32605,
    '7F5E' => 32606,'7F60' => 32608,
    '7F63' => 32611,'7F65' => 32613,
    '7F66' => 32614,'7F67' => 32615,
    '7F6B' => 32619,'7F6C' => 32620,
    '7F6D' => 32621,'7F70' => 32624,
    '7F73' => 32627,'7F75' => 32629,
    '7F76' => 32630,'7F77' => 32631,
    '7F7A' => 32634,'7F7B' => 32635,
    '7F7C' => 32636,'7F7D' => 32637,
    '7F7F' => 32639,'7F83' => 32643,
    '7F85' => 32645,'7F86' => 32646,
    '7F87' => 32647,'7F88' => 32648,
    '7F89' => 32649,'7F8B' => 32651,
    '7F8D' => 32653,'7F91' => 32657,
    '7F92' => 32658,'7F95' => 32661,
    '7F96' => 32662,'7F9B' => 32667,
    '7F9C' => 32668,'7FA0' => 32672,
    '7FA2' => 32674,'7FA5' => 32677,
    '7FA6' => 32678,'7FA8' => 32680,
    '7FA9' => 32681,'7FAC' => 32684,
    '7FAD' => 32685,'7FB1' => 32689,
    '7FB3' => 32691,'7FB5' => 32693,
    '7FB6' => 32694,'7FB7' => 32695,
    '7FBA' => 32698,'7FBB' => 32699,
    '7FBE' => 32702,'7FC0' => 32704,
    '7FC2' => 32706,'7FC3' => 32707,
    '7FC7' => 32711,'7FC9' => 32713,
    '7FCB' => 32715,'7FCD' => 32717,
    '7FCF' => 32719,'7FD0' => 32720,
    '7FD1' => 32721,'7FD2' => 32722,
    '7FD7' => 32727,'7FDB' => 32731,
    '7FDC' => 32732,'7FDE' => 32734,
    '7FE2' => 32738,'7FE3' => 32739,
    '7FE8' => 32744,'7FEA' => 32746,
    '7FEB' => 32747,'7FEC' => 32748,
    '7FED' => 32749,'7FEF' => 32751,
    '7FF2' => 32754,'7FF4' => 32756,
    '7FF5' => 32757,'7FF7' => 32759,
    '7FF8' => 32760,'7FF9' => 32761,
    '7FFD' => 32765,'7FFE' => 32766,
    '7FFF' => 32767,'8007' => 32775,
    '800E' => 32782,'800F' => 32783,
    '8011' => 32785,'801B' => 32795,
    '801E' => 32798,'801F' => 32799,
    '8021' => 32801,'8024' => 32804,
    '802C' => 32812,'8030' => 32816,
    '8034' => 32820,'8039' => 32825,
    '803E' => 32830,'8047' => 32839,
    '8048' => 32840,'804F' => 32847,
    '8050' => 32848,'8051' => 32849,
    '8056' => 32854,'805C' => 32860,
    '805D' => 32861,'805E' => 32862,
    '8064' => 32868,'8067' => 32871,
    '806C' => 32876,'806F' => 32879,
    '8070' => 32880,'8072' => 32882,
    '8073' => 32883,'8075' => 32885,
    '8076' => 32886,'8077' => 32887,
    '8078' => 32888,'8079' => 32889,
    '807D' => 32893,'807E' => 32894,
    '8082' => 32898,'8085' => 32901,
    '808A' => 32906,'808F' => 32911,
    '8090' => 32912,'8092' => 32914,
    '8095' => 32917,'8099' => 32921,
    '80A3' => 32931,'80B5' => 32949,
    '80B8' => 32952,'80C5' => 32965,
    '80C7' => 32967,'80C8' => 32968,
    '80C9' => 32969,'80CA' => 32970,
    '80CF' => 32975,'80D0' => 32976,
    '80D1' => 32977,'80D4' => 32980,
    '80D5' => 32981,'80D8' => 32984,
    '80E0' => 32992,'80E3' => 32995,
    '80E6' => 32998,'80F5' => 33013,
    '80F9' => 33017,'80FB' => 33019,
    '80FE' => 33022,'8100' => 33024,
    '8101' => 33025,'8105' => 33029,
    '8108' => 33032,'8115' => 33045,
    '8119' => 33049,'811B' => 33051,
    '811D' => 33053,'811F' => 33055,
    '8121' => 33057,'8122' => 33058,
    '8123' => 33059,'8124' => 33060,
    '8125' => 33061,'8127' => 33063,
    '8129' => 33065,'812B' => 33067,
    '812D' => 33069,'8130' => 33072,
    '8139' => 33081,'813A' => 33082,
    '813D' => 33085,'8143' => 33091,
    '8144' => 33092,'8147' => 33095,
    '814D' => 33101,'814E' => 33102,
    '814F' => 33103,'8152' => 33106,
    '815B' => 33115,'815C' => 33116,
    '815E' => 33118,'8161' => 33121,
    '8162' => 33122,'8164' => 33124,
    '8166' => 33126,'816B' => 33131,
    '816F' => 33135,'8172' => 33138,
    '8173' => 33139,'8176' => 33142,
    '8177' => 33143,'8178' => 33144,
    '8183' => 33155,'8186' => 33158,
    '8187' => 33159,'8189' => 33161,
    '818B' => 33163,'818C' => 33164,
    '818D' => 33165,'8195' => 33173,
    '8197' => 33175,'8199' => 33177,
    '819A' => 33178,'819E' => 33182,
    '819F' => 33183,'81A0' => 33184,
    '81A2' => 33186,'81A7' => 33191,
    '81A9' => 33193,'81AB' => 33195,
    '81AC' => 33196,'81AE' => 33198,
    '81B0' => 33200,'81B1' => 33201,
    '81B2' => 33202,'81B4' => 33204,
    '81B5' => 33205,'81B7' => 33207,
    '81B9' => 33209,'81BC' => 33212,
    '81BD' => 33213,'81BE' => 33214,
    '81BF' => 33215,'81C4' => 33220,
    '81C5' => 33221,'81C7' => 33223,
    '81C9' => 33225,'81CD' => 33229,
    '81CF' => 33231,'81D0' => 33232,
    '81D1' => 33233,'81D2' => 33234,
    '81D5' => 33237,'81D7' => 33239,
    '81D8' => 33240,'81D9' => 33241,
    '81DA' => 33242,'81DB' => 33243,
    '81DD' => 33245,'81DE' => 33246,
    '81DF' => 33247,'81E0' => 33248,
    '81E1' => 33249,'81E2' => 33250,
    '81E5' => 33253,'81E6' => 33254,
    '81E8' => 33256,'81E9' => 33257,
    '81EE' => 33262,'81F2' => 33266,
    '81F7' => 33271,'81F8' => 33272,
    '81F9' => 33273,'81FA' => 33274,
    '81FF' => 33279,'8207' => 33287,
    '8208' => 33288,'8209' => 33289,
    '820A' => 33290,'820B' => 33291,
    '8211' => 33297,'8215' => 33301,
    '8216' => 33302,'821D' => 33309,
    '8220' => 33312,'8225' => 33317,
    '8232' => 33330,'823A' => 33338,
    '823C' => 33340,'823D' => 33341,
    '823F' => 33343,'8240' => 33344,
    '8242' => 33346,'8245' => 33349,
    '824E' => 33358,'8250' => 33360,
    '8251' => 33361,'8252' => 33362,
    '8253' => 33363,'8255' => 33365,
    '8256' => 33366,'8257' => 33367,
    '8259' => 33369,'825B' => 33371,
    '825C' => 33372,'825E' => 33374,
    '8261' => 33377,'8263' => 33379,
    '8264' => 33380,'8266' => 33382,
    '8269' => 33385,'826B' => 33387,
    '826C' => 33388,'826D' => 33389,
    '8271' => 33393,'8275' => 33397,
    '8277' => 33399,'8278' => 33400,
    '827C' => 33404,'8280' => 33408,
    '8283' => 33411,'8285' => 33413,
    '8290' => 33424,'8293' => 33427,
    '8294' => 33428,'829A' => 33434,
    '829B' => 33435,'829E' => 33438,
    '82A0' => 33440,'82A2' => 33442,
    '82A3' => 33443,'82A7' => 33447,
    '82B5' => 33461,'82B6' => 33462,
    '82BA' => 33466,'82BB' => 33467,
    '82BC' => 33468,'82C0' => 33472,
    '82C2' => 33474,'82C3' => 33475,
    '82D6' => 33494,'82D9' => 33497,
    '82E7' => 33511,'82E8' => 33512,
    '82EA' => 33514,'82EC' => 33516,
    '82ED' => 33517,'82F0' => 33520,
    '82F2' => 33522,'82F3' => 33523,
    '82F5' => 33525,'82F6' => 33526,
    '82FA' => 33530,'82FE' => 33534,
    '8300' => 33536,'830D' => 33549,
    '8316' => 33558,'8319' => 33561,
    '831E' => 33566,'8320' => 33568,
    '8322' => 33570,'8324' => 33572,
    '8325' => 33573,'8326' => 33574,
    '8329' => 33577,'832A' => 33578,
    '8332' => 33586,'8337' => 33591,
    '833B' => 33595,'833F' => 33599,
    '8341' => 33601,'8342' => 33602,
    '8344' => 33604,'8345' => 33605,
    '8348' => 33608,'834A' => 33610,
    '834B' => 33611,'834C' => 33612,
    '834D' => 33613,'834E' => 33614,
    '8353' => 33619,'8356' => 33622,
    '8373' => 33651,'8374' => 33652,
    '8375' => 33653,'8376' => 33654,
    '837A' => 33658,'837E' => 33662,
    '837F' => 33663,'8381' => 33665,
    '8383' => 33667,'8387' => 33671,
    '8388' => 33672,'838A' => 33674,
    '838B' => 33675,'838C' => 33676,
    '838D' => 33677,'838F' => 33679,
    '8390' => 33680,'8394' => 33684,
    '8395' => 33685,'8396' => 33686,
    '8397' => 33687,'8399' => 33689,
    '839A' => 33690,'839D' => 33693,
    '83A2' => 33698,'83A3' => 33699,
    '83A4' => 33700,'83A5' => 33701,
    '83A6' => 33702,'83A7' => 33703,
    '83AE' => 33710,'83AF' => 33711,
    '83BF' => 33727,'83C2' => 33730,
    '83C3' => 33731,'83C4' => 33732,
    '83C6' => 33734,'83C8' => 33736,
    '83C9' => 33737,'83CB' => 33739,
    '83CE' => 33742,'83D1' => 33745,
    '83D5' => 33749,'83D7' => 33751,
    '83D9' => 33753,'83DB' => 33755,
    '83DE' => 33758,'83E2' => 33762,
    '83E3' => 33763,'83E4' => 33764,
    '83E7' => 33767,'83E8' => 33768,
    '83EB' => 33771,'83EC' => 33772,
    '83EE' => 33774,'83EF' => 33775,
    '83F3' => 33779,'83F4' => 33780,
    '83F5' => 33781,'83F6' => 33782,
    '83FA' => 33786,'83FB' => 33787,
    '83FC' => 33788,'83FE' => 33790,
    '83FF' => 33791,'8407' => 33799,
    '8409' => 33801,'840A' => 33802,
    '8410' => 33808,'8412' => 33810,
    '8413' => 33811,'841B' => 33819,
    '8423' => 33827,'8429' => 33833,
    '842B' => 33835,'842C' => 33836,
    '842D' => 33837,'842F' => 33839,
    '8430' => 33840,'8432' => 33842,
    '8433' => 33843,'8434' => 33844,
    '8435' => 33845,'8436' => 33846,
    '8437' => 33847,'8439' => 33849,
    '843A' => 33850,'843B' => 33851,
    '843F' => 33855,'8440' => 33856,
    '8442' => 33858,'8443' => 33859,
    '8444' => 33860,'8445' => 33861,
    '8447' => 33863,'8449' => 33865,
    '844B' => 33867,'844C' => 33868,
    '844D' => 33869,'844E' => 33870,
    '8450' => 33872,'8452' => 33874,
    '8454' => 33876,'8456' => 33878,
    '845D' => 33885,'845E' => 33886,
    '845F' => 33887,'8460' => 33888,
    '8465' => 33893,'8466' => 33894,
    '8467' => 33895,'8468' => 33896,
    '846E' => 33902,'846F' => 33903,
    '8470' => 33904,'8474' => 33908,
    '8477' => 33911,'8479' => 33913,
    '847D' => 33917,'847E' => 33918,
    '8486' => 33926,'848D' => 33933,
    '848F' => 33935,'8490' => 33936,
    '8491' => 33937,'8494' => 33940,
    '8498' => 33944,'849A' => 33946,
    '849B' => 33947,'849D' => 33949,
    '849E' => 33950,'849F' => 33951,
    '84A0' => 33952,'84A2' => 33954,
    '84A4' => 33956,'84A7' => 33959,
    '84A8' => 33960,'84A9' => 33961,
    '84AA' => 33962,'84AB' => 33963,
    '84AC' => 33964,'84AE' => 33966,
    '84B0' => 33968,'84B1' => 33969,
    '84B6' => 33974,'84BB' => 33979,
    '84BC' => 33980,'84C0' => 33984,
    '84C2' => 33986,'84C5' => 33989,
    '84C6' => 33990,'84C7' => 33991,
    '84CB' => 33995,'84CC' => 33996,
    '84CE' => 33998,'84CF' => 33999,
    '84D2' => 34002,'84D4' => 34004,
    '84D7' => 34007,'84DB' => 34011,
    '84E7' => 34023,'84E8' => 34024,
    '84E9' => 34025,'84EA' => 34026,
    '84EB' => 34027,'84EE' => 34030,
    '84EF' => 34031,'84F1' => 34033,
    '84F2' => 34034,'84F3' => 34035,
    '84F4' => 34036,'84F6' => 34038,
    '84F7' => 34039,'84F9' => 34041,
    '84FA' => 34042,'84FB' => 34043,
    '84FD' => 34045,'84FE' => 34046,
    '8500' => 34048,'8502' => 34050,
    '8506' => 34054,'8507' => 34055,
    '8508' => 34056,'8509' => 34057,
    '850A' => 34058,'850B' => 34059,
    '850D' => 34061,'850E' => 34062,
    '850F' => 34063,'8512' => 34066,
    '8514' => 34068,'8515' => 34069,
    '8516' => 34070,'8518' => 34072,
    '8519' => 34073,'851C' => 34076,
    '851D' => 34077,'851E' => 34078,
    '8520' => 34080,'8523' => 34083,
    '8524' => 34084,'8525' => 34085,
    '8526' => 34086,'8527' => 34087,
    '8528' => 34088,'8529' => 34089,
    '852A' => 34090,'852D' => 34093,
    '852E' => 34094,'852F' => 34095,
    '8530' => 34096,'8531' => 34097,
    '853E' => 34110,'8540' => 34112,
    '8541' => 34113,'8544' => 34116,
    '8545' => 34117,'8546' => 34118,
    '8547' => 34119,'854D' => 34125,
    '854E' => 34126,'8551' => 34129,
    '8553' => 34131,'8554' => 34132,
    '8555' => 34133,'8557' => 34135,
    '8558' => 34136,'855B' => 34139,
    '855D' => 34141,'8560' => 34144,
    '8561' => 34145,'8562' => 34146,
    '8563' => 34147,'8565' => 34149,
    '8566' => 34150,'8567' => 34151,
    '8569' => 34153,'856A' => 34154,
    '856B' => 34155,'856C' => 34156,
    '856D' => 34157,'856E' => 34158,
    '8571' => 34161,'8575' => 34165,
    '8576' => 34166,'8577' => 34167,
    '8578' => 34168,'857C' => 34172,
    '8580' => 34176,'8581' => 34177,
    '8582' => 34178,'8583' => 34179,
    '8586' => 34182,'8588' => 34184,
    '8589' => 34185,'858A' => 34186,
    '858B' => 34187,'858C' => 34188,
    '858D' => 34189,'858E' => 34190,
    '8590' => 34192,'8591' => 34193,
    '8594' => 34196,'8595' => 34197,
    '8596' => 34198,'8598' => 34200,
    '8599' => 34201,'859A' => 34202,
    '859D' => 34205,'859E' => 34206,
    '859F' => 34207,'85A0' => 34208,
    '85A1' => 34209,'85A2' => 34210,
    '85A3' => 34211,'85A6' => 34214,
    '85A7' => 34215,'85A9' => 34217,
    '85B1' => 34225,'85B3' => 34227,
    '85B4' => 34228,'85B5' => 34229,
    '85B6' => 34230,'85B8' => 34232,
    '85BA' => 34234,'85BD' => 34237,
    '85BE' => 34238,'85BF' => 34239,
    '85C0' => 34240,'85C2' => 34242,
    '85C3' => 34243,'85C4' => 34244,
    '85C5' => 34245,'85C6' => 34246,
    '85C7' => 34247,'85C8' => 34248,
    '85CB' => 34251,'85CD' => 34253,
    '85CE' => 34254,'85D1' => 34257,
    '85D2' => 34258,'85D7' => 34263,
    '85D8' => 34264,'85D9' => 34265,
    '85DA' => 34266,'85DD' => 34269,
    '85DE' => 34270,'85DF' => 34271,
    '85E1' => 34273,'85E2' => 34274,
    '85E3' => 34275,'85E5' => 34277,
    '85E6' => 34278,'85E8' => 34280,
    '85EA' => 34282,'85EB' => 34283,
    '85EC' => 34284,'85ED' => 34285,
    '85EF' => 34287,'85F0' => 34288,
    '85F1' => 34289,'85F2' => 34290,
    '85F6' => 34294,'85F7' => 34295,
    '85F8' => 34296,'85F9' => 34297,
    '85FA' => 34298,'85FD' => 34301,
    '85FE' => 34302,'8600' => 34304,
    '8601' => 34305,'8604' => 34308,
    '8606' => 34310,'8607' => 34311,
    '8609' => 34313,'860A' => 34314,
    '860B' => 34315,'860C' => 34316,
    '8617' => 34327,'8618' => 34328,
    '8619' => 34329,'861A' => 34330,
    '861B' => 34331,'861C' => 34332,
    '861E' => 34334,'861F' => 34335,
    '8620' => 34336,'8621' => 34337,
    '8622' => 34338,'8623' => 34339,
    '8624' => 34340,'8625' => 34341,
    '8626' => 34342,'862A' => 34346,
    '862C' => 34348,'862D' => 34349,
    '862E' => 34350,'8631' => 34353,
    '8632' => 34354,'8633' => 34355,
    '8634' => 34356,'8635' => 34357,
    '8636' => 34358,'8639' => 34361,
    '863A' => 34362,'863B' => 34363,
    '863E' => 34366,'863F' => 34367,
    '8640' => 34368,'8643' => 34371,
    '8646' => 34374,'8647' => 34375,
    '8648' => 34376,'864B' => 34379,
    '864C' => 34380,'8652' => 34386,
    '8653' => 34387,'8655' => 34389,
    '8656' => 34390,'8659' => 34393,
    '865B' => 34395,'865C' => 34396,
    '865F' => 34399,'8661' => 34401,
    '8663' => 34403,'8664' => 34404,
    '8665' => 34405,'8667' => 34407,
    '8668' => 34408,'8669' => 34409,
    '866A' => 34410,'866D' => 34413,
    '866F' => 34415,'8670' => 34416,
    '8673' => 34419,'8674' => 34420,
    '8677' => 34423,'8685' => 34437,
    '8686' => 34438,'8687' => 34439,
    '868E' => 34446,'8690' => 34448,
    '8691' => 34449,'8694' => 34452,
    '8696' => 34454,'8697' => 34455,
    '8698' => 34456,'8699' => 34457,
    '869A' => 34458,'869E' => 34462,
    '86A1' => 34465,'86A2' => 34466,
    '86A5' => 34469,'86B3' => 34483,
    '86B7' => 34487,'86B8' => 34488,
    '86B9' => 34489,'86BB' => 34491,
    '86BC' => 34492,'86BD' => 34493,
    '86BE' => 34494,'86BF' => 34495,
    '86C1' => 34497,'86C2' => 34498,
    '86C3' => 34499,'86C5' => 34501,
    '86C8' => 34504,'86CC' => 34508,
    '86D3' => 34515,'86D6' => 34518,
    '86D7' => 34519,'86DA' => 34522,
    '86DC' => 34524,'86DD' => 34525,
    '86E2' => 34530,'86E3' => 34531,
    '86E6' => 34534,'86E8' => 34536,
    '86EA' => 34538,'86EB' => 34539,
    '86EC' => 34540,'86F5' => 34549,
    '86F6' => 34550,'86F7' => 34551,
    '86FA' => 34554,'86FB' => 34555,
    '8701' => 34561,'8704' => 34564,
    '8705' => 34565,'8706' => 34566,
    '870B' => 34571,'870C' => 34572,
    '870E' => 34574,'8711' => 34577,
    '8719' => 34585,'871B' => 34587,
    '8720' => 34592,'8724' => 34596,
    '8726' => 34598,'8727' => 34599,
    '8728' => 34600,'872A' => 34602,
    '872C' => 34604,'872D' => 34605,
    '8730' => 34608,'8732' => 34610,
    '8733' => 34611,'8735' => 34613,
    '8738' => 34616,'873A' => 34618,
    '873C' => 34620,'8740' => 34624,
    '8741' => 34625,'8742' => 34626,
    '8743' => 34627,'8746' => 34630,
    '874D' => 34637,'874F' => 34639,
    '8750' => 34640,'8751' => 34641,
    '8752' => 34642,'8754' => 34644,
    '8755' => 34645,'8756' => 34646,
    '8758' => 34648,'875A' => 34650,
    '875B' => 34651,'875C' => 34652,
    '875D' => 34653,'875E' => 34654,
    '875F' => 34655,'8761' => 34657,
    '8762' => 34658,'8766' => 34662,
    '8767' => 34663,'8768' => 34664,
    '8769' => 34665,'876A' => 34666,
    '876B' => 34667,'876C' => 34668,
    '876D' => 34669,'876F' => 34671,
    '8773' => 34675,'8775' => 34677,
    '8777' => 34679,'8778' => 34680,
    '8779' => 34681,'877A' => 34682,
    '8781' => 34689,'8784' => 34692,
    '8787' => 34695,'8789' => 34697,
    '878F' => 34703,'8790' => 34704,
    '8791' => 34705,'8792' => 34706,
    '8794' => 34708,'8796' => 34710,
    '8798' => 34712,'879A' => 34714,
    '879B' => 34715,'879C' => 34716,
    '879D' => 34717,'879E' => 34718,
    '87A2' => 34722,'87A3' => 34723,
    '87A4' => 34724,'87AA' => 34730,
    '87AE' => 34734,'87B0' => 34736,
    '87B2' => 34738,'87B4' => 34740,
    '87B6' => 34742,'87B7' => 34743,
    '87B8' => 34744,'87B9' => 34745,
    '87BB' => 34747,'87BC' => 34748,
    '87BE' => 34750,'87BF' => 34751,
    '87C2' => 34754,'87C3' => 34755,
    '87C4' => 34756,'87C5' => 34757,
    '87C8' => 34760,'87C9' => 34761,
    '87CC' => 34764,'87D4' => 34772,
    '87D7' => 34775,'87D8' => 34776,
    '87D9' => 34777,'87DC' => 34780,
    '87DD' => 34781,'87DE' => 34782,
    '87DF' => 34783,'87E1' => 34785,
    '87E2' => 34786,'87E3' => 34787,
    '87E4' => 34788,'87E6' => 34790,
    '87E7' => 34791,'87E8' => 34792,
    '87EB' => 34795,'87EC' => 34796,
    '87ED' => 34797,'87EF' => 34799,
    '87F2' => 34802,'87F3' => 34803,
    '87F4' => 34804,'87F6' => 34806,
    '87F7' => 34807,'87FA' => 34810,
    '87FB' => 34811,'87FC' => 34812,
    '87FF' => 34815,'8800' => 34816,
    '8801' => 34817,'8802' => 34818,
    '8805' => 34821,'8806' => 34822,
    '8808' => 34824,'8809' => 34825,
    '880B' => 34827,'880C' => 34828,
    '880D' => 34829,'8810' => 34832,
    '8811' => 34833,'8814' => 34836,
    '8817' => 34839,'8819' => 34841,
    '881C' => 34844,'881D' => 34845,
    '881F' => 34847,'8820' => 34848,
    '8823' => 34851,'8824' => 34852,
    '8825' => 34853,'8826' => 34854,
    '8828' => 34856,'8829' => 34857,
    '882A' => 34858,'882B' => 34859,
    '882C' => 34860,'882E' => 34862,
    '882F' => 34863,'8830' => 34864,
    '8831' => 34865,'8833' => 34867,
    '8835' => 34869,'8836' => 34870,
    '8837' => 34871,'8838' => 34872,
    '883B' => 34875,'883D' => 34877,
    '883E' => 34878,'883F' => 34879,
    '8841' => 34881,'8843' => 34883,
    '8848' => 34888,'884A' => 34890,
    '884B' => 34891,'884E' => 34894,
    '8852' => 34898,'8853' => 34899,
    '8855' => 34901,'8856' => 34902,
    '885A' => 34906,'885B' => 34907,
    '885D' => 34909,'8867' => 34919,
    '886A' => 34922,'886D' => 34925,
    '886F' => 34927,'8871' => 34929,
    '8874' => 34932,'8875' => 34933,
    '8876' => 34934,'8879' => 34937,
    '887C' => 34940,'8880' => 34944,
    '8883' => 34947,'8889' => 34953,
    '888C' => 34956,'888E' => 34958,
    '8891' => 34961,'8893' => 34963,
    '8895' => 34965,'8897' => 34967,
    '8898' => 34968,'8899' => 34969,
    '889A' => 34970,'889B' => 34971,
    '889E' => 34974,'889F' => 34975,
    '88A1' => 34977,'88A7' => 34983,
    '88A8' => 34984,'88AA' => 34986,
    '88AC' => 34988,'88B2' => 34994,
    '88B6' => 34998,'88B8' => 35000,
    '88B9' => 35001,'88BA' => 35002,
    '88BD' => 35005,'88BE' => 35006,
    '88C0' => 35008,'88CA' => 35018,
    '88CB' => 35019,'88CC' => 35020,
    '88CD' => 35021,'88CF' => 35023,
    '88D0' => 35024,'88D6' => 35030,
    '88D7' => 35031,'88DA' => 35034,
    '88DB' => 35035,'88DC' => 35036,
    '88DD' => 35037,'88DE' => 35038,
    '88E1' => 35041,'88E7' => 35047,
    '88EB' => 35051,'88EC' => 35052,
    '88EE' => 35054,'88EF' => 35055,
    '88F2' => 35058,'88F6' => 35062,
    '88F7' => 35063,'88FA' => 35066,
    '88FB' => 35067,'88FD' => 35069,
    '8901' => 35073,'8905' => 35077,
    '8906' => 35078,'8907' => 35079,
    '8909' => 35081,'890B' => 35083,
    '890C' => 35084,'890E' => 35086,
    '8911' => 35089,'8914' => 35092,
    '8915' => 35093,'8916' => 35094,
    '8917' => 35095,'8918' => 35096,
    '891E' => 35102,'891F' => 35103,
    '8922' => 35106,'8923' => 35107,
    '8926' => 35110,'8927' => 35111,
    '8929' => 35113,'892C' => 35116,
    '892D' => 35117,'892E' => 35118,
    '892F' => 35119,'8931' => 35121,
    '8932' => 35122,'8933' => 35123,
    '8935' => 35125,'8937' => 35127,
    '8938' => 35128,'893B' => 35131,
    '893C' => 35132,'893D' => 35133,
    '893E' => 35134,'8942' => 35138,
    '8946' => 35142,'8949' => 35145,
    '894B' => 35147,'894C' => 35148,
    '894F' => 35151,'8950' => 35152,
    '8951' => 35153,'8952' => 35154,
    '8953' => 35155,'8956' => 35158,
    '8957' => 35159,'8958' => 35160,
    '8959' => 35161,'895A' => 35162,
    '895B' => 35163,'895C' => 35164,
    '895D' => 35165,'8960' => 35168,
    '8961' => 35169,'8962' => 35170,
    '8963' => 35171,'8964' => 35172,
    '8969' => 35177,'896A' => 35178,
    '896B' => 35179,'896C' => 35180,
    '896D' => 35181,'896E' => 35182,
    '896F' => 35183,'8971' => 35185,
    '8972' => 35186,'8973' => 35187,
    '8974' => 35188,'8976' => 35190,
    '8979' => 35193,'897A' => 35194,
    '897C' => 35196,'897E' => 35198,
    '8982' => 35202,'8985' => 35205,
    '8988' => 35208,'898B' => 35211,
    '898F' => 35215,'8993' => 35219,
    '8995' => 35221,'8996' => 35222,
    '8997' => 35223,'8998' => 35224,
    '899B' => 35227,'899C' => 35228,
    '899D' => 35229,'899E' => 35230,
    '899F' => 35231,'89A1' => 35233,
    '89A2' => 35234,'89A3' => 35235,
    '89A4' => 35236,'89A6' => 35238,
    '89AA' => 35242,'89AC' => 35244,
    '89AD' => 35245,'89AE' => 35246,
    '89AF' => 35247,'89B2' => 35250,
    '89B6' => 35254,'89B7' => 35255,
    '89B9' => 35257,'89BA' => 35258,
    '89BD' => 35261,'89BE' => 35262,
    '89BF' => 35263,'89C0' => 35264,
    '89D3' => 35283,'89D4' => 35284,
    '89D5' => 35285,'89D9' => 35289,
    '89DB' => 35291,'89DD' => 35293,
    '89DF' => 35295,'89E0' => 35296,
    '89E1' => 35297,'89E2' => 35298,
    '89E4' => 35300,'89E8' => 35304,
    '89E9' => 35305,'89EC' => 35308,
    '89ED' => 35309,'89F0' => 35312,
    '89F1' => 35313,'89F2' => 35314,
    '89F4' => 35316,'89F6' => 35318,
    '89F7' => 35319,'89F8' => 35320,
    '89FA' => 35322,'89FB' => 35323,
    '89FC' => 35324,'89FE' => 35326,
    '89FF' => 35327,'8A02' => 35330,
    '8A03' => 35331,'8A04' => 35332,
    '8A08' => 35336,'8A0A' => 35338,
    '8A0C' => 35340,'8A0E' => 35342,
    '8A0F' => 35343,'8A10' => 35344,
    '8A11' => 35345,'8A12' => 35346,
    '8A13' => 35347,'8A15' => 35349,
    '8A16' => 35350,'8A17' => 35351,
    '8A18' => 35352,'8A1B' => 35355,
    '8A1D' => 35357,'8A1E' => 35358,
    '8A1F' => 35359,'8A22' => 35362,
    '8A23' => 35363,'8A25' => 35365,
    '8A27' => 35367,'8A2A' => 35370,
    '8A2C' => 35372,'8A2D' => 35373,
    '8A30' => 35376,'8A31' => 35377,
    '8A34' => 35380,'8A36' => 35382,
    '8A39' => 35385,'8A3A' => 35386,
    '8A3B' => 35387,'8A3C' => 35388,
    '8A3F' => 35391,'8A40' => 35392,
    '8A41' => 35393,'8A44' => 35396,
    '8A45' => 35397,'8A46' => 35398,
    '8A4A' => 35402,'8A4C' => 35404,
    '8A4D' => 35405,'8A4E' => 35406,
    '8A4F' => 35407,'8A50' => 35408,
    '8A51' => 35409,'8A52' => 35410,
    '8A54' => 35412,'8A55' => 35413,
    '8A56' => 35414,'8A57' => 35415,
    '8A58' => 35416,'8A59' => 35417,
    '8A5B' => 35419,'8A5E' => 35422,
    '8A60' => 35424,'8A61' => 35425,
    '8A62' => 35426,'8A63' => 35427,
    '8A66' => 35430,'8A68' => 35432,
    '8A69' => 35433,'8A6B' => 35435,
    '8A6C' => 35436,'8A6D' => 35437,
    '8A6E' => 35438,'8A70' => 35440,
    '8A71' => 35441,'8A72' => 35442,
    '8A73' => 35443,'8A74' => 35444,
    '8A75' => 35445,'8A76' => 35446,
    '8A77' => 35447,'8A7A' => 35450,
    '8A7B' => 35451,'8A7C' => 35452,
    '8A7F' => 35455,'8A81' => 35457,
    '8A82' => 35458,'8A83' => 35459,
    '8A84' => 35460,'8A85' => 35461,
    '8A86' => 35462,'8A87' => 35463,
    '8A8B' => 35467,'8A8C' => 35468,
    '8A8D' => 35469,'8A8F' => 35471,
    '8A91' => 35473,'8A92' => 35474,
    '8A95' => 35477,'8A96' => 35478,
    '8A98' => 35480,'8A99' => 35481,
    '8A9A' => 35482,'8A9E' => 35486,
    '8AA0' => 35488,'8AA1' => 35489,
    '8AA3' => 35491,'8AA4' => 35492,
    '8AA5' => 35493,'8AA6' => 35494,
    '8AA7' => 35495,'8AA8' => 35496,
    '8AAA' => 35498,'8AAB' => 35499,
    '8AB0' => 35504,'8AB2' => 35506,
    '8AB6' => 35510,'8AB8' => 35512,
    '8AB9' => 35513,'8ABA' => 35514,
    '8ABB' => 35515,'8ABC' => 35516,
    '8ABD' => 35517,'8ABE' => 35518,
    '8ABF' => 35519,'8AC0' => 35520,
    '8AC2' => 35522,'8AC3' => 35523,
    '8AC4' => 35524,'8AC5' => 35525,
    '8AC6' => 35526,'8AC7' => 35527,
    '8AC8' => 35528,'8AC9' => 35529,
    '8ACB' => 35531,'8ACD' => 35533,
    '8ACF' => 35535,'8AD1' => 35537,
    '8AD2' => 35538,'8AD3' => 35539,
    '8AD4' => 35540,'8AD5' => 35541,
    '8AD6' => 35542,'8AD7' => 35543,
    '8AD8' => 35544,'8AD9' => 35545,
    '8ADB' => 35547,'8ADC' => 35548,
    '8ADD' => 35549,'8ADE' => 35550,
    '8ADF' => 35551,'8AE0' => 35552,
    '8AE1' => 35553,'8AE2' => 35554,
    '8AE4' => 35556,'8AE6' => 35558,
    '8AE7' => 35559,'8AE8' => 35560,
    '8AEB' => 35563,'8AED' => 35565,
    '8AEE' => 35566,'8AEF' => 35567,
    '8AF0' => 35568,'8AF1' => 35569,
    '8AF2' => 35570,'8AF3' => 35571,
    '8AF4' => 35572,'8AF5' => 35573,
    '8AF6' => 35574,'8AF7' => 35575,
    '8AF8' => 35576,'8AFA' => 35578,
    '8AFB' => 35579,'8AFC' => 35580,
    '8AFE' => 35582,'8AFF' => 35583,
    '8B00' => 35584,'8B01' => 35585,
    '8B02' => 35586,'8B04' => 35588,
    '8B05' => 35589,'8B06' => 35590,
    '8B08' => 35592,'8B0A' => 35594,
    '8B0B' => 35595,'8B0D' => 35597,
    '8B0E' => 35598,'8B0F' => 35599,
    '8B10' => 35600,'8B11' => 35601,
    '8B12' => 35602,'8B13' => 35603,
    '8B14' => 35604,'8B15' => 35605,
    '8B16' => 35606,'8B17' => 35607,
    '8B18' => 35608,'8B19' => 35609,
    '8B1A' => 35610,'8B1B' => 35611,
    '8B1C' => 35612,'8B1D' => 35613,
    '8B1E' => 35614,'8B20' => 35616,
    '8B22' => 35618,'8B23' => 35619,
    '8B24' => 35620,'8B25' => 35621,
    '8B27' => 35623,'8B28' => 35624,
    '8B2A' => 35626,'8B2B' => 35627,
    '8B2C' => 35628,'8B2E' => 35630,
    '8B2F' => 35631,'8B30' => 35632,
    '8B31' => 35633,'8B33' => 35635,
    '8B35' => 35637,'8B36' => 35638,
    '8B37' => 35639,'8B39' => 35641,
    '8B3A' => 35642,'8B3B' => 35643,
    '8B3C' => 35644,'8B3D' => 35645,
    '8B3E' => 35646,'8B40' => 35648,
    '8B41' => 35649,'8B42' => 35650,
    '8B45' => 35653,'8B46' => 35654,
    '8B47' => 35655,'8B48' => 35656,
    '8B49' => 35657,'8B4A' => 35658,
    '8B4B' => 35659,'8B4E' => 35662,
    '8B4F' => 35663,'8B50' => 35664,
    '8B51' => 35665,'8B52' => 35666,
    '8B53' => 35667,'8B54' => 35668,
    '8B55' => 35669,'8B56' => 35670,
    '8B57' => 35671,'8B58' => 35672,
    '8B59' => 35673,'8B5A' => 35674,
    '8B5C' => 35676,'8B5D' => 35677,
    '8B5F' => 35679,'8B60' => 35680,
    '8B63' => 35683,'8B65' => 35685,
    '8B67' => 35687,'8B68' => 35688,
    '8B6A' => 35690,'8B6B' => 35691,
    '8B6D' => 35693,'8B6F' => 35695,
    '8B70' => 35696,'8B74' => 35700,
    '8B77' => 35703,'8B78' => 35704,
    '8B79' => 35705,'8B7A' => 35706,
    '8B7B' => 35707,'8B7D' => 35709,
    '8B7E' => 35710,'8B7F' => 35711,
    '8B80' => 35712,'8B82' => 35714,
    '8B84' => 35716,'8B85' => 35717,
    '8B86' => 35718,'8B88' => 35720,
    '8B8A' => 35722,'8B8B' => 35723,
    '8B8C' => 35724,'8B8E' => 35726,
    '8B92' => 35730,'8B93' => 35731,
    '8B94' => 35732,'8B95' => 35733,
    '8B96' => 35734,'8B98' => 35736,
    '8B99' => 35737,'8B9A' => 35738,
    '8B9C' => 35740,'8B9E' => 35742,
    '8B9F' => 35743,'8C39' => 35897,
    '8C3B' => 35899,'8C3C' => 35900,
    '8C3D' => 35901,'8C3E' => 35902,
    '8C3F' => 35903,'8C42' => 35906,
    '8C43' => 35907,'8C45' => 35909,
    '8C48' => 35912,'8C4A' => 35914,
    '8C4B' => 35915,'8C4D' => 35917,
    '8C4E' => 35918,'8C4F' => 35919,
    '8C50' => 35920,'8C54' => 35924,
    '8C56' => 35926,'8C57' => 35927,
    '8C5C' => 35932,'8C5D' => 35933,
    '8C5F' => 35935,'8C64' => 35940,
    '8C65' => 35941,'8C66' => 35942,
    '8C68' => 35944,'8C69' => 35945,
    '8C6C' => 35948,'8C6D' => 35949,
    '8C6F' => 35951,'8C70' => 35952,
    '8C71' => 35953,'8C72' => 35954,
    '8C75' => 35957,'8C76' => 35958,
    '8C77' => 35959,'8C7B' => 35963,
    '8C7D' => 35965,'8C80' => 35968,
    '8C81' => 35969,'8C84' => 35972,
    '8C86' => 35974,'8C8D' => 35981,
    '8C8F' => 35983,'8C90' => 35984,
    '8C91' => 35985,'8C92' => 35986,
    '8C93' => 35987,'8C95' => 35989,
    '8C97' => 35991,'8C99' => 35993,
    '8C9A' => 35994,'8C9C' => 35996,
    '8C9D' => 35997,'8C9E' => 35998,
    '8CA0' => 36000,'8CA1' => 36001,
    '8CA2' => 36002,'8CA3' => 36003,
    '8CA4' => 36004,'8CA5' => 36005,
    '8CA7' => 36007,'8CA8' => 36008,
    '8CA9' => 36009,'8CAA' => 36010,
    '8CAB' => 36011,'8CAC' => 36012,
    '8CAF' => 36015,'8CB0' => 36016,
    '8CB2' => 36018,'8CB3' => 36019,
    '8CB4' => 36020,'8CB5' => 36021,
    '8CB6' => 36022,'8CB7' => 36023,
    '8CB8' => 36024,'8CB9' => 36025,
    '8CBA' => 36026,'8CBB' => 36027,
    '8CBC' => 36028,'8CBD' => 36029,
    '8CBE' => 36030,'8CBF' => 36031,
    '8CC0' => 36032,'8CC1' => 36033,
    '8CC2' => 36034,'8CC3' => 36035,
    '8CC4' => 36036,'8CC5' => 36037,
    '8CC7' => 36039,'8CC8' => 36040,
    '8CCA' => 36042,'8CCC' => 36044,
    '8CCF' => 36047,'8CD1' => 36049,
    '8CD2' => 36050,'8CD3' => 36051,
    '8CD5' => 36053,'8CD7' => 36055,
    '8CD9' => 36057,'8CDA' => 36058,
    '8CDC' => 36060,'8CDD' => 36061,
    '8CDE' => 36062,'8CDF' => 36063,
    '8CE0' => 36064,'8CE1' => 36065,
    '8CE2' => 36066,'8CE3' => 36067,
    '8CE4' => 36068,'8CE5' => 36069,
    '8CE6' => 36070,'8CE7' => 36071,
    '8CE8' => 36072,'8CEA' => 36074,
    '8CEC' => 36076,'8CED' => 36077,
    '8CEE' => 36078,'8CF0' => 36080,
    '8CF1' => 36081,'8CF3' => 36083,
    '8CF4' => 36084,'8CF5' => 36085,
    '8CF8' => 36088,'8CF9' => 36089,
    '8CFA' => 36090,'8CFB' => 36091,
    '8CFC' => 36092,'8CFD' => 36093,
    '8CFE' => 36094,'8D00' => 36096,
    '8D02' => 36098,'8D04' => 36100,
    '8D05' => 36101,'8D06' => 36102,
    '8D07' => 36103,'8D08' => 36104,
    '8D09' => 36105,'8D0A' => 36106,
    '8D0D' => 36109,'8D0F' => 36111,
    '8D10' => 36112,'8D13' => 36115,
    '8D14' => 36116,'8D15' => 36117,
    '8D16' => 36118,'8D17' => 36119,
    '8D19' => 36121,'8D1B' => 36123,
    '8D68' => 36200,'8D69' => 36201,
    '8D6C' => 36204,'8D6E' => 36206,
    '8D6F' => 36207,'8D72' => 36210,
    '8D78' => 36216,'8D79' => 36217,
    '8D7B' => 36219,'8D7D' => 36221,
    '8D80' => 36224,'8D89' => 36233,
    '8D8C' => 36236,'8D8D' => 36237,
    '8D8E' => 36238,'8D8F' => 36239,
    '8D90' => 36240,'8D92' => 36242,
    '8D93' => 36243,'8D95' => 36245,
    '8D96' => 36246,'8D99' => 36249,
    '8D9B' => 36251,'8D9C' => 36252,
    '8DA0' => 36256,'8DA1' => 36257,
    '8DA5' => 36261,'8DA7' => 36263,
    '8DA8' => 36264,'8DAA' => 36266,
    '8DAB' => 36267,'8DAC' => 36268,
    '8DAD' => 36269,'8DAE' => 36270,
    '8DAF' => 36271,'8DB2' => 36274,
    '8DB6' => 36278,'8DB7' => 36279,
    '8DB9' => 36281,'8DC1' => 36289,
    '8DC2' => 36290,'8DC5' => 36293,
    '8DC7' => 36295,'8DC8' => 36296,
    '8DCD' => 36301,'8DD0' => 36304,
    '8DD3' => 36307,'8DD5' => 36309,
    '8DD8' => 36312,'8DD9' => 36313,
    '8DDC' => 36316,'8DE0' => 36320,
    '8DE1' => 36321,'8DE2' => 36322,
    '8DE6' => 36326,'8DE7' => 36327,
    '8DE9' => 36329,'8DEE' => 36334,
    '8DF0' => 36336,'8DF1' => 36337,
    '8DF2' => 36338,'8DF4' => 36340,
    '8DFC' => 36348,'8DFE' => 36350,
    '8DFF' => 36351,'8E00' => 36352,
    '8E02' => 36354,'8E03' => 36355,
    '8E04' => 36356,'8E06' => 36358,
    '8E07' => 36359,'8E0D' => 36365,
    '8E10' => 36368,'8E11' => 36369,
    '8E12' => 36370,'8E13' => 36371,
    '8E15' => 36373,'8E16' => 36374,
    '8E17' => 36375,'8E18' => 36376,
    '8E19' => 36377,'8E1A' => 36378,
    '8E1B' => 36379,'8E1C' => 36380,
    '8E20' => 36384,'8E21' => 36385,
    '8E24' => 36388,'8E25' => 36389,
    '8E26' => 36390,'8E27' => 36391,
    '8E2B' => 36395,'8E30' => 36400,
    '8E33' => 36403,'8E34' => 36404,
    '8E36' => 36406,'8E38' => 36408,
    '8E3C' => 36412,'8E3E' => 36414,
    '8E3F' => 36415,'8E45' => 36421,
    '8E4C' => 36428,'8E4D' => 36429,
    '8E4E' => 36430,'8E50' => 36432,
    '8E53' => 36435,'8E54' => 36436,
    '8E55' => 36437,'8E56' => 36438,
    '8E57' => 36439,'8E5A' => 36442,
    '8E5B' => 36443,'8E5C' => 36444,
    '8E5D' => 36445,'8E5E' => 36446,
    '8E5F' => 36447,'8E60' => 36448,
    '8E61' => 36449,'8E62' => 36450,
    '8E63' => 36451,'8E64' => 36452,
    '8E65' => 36453,'8E67' => 36455,
    '8E6A' => 36458,'8E73' => 36467,
    '8E78' => 36472,'8E7A' => 36474,
    '8E7B' => 36475,'8E82' => 36482,
    '8E84' => 36484,'8E86' => 36486,
    '8E88' => 36488,'8E89' => 36489,
    '8E8A' => 36490,'8E8B' => 36491,
    '8E8C' => 36492,'8E8D' => 36493,
    '8E8E' => 36494,'8E91' => 36497,
    '8E92' => 36498,'8E93' => 36499,
    '8E95' => 36501,'8E96' => 36502,
    '8E97' => 36503,'8E98' => 36504,
    '8E9A' => 36506,'8E9D' => 36509,
    '8E9F' => 36511,'8EA0' => 36512,
    '8EA1' => 36513,'8EA3' => 36515,
    '8EA4' => 36516,'8EA5' => 36517,
    '8EA6' => 36518,'8EA8' => 36520,
    '8EA9' => 36521,'8EAA' => 36522,
    '8EBD' => 36541,'8EC0' => 36544,
    '8EC2' => 36546,'8EC9' => 36553,
    '8ECA' => 36554,'8ECB' => 36555,
    '8ECC' => 36556,'8ECD' => 36557,
    '8ECF' => 36559,'8ED1' => 36561,
    '8ED2' => 36562,'8ED3' => 36563,
    '8ED4' => 36564,'8ED7' => 36567,
    '8ED8' => 36568,'8EDB' => 36571,
    '8EDC' => 36572,'8EDD' => 36573,
    '8EDE' => 36574,'8EDF' => 36575,
    '8EE0' => 36576,'8EE1' => 36577,
    '8EE5' => 36581,'8EE6' => 36582,
    '8EE7' => 36583,'8EE8' => 36584,
    '8EE9' => 36585,'8EEB' => 36587,
    '8EEC' => 36588,'8EEE' => 36590,
    '8EEF' => 36591,'8EF1' => 36593,
    '8EF4' => 36596,'8EF5' => 36597,
    '8EF6' => 36598,'8EF7' => 36599,
    '8EF8' => 36600,'8EF9' => 36601,
    '8EFA' => 36602,'8EFB' => 36603,
    '8EFC' => 36604,'8EFE' => 36606,
    '8EFF' => 36607,'8F00' => 36608,
    '8F01' => 36609,'8F02' => 36610,
    '8F03' => 36611,'8F05' => 36613,
    '8F06' => 36614,'8F07' => 36615,
    '8F08' => 36616,'8F09' => 36617,
    '8F0A' => 36618,'8F0B' => 36619,
    '8F0D' => 36621,'8F0E' => 36622,
    '8F10' => 36624,'8F11' => 36625,
    '8F12' => 36626,'8F13' => 36627,
    '8F14' => 36628,'8F15' => 36629,
    '8F16' => 36630,'8F17' => 36631,
    '8F18' => 36632,'8F1A' => 36634,
    '8F1B' => 36635,'8F1C' => 36636,
    '8F1D' => 36637,'8F1E' => 36638,
    '8F1F' => 36639,'8F20' => 36640,
    '8F23' => 36643,'8F24' => 36644,
    '8F25' => 36645,'8F26' => 36646,
    '8F29' => 36649,'8F2A' => 36650,
    '8F2C' => 36652,'8F2E' => 36654,
    '8F2F' => 36655,'8F32' => 36658,
    '8F33' => 36659,'8F34' => 36660,
    '8F35' => 36661,'8F36' => 36662,
    '8F37' => 36663,'8F38' => 36664,
    '8F39' => 36665,'8F3B' => 36667,
    '8F3E' => 36670,'8F3F' => 36671,
    '8F40' => 36672,'8F42' => 36674,
    '8F43' => 36675,'8F44' => 36676,
    '8F45' => 36677,'8F46' => 36678,
    '8F47' => 36679,'8F48' => 36680,
    '8F49' => 36681,'8F4B' => 36683,
    '8F4D' => 36685,'8F4E' => 36686,
    '8F4F' => 36687,'8F50' => 36688,
    '8F51' => 36689,'8F52' => 36690,
    '8F53' => 36691,'8F54' => 36692,
    '8F55' => 36693,'8F56' => 36694,
    '8F57' => 36695,'8F58' => 36696,
    '8F59' => 36697,'8F5A' => 36698,
    '8F5B' => 36699,'8F5D' => 36701,
    '8F5E' => 36702,'8F5F' => 36703,
    '8F60' => 36704,'8F61' => 36705,
    '8F62' => 36706,'8F63' => 36707,
    '8F64' => 36708,'8FA6' => 36774,
    '8FAD' => 36781,'8FAE' => 36782,
    '8FAF' => 36783,'8FB2' => 36786,
    '8FB4' => 36788,'8FBF' => 36799,
    '8FC6' => 36806,'8FC9' => 36809,
    '8FCB' => 36811,'8FCD' => 36813,
    '8FD2' => 36818,'8FD6' => 36822,
    '8FD7' => 36823,'8FE0' => 36832,
    '8FE1' => 36833,'8FE3' => 36835,
    '8FF4' => 36852,'8FF5' => 36853,
    '8FF6' => 36854,'8FFA' => 36858,
    '8FFB' => 36859,'8FFC' => 36860,
    '8FFE' => 36862,'8FFF' => 36863,
    '900C' => 36876,'9015' => 36885,
    '9019' => 36889,'901C' => 36892,
    '9023' => 36899,'9024' => 36900,
    '9031' => 36913,'9032' => 36914,
    '9034' => 36916,'903D' => 36925,
    '903F' => 36927,'9049' => 36937,
    '904A' => 36938,'904B' => 36939,
    '904E' => 36942,'9054' => 36948,
    '9055' => 36949,'9059' => 36953,
    '905C' => 36956,'905D' => 36957,
    '905E' => 36958,'9060' => 36960,
    '9067' => 36967,'9069' => 36969,
    '906B' => 36971,'906F' => 36975,
    '9070' => 36976,'9072' => 36978,
    '9073' => 36979,'9076' => 36982,
    '9077' => 36983,'9078' => 36984,
    '9079' => 36985,'907A' => 36986,
    '907B' => 36987,'907C' => 36988,
    '907E' => 36990,'9081' => 36993,
    '9084' => 36996,'9085' => 36997,
    '9086' => 36998,'9087' => 36999,
    '908A' => 37002,'908D' => 37005,
    '908F' => 37007,'9090' => 37008,
    '9094' => 37012,'9098' => 37016,
    '909E' => 37022,'909F' => 37023,
    '90A0' => 37024,'90A5' => 37029,
    '90A7' => 37031,'90B2' => 37042,
    '90BD' => 37053,'90BF' => 37055,
    '90C3' => 37059,'90C8' => 37064,
    '90CB' => 37067,'90D4' => 37076,
    '90D5' => 37077,'90D6' => 37078,
    '90D8' => 37080,'90D9' => 37081,
    '90DA' => 37082,'90DF' => 37087,
    '90E0' => 37088,'90E3' => 37091,
    '90E4' => 37092,'90E5' => 37093,
    '90E9' => 37097,'90EA' => 37098,
    '90EC' => 37100,'90F0' => 37104,
    '90F1' => 37105,'90F2' => 37106,
    '90F3' => 37107,'90F5' => 37109,
    '90F9' => 37113,'90FA' => 37114,
    '90FB' => 37115,'90FC' => 37116,
    '90FF' => 37119,'9100' => 37120,
    '9101' => 37121,'9103' => 37123,
    '9105' => 37125,'9106' => 37126,
    '9107' => 37127,'9108' => 37128,
    '9109' => 37129,'910B' => 37131,
    '910D' => 37133,'910E' => 37134,
    '910F' => 37135,'9110' => 37136,
    '9111' => 37137,'9112' => 37138,
    '9114' => 37140,'9116' => 37142,
    '9117' => 37143,'9118' => 37144,
    '911A' => 37146,'911B' => 37147,
    '911C' => 37148,'911D' => 37149,
    '911F' => 37151,'9120' => 37152,
    '9121' => 37153,'9124' => 37156,
    '9126' => 37158,'9127' => 37159,
    '9128' => 37160,'9129' => 37161,
    '912A' => 37162,'912B' => 37163,
    '912C' => 37164,'912D' => 37165,
    '912E' => 37166,'9130' => 37168,
    '9132' => 37170,'9133' => 37171,
    '9134' => 37172,'9135' => 37173,
    '9136' => 37174,'9138' => 37176,
    '913A' => 37178,'913B' => 37179,
    '913E' => 37182,'913F' => 37183,
    '9140' => 37184,'9141' => 37185,
    '9144' => 37188,'9145' => 37189,
    '9147' => 37191,'9148' => 37192,
    '9153' => 37203,'9155' => 37205,
    '9156' => 37206,'9158' => 37208,
    '915F' => 37215,'9160' => 37216,
    '9168' => 37224,'9173' => 37235,
    '917A' => 37242,'9180' => 37248,
    '9181' => 37249,'9182' => 37250,
    '9183' => 37251,'9184' => 37252,
    '9186' => 37254,'918A' => 37258,
    '918F' => 37263,'9193' => 37267,
    '9199' => 37273,'919C' => 37276,
    '919D' => 37277,'919E' => 37278,
    '919F' => 37279,'91A0' => 37280,
    '91A1' => 37281,'91A5' => 37285,
    '91A7' => 37287,'91A8' => 37288,
    '91AB' => 37291,'91AC' => 37292,
    '91B0' => 37296,'91B1' => 37297,
    '91B2' => 37298,'91B3' => 37299,
    '91B7' => 37303,'91B9' => 37305,
    '91BC' => 37308,'91BD' => 37309,
    '91BE' => 37310,'91C0' => 37312,
    '91C1' => 37313,'91C2' => 37314,
    '91C3' => 37315,'91C5' => 37317,
    '91C6' => 37318,'91CB' => 37323,
    '91D0' => 37328,'91D3' => 37331,
    '91D4' => 37332,'91D5' => 37333,
    '91D7' => 37335,'91D8' => 37336,
    '91D9' => 37337,'91DA' => 37338,
    '91DD' => 37341,'91E2' => 37346,
    '91E3' => 37347,'91E4' => 37348,
    '91E6' => 37350,'91E7' => 37351,
    '91E8' => 37352,'91E9' => 37353,
    '91EA' => 37354,'91EB' => 37355,
    '91EC' => 37356,'91ED' => 37357,
    '91EE' => 37358,'91F1' => 37361,
    '91F3' => 37363,'91F4' => 37364,
    '91F5' => 37365,'91F7' => 37367,
    '91F8' => 37368,'91F9' => 37369,
    '91FD' => 37373,'91FF' => 37375,
    '9200' => 37376,'9201' => 37377,
    '9202' => 37378,'9203' => 37379,
    '9204' => 37380,'9205' => 37381,
    '9206' => 37382,'9207' => 37383,
    '9209' => 37385,'920A' => 37386,
    '920C' => 37388,'920D' => 37389,
    '920F' => 37391,'9210' => 37392,
    '9211' => 37393,'9212' => 37394,
    '9214' => 37396,'9215' => 37397,
    '9216' => 37398,'9217' => 37399,
    '9219' => 37401,'921A' => 37402,
    '921C' => 37404,'921E' => 37406,
    '9223' => 37411,'9224' => 37412,
    '9225' => 37413,'9226' => 37414,
    '9227' => 37415,'922D' => 37421,
    '922E' => 37422,'9230' => 37424,
    '9231' => 37425,'9232' => 37426,
    '9233' => 37427,'9234' => 37428,
    '9236' => 37430,'9237' => 37431,
    '9238' => 37432,'9239' => 37433,
    '923A' => 37434,'923D' => 37437,
    '923E' => 37438,'923F' => 37439,
    '9240' => 37440,'9245' => 37445,
    '9246' => 37446,'9248' => 37448,
    '9249' => 37449,'924A' => 37450,
    '924B' => 37451,'924C' => 37452,
    '924D' => 37453,'924E' => 37454,
    '924F' => 37455,'9250' => 37456,
    '9251' => 37457,'9252' => 37458,
    '9253' => 37459,'9254' => 37460,
    '9256' => 37462,'9257' => 37463,
    '925A' => 37466,'925B' => 37467,
    '925E' => 37470,'9260' => 37472,
    '9261' => 37473,'9263' => 37475,
    '9264' => 37476,'9265' => 37477,
    '9266' => 37478,'9267' => 37479,
    '926C' => 37484,'926D' => 37485,
    '926F' => 37487,'9270' => 37488,
    '9272' => 37490,'9276' => 37494,
    '9278' => 37496,'9279' => 37497,
    '927A' => 37498,'927B' => 37499,
    '927C' => 37500,'927D' => 37501,
    '927E' => 37502,'927F' => 37503,
    '9280' => 37504,'9282' => 37506,
    '9283' => 37507,'9285' => 37509,
    '9286' => 37510,'9287' => 37511,
    '9288' => 37512,'928A' => 37514,
    '928B' => 37515,'928C' => 37516,
    '928D' => 37517,'9291' => 37521,
    '9293' => 37523,'9294' => 37524,
    '9295' => 37525,'9296' => 37526,
    '9297' => 37527,'9298' => 37528,
    '9299' => 37529,'929A' => 37530,
    '929B' => 37531,'929C' => 37532,
    '929D' => 37533,'92A0' => 37536,
    '92A1' => 37537,'92A2' => 37538,
    '92A3' => 37539,'92A4' => 37540,
    '92A5' => 37541,'92A6' => 37542,
    '92A7' => 37543,'92A8' => 37544,
    '92A9' => 37545,'92AA' => 37546,
    '92AB' => 37547,'92AC' => 37548,
    '92B2' => 37554,'92B3' => 37555,
    '92B4' => 37556,'92B5' => 37557,
    '92B6' => 37558,'92B7' => 37559,
    '92B9' => 37561,'92BB' => 37563,
    '92BC' => 37564,'92C0' => 37568,
    '92C1' => 37569,'92C2' => 37570,
    '92C3' => 37571,'92C4' => 37572,
    '92C5' => 37573,'92C6' => 37574,
    '92C7' => 37575,'92C9' => 37577,
    '92CA' => 37578,'92CB' => 37579,
    '92CC' => 37580,'92CD' => 37581,
    '92CE' => 37582,'92CF' => 37583,
    '92D0' => 37584,'92D1' => 37585,
    '92D2' => 37586,'92D3' => 37587,
    '92D5' => 37589,'92D7' => 37591,
    '92D8' => 37592,'92D9' => 37593,
    '92DD' => 37597,'92DE' => 37598,
    '92DF' => 37599,'92E0' => 37600,
    '92E1' => 37601,'92E4' => 37604,
    '92E6' => 37606,'92E7' => 37607,
    '92E8' => 37608,'92E9' => 37609,
    '92EA' => 37610,'92EE' => 37614,
    '92EF' => 37615,'92F0' => 37616,
    '92F1' => 37617,'92F7' => 37623,
    '92F8' => 37624,'92F9' => 37625,
    '92FA' => 37626,'92FB' => 37627,
    '92FC' => 37628,'92FE' => 37630,
    '92FF' => 37631,'9300' => 37632,
    '9301' => 37633,'9302' => 37634,
    '9304' => 37636,'9306' => 37638,
    '9308' => 37640,'9309' => 37641,
    '930B' => 37643,'930C' => 37644,
    '930D' => 37645,'930E' => 37646,
    '930F' => 37647,'9310' => 37648,
    '9312' => 37650,'9313' => 37651,
    '9314' => 37652,'9315' => 37653,
    '9316' => 37654,'9318' => 37656,
    '9319' => 37657,'931A' => 37658,
    '931B' => 37659,'931D' => 37661,
    '931E' => 37662,'931F' => 37663,
    '9320' => 37664,'9321' => 37665,
    '9322' => 37666,'9323' => 37667,
    '9324' => 37668,'9325' => 37669,
    '9326' => 37670,'9327' => 37671,
    '9328' => 37672,'9329' => 37673,
    '932A' => 37674,'932B' => 37675,
    '932D' => 37677,'932E' => 37678,
    '932F' => 37679,'9333' => 37683,
    '9334' => 37684,'9335' => 37685,
    '9336' => 37686,'9338' => 37688,
    '9339' => 37689,'933C' => 37692,
    '9346' => 37702,'9347' => 37703,
    '9349' => 37705,'934A' => 37706,
    '934B' => 37707,'934C' => 37708,
    '934D' => 37709,'934E' => 37710,
    '934F' => 37711,'9350' => 37712,
    '9351' => 37713,'9352' => 37714,
    '9354' => 37716,'9355' => 37717,
    '9356' => 37718,'9357' => 37719,
    '9358' => 37720,'9359' => 37721,
    '935A' => 37722,'935B' => 37723,
    '935C' => 37724,'935E' => 37726,
    '9360' => 37728,'9361' => 37729,
    '9363' => 37731,'9364' => 37732,
    '9365' => 37733,'9367' => 37735,
    '936C' => 37740,'936D' => 37741,
    '9370' => 37744,'9371' => 37745,
    '9375' => 37749,'9376' => 37750,
    '9377' => 37751,'9379' => 37753,
    '937A' => 37754,'937B' => 37755,
    '937C' => 37756,'937E' => 37758,
    '9380' => 37760,'9382' => 37762,
    '9383' => 37763,'9388' => 37768,
    '9389' => 37769,'938A' => 37770,
    '938C' => 37772,'938D' => 37773,
    '938E' => 37774,'9391' => 37777,
    '9392' => 37778,'9394' => 37780,
    '9395' => 37781,'9396' => 37782,
    '9397' => 37783,'9398' => 37784,
    '9399' => 37785,'939A' => 37786,
    '939B' => 37787,'939D' => 37789,
    '939E' => 37790,'939F' => 37791,
    '93A1' => 37793,'93A2' => 37794,
    '93A3' => 37795,'93A4' => 37796,
    '93A5' => 37797,'93A6' => 37798,
    '93A7' => 37799,'93A8' => 37800,
    '93A9' => 37801,'93AA' => 37802,
    '93AC' => 37804,'93AE' => 37806,
    '93AF' => 37807,'93B0' => 37808,
    '93B1' => 37809,'93B2' => 37810,
    '93B3' => 37811,'93B4' => 37812,
    '93B5' => 37813,'93B7' => 37815,
    '93C0' => 37824,'93C2' => 37826,
    '93C3' => 37827,'93C4' => 37828,
    '93C7' => 37831,'93C8' => 37832,
    '93CC' => 37836,'93CD' => 37837,
    '93CE' => 37838,'93CF' => 37839,
    '93D0' => 37840,'93D1' => 37841,
    '93D2' => 37842,'93D4' => 37844,
    '93D5' => 37845,'93D7' => 37847,
    '93D8' => 37848,'93D9' => 37849,
    '93DA' => 37850,'93DC' => 37852,
    '93DD' => 37853,'93DE' => 37854,
    '93DF' => 37855,'93E1' => 37857,
    '93E2' => 37858,'93E3' => 37859,
    '93E4' => 37860,'93E6' => 37862,
    '93E7' => 37863,'93E8' => 37864,
    '93EC' => 37868,'93EE' => 37870,
    '93F5' => 37877,'93F6' => 37878,
    '93F7' => 37879,'93F8' => 37880,
    '93F9' => 37881,'93FA' => 37882,
    '93FB' => 37883,'93FC' => 37884,
    '93FD' => 37885,'93FE' => 37886,
    '93FF' => 37887,'9400' => 37888,
    '9403' => 37891,'9406' => 37894,
    '9407' => 37895,'9409' => 37897,
    '940A' => 37898,'940B' => 37899,
    '940C' => 37900,'940D' => 37901,
    '940E' => 37902,'940F' => 37903,
    '9410' => 37904,'9411' => 37905,
    '9412' => 37906,'9413' => 37907,
    '9414' => 37908,'9415' => 37909,
    '9416' => 37910,'9418' => 37912,
    '9419' => 37913,'9420' => 37920,
    '9428' => 37928,'9429' => 37929,
    '942A' => 37930,'942B' => 37931,
    '942C' => 37932,'942E' => 37934,
    '9430' => 37936,'9431' => 37937,
    '9432' => 37938,'9433' => 37939,
    '9435' => 37941,'9436' => 37942,
    '9437' => 37943,'9438' => 37944,
    '9439' => 37945,'943A' => 37946,
    '943B' => 37947,'943C' => 37948,
    '943D' => 37949,'943F' => 37951,
    '9440' => 37952,'9444' => 37956,
    '9445' => 37957,'9446' => 37958,
    '9447' => 37959,'9448' => 37960,
    '9449' => 37961,'944A' => 37962,
    '944B' => 37963,'944C' => 37964,
    '944F' => 37967,'9450' => 37968,
    '9451' => 37969,'9452' => 37970,
    '9455' => 37973,'9457' => 37975,
    '945D' => 37981,'945E' => 37982,
    '9460' => 37984,'9462' => 37986,
    '9463' => 37987,'9464' => 37988,
    '9468' => 37992,'9469' => 37993,
    '946A' => 37994,'946D' => 37997,
    '946E' => 37998,'946F' => 37999,
    '9470' => 38000,'9471' => 38001,
    '9472' => 38002,'9473' => 38003,
    '9474' => 38004,'9475' => 38005,
    '9476' => 38006,'9477' => 38007,
    '9478' => 38008,'947C' => 38012,
    '947D' => 38013,'947E' => 38014,
    '947F' => 38015,'9480' => 38016,
    '9481' => 38017,'9482' => 38018,
    '9483' => 38019,'9577' => 38263,
    '957A' => 38266,'957B' => 38267,
    '957C' => 38268,'957D' => 38269,
    '9580' => 38272,'9582' => 38274,
    '9583' => 38275,'9586' => 38278,
    '9588' => 38280,'9589' => 38281,
    '958B' => 38283,'958C' => 38284,
    '958D' => 38285,'958E' => 38286,
    '958F' => 38287,'9590' => 38288,
    '9591' => 38289,'9592' => 38290,
    '9593' => 38291,'9594' => 38292,
    '9598' => 38296,'959B' => 38299,
    '959C' => 38300,'959E' => 38302,
    '959F' => 38303,'95A1' => 38305,
    '95A3' => 38307,'95A4' => 38308,
    '95A5' => 38309,'95A8' => 38312,
    '95A9' => 38313,'95AB' => 38315,
    '95AC' => 38316,'95AD' => 38317,
    '95AE' => 38318,'95B0' => 38320,
    '95B1' => 38321,'95B5' => 38325,
    '95B6' => 38326,'95B7' => 38327,
    '95B9' => 38329,'95BA' => 38330,
    '95BB' => 38331,'95BC' => 38332,
    '95BD' => 38333,'95BE' => 38334,
    '95BF' => 38335,'95C0' => 38336,
    '95C3' => 38339,'95C5' => 38341,
    '95C6' => 38342,'95C7' => 38343,
    '95C8' => 38344,'95C9' => 38345,
    '95CA' => 38346,'95CB' => 38347,
    '95CC' => 38348,'95CD' => 38349,
    '95D0' => 38352,'95D1' => 38353,
    '95D2' => 38354,'95D3' => 38355,
    '95D4' => 38356,'95D5' => 38357,
    '95D6' => 38358,'95DA' => 38362,
    '95DB' => 38363,'95DC' => 38364,
    '95DE' => 38366,'95DF' => 38367,
    '95E0' => 38368,'95E1' => 38369,
    '95E2' => 38370,'95E3' => 38371,
    '95E4' => 38372,'95E5' => 38373,
    '961E' => 38430,'9620' => 38432,
    '9623' => 38435,'9624' => 38436,
    '9628' => 38440,'962C' => 38444,
    '962D' => 38445,'962F' => 38447,
    '9630' => 38448,'9639' => 38457,
    '963A' => 38458,'9643' => 38467,
    '964A' => 38474,'964E' => 38478,
    '964F' => 38479,'9651' => 38481,
    '9653' => 38483,'9658' => 38488,
    '965C' => 38492,'965D' => 38493,
    '965E' => 38494,'9663' => 38499,
    '966B' => 38507,'966D' => 38509,
    '966F' => 38511,'9670' => 38512,
    '9671' => 38513,'9673' => 38515,
    '9678' => 38520,'967C' => 38524,
    '967D' => 38525,'967E' => 38526,
    '9680' => 38528,'9683' => 38531,
    '9684' => 38532,'9687' => 38535,
    '9689' => 38537,'968A' => 38538,
    '968E' => 38542,'9691' => 38545,
    '9692' => 38546,'9693' => 38547,
    '9695' => 38549,'969B' => 38555,
    '969E' => 38558,'96A1' => 38561,
    '96A2' => 38562,'96A4' => 38564,
    '96A8' => 38568,'96A9' => 38569,
    '96AA' => 38570,'96AC' => 38572,
    '96AE' => 38574,'96B1' => 38577,
    '96B4' => 38580,'96B8' => 38584,
    '96BB' => 38587,'96BF' => 38591,
    '96C2' => 38594,'96C3' => 38595,
    '96C8' => 38600,'96CA' => 38602,
    '96CB' => 38603,'96D3' => 38611,
    '96D4' => 38612,'96D6' => 38614,
    '96D7' => 38615,'96D8' => 38616,
    '96D9' => 38617,'96DA' => 38618,
    '96DB' => 38619,'96DC' => 38620,
    '96DD' => 38621,'96DE' => 38622,
    '96DF' => 38623,'96E1' => 38625,
    '96E2' => 38626,'96E3' => 38627,
    '96E5' => 38629,'96F0' => 38640,
    '96F1' => 38641,'96F2' => 38642,
    '96F5' => 38645,'96F8' => 38648,
    '96FA' => 38650,'96FB' => 38651,
    '96FD' => 38653,'96FF' => 38655,
    '9702' => 38658,'9705' => 38661,
    '970B' => 38667,'9710' => 38672,
    '9711' => 38673,'9712' => 38674,
    '9718' => 38680,'9719' => 38681,
    '971D' => 38685,'971F' => 38687,
    '9720' => 38688,'9722' => 38690,
    '9723' => 38691,'9724' => 38692,
    '9725' => 38693,'9726' => 38694,
    '9727' => 38695,'9728' => 38696,
    '9729' => 38697,'972B' => 38699,
    '972C' => 38700,'972E' => 38702,
    '972F' => 38703,'9735' => 38709,
    '973A' => 38714,'973D' => 38717,
    '973F' => 38719,'9742' => 38722,
    '9743' => 38723,'9744' => 38724,
    '9746' => 38726,'9747' => 38727,
    '9748' => 38728,'9749' => 38729,
    '974B' => 38731,'9758' => 38744,
    '975A' => 38746,'975C' => 38748,
    '9766' => 38758,'9768' => 38760,
    '976A' => 38762,'976C' => 38764,
    '976E' => 38766,'9770' => 38768,
    '9772' => 38770,'9777' => 38775,
    '9778' => 38776,'977A' => 38778,
    '977B' => 38779,'977D' => 38781,
    '977E' => 38782,'977F' => 38783,
    '9780' => 38784,'9781' => 38785,
    '9782' => 38786,'9783' => 38787,
    '9784' => 38788,'9788' => 38792,
    '978A' => 38794,'978E' => 38798,
    '978F' => 38799,'9797' => 38807,
    '9799' => 38809,'979A' => 38810,
    '979C' => 38812,'979D' => 38813,
    '979E' => 38814,'97A1' => 38817,
    '97A2' => 38818,'97A4' => 38820,
    '97A5' => 38821,'97A6' => 38822,
    '97A8' => 38824,'97AA' => 38826,
    '97AC' => 38828,'97AE' => 38830,
    '97B3' => 38835,'97B6' => 38838,
    '97B7' => 38839,'97B9' => 38841,
    '97BB' => 38843,'97BF' => 38847,
    '97C1' => 38849,'97C3' => 38851,
    '97C4' => 38852,'97C5' => 38853,
    '97C6' => 38854,'97C7' => 38855,
    '97C9' => 38857,'97CB' => 38859,
    '97CC' => 38860,'97CD' => 38861,
    '97CE' => 38862,'97CF' => 38863,
    '97D0' => 38864,'97D3' => 38867,
    '97D4' => 38868,'97D5' => 38869,
    '97D6' => 38870,'97D7' => 38871,
    '97D8' => 38872,'97D9' => 38873,
    '97DC' => 38876,'97DD' => 38877,
    '97DE' => 38878,'97DF' => 38879,
    '97E1' => 38881,'97E3' => 38883,
    '97E5' => 38885,'97F0' => 38896,
    '97F1' => 38897,'97F8' => 38904,
    '97F9' => 38905,'97FA' => 38906,
    '97FB' => 38907,'97FD' => 38909,
    '97FE' => 38910,'97FF' => 38911,
    '9800' => 38912,'9801' => 38913,
    '9802' => 38914,'9803' => 38915,
    '9804' => 38916,'9805' => 38917,
    '9806' => 38918,'9807' => 38919,
    '9808' => 38920,'980A' => 38922,
    '980C' => 38924,'980D' => 38925,
    '980E' => 38926,'980F' => 38927,
    '9810' => 38928,'9811' => 38929,
    '9812' => 38930,'9813' => 38931,
    '9816' => 38934,'9817' => 38935,
    '9818' => 38936,'981B' => 38939,
    '981C' => 38940,'981D' => 38941,
    '981E' => 38942,'9820' => 38944,
    '9821' => 38945,'9824' => 38948,
    '9826' => 38950,'9827' => 38951,
    '9828' => 38952,'9829' => 38953,
    '982B' => 38955,'982D' => 38957,
    '982F' => 38959,'9830' => 38960,
    '9832' => 38962,'9835' => 38965,
    '9837' => 38967,'9838' => 38968,
    '9839' => 38969,'983B' => 38971,
    '9841' => 38977,'9843' => 38979,
    '9844' => 38980,'9845' => 38981,
    '9846' => 38982,'9848' => 38984,
    '9849' => 38985,'984A' => 38986,
    '984C' => 38988,'984D' => 38989,
    '984E' => 38990,'984F' => 38991,
    '9850' => 38992,'9851' => 38993,
    '9852' => 38994,'9853' => 38995,
    '9857' => 38999,'9858' => 39000,
    '9859' => 39001,'985B' => 39003,
    '985C' => 39004,'985D' => 39005,
    '985E' => 39006,'985F' => 39007,
    '9860' => 39008,'9862' => 39010,
    '9863' => 39011,'9864' => 39012,
    '9865' => 39013,'9867' => 39015,
    '9869' => 39017,'986A' => 39018,
    '986B' => 39019,'986F' => 39023,
    '9870' => 39024,'9871' => 39025,
    '9872' => 39026,'9873' => 39027,
    '9874' => 39028,'98A8' => 39080,
    '98A9' => 39081,'98AC' => 39084,
    '98AD' => 39085,'98AE' => 39086,
    '98AF' => 39087,'98B1' => 39089,
    '98B2' => 39090,'98B3' => 39091,
    '98B6' => 39094,'98B8' => 39096,
    '98BA' => 39098,'98BB' => 39099,
    '98BC' => 39100,'98BD' => 39101,
    '98BE' => 39102,'98BF' => 39103,
    '98C0' => 39104,'98C1' => 39105,
    '98C2' => 39106,'98C4' => 39108,
    '98C6' => 39110,'98C9' => 39113,
    '98CB' => 39115,'98CC' => 39116,
    '98DB' => 39131,'98E2' => 39138,
    '98E3' => 39139,'98E5' => 39141,
    '98E9' => 39145,'98EA' => 39146,
    '98EB' => 39147,'98ED' => 39149,
    '98EF' => 39151,'98F2' => 39154,
    '98F4' => 39156,'98F6' => 39158,
    '98F9' => 39161,'98FA' => 39162,
    '98FC' => 39164,'98FD' => 39165,
    '98FE' => 39166,'9900' => 39168,
    '9902' => 39170,'9903' => 39171,
    '9905' => 39173,'9907' => 39175,
    '9908' => 39176,'9909' => 39177,
    '990A' => 39178,'990C' => 39180,
    '9911' => 39185,'9912' => 39186,
    '9913' => 39187,'9914' => 39188,
    '9915' => 39189,'9916' => 39190,
    '9917' => 39191,'9918' => 39192,
    '991A' => 39194,'991B' => 39195,
    '991E' => 39198,'991F' => 39199,
    '9921' => 39201,'9924' => 39204,
    '9925' => 39205,'9927' => 39207,
    '9928' => 39208,'9929' => 39209,
    '992A' => 39210,'992B' => 39211,
    '992C' => 39212,'992D' => 39213,
    '992F' => 39215,'9930' => 39216,
    '9931' => 39217,'9932' => 39218,
    '9933' => 39219,'9935' => 39221,
    '993A' => 39226,'993C' => 39228,
    '993D' => 39229,'993E' => 39230,
    '993F' => 39231,'9941' => 39233,
    '9943' => 39235,'9945' => 39237,
    '9947' => 39239,'9948' => 39240,
    '9949' => 39241,'994B' => 39243,
    '994C' => 39244,'994E' => 39246,
    '9950' => 39248,'9951' => 39249,
    '9952' => 39250,'9953' => 39251,
    '9956' => 39254,'9957' => 39255,
    '9958' => 39256,'9959' => 39257,
    '995B' => 39259,'995C' => 39260,
    '995E' => 39262,'995F' => 39263,
    '9961' => 39265,'999C' => 39324,
    '999D' => 39325,'999E' => 39326,
    '99A1' => 39329,'99A3' => 39331,
    '99A6' => 39334,'99A7' => 39335,
    '99AB' => 39339,'99AC' => 39340,
    '99AD' => 39341,'99AE' => 39342,
    '99AF' => 39343,'99B0' => 39344,
    '99B1' => 39345,'99B2' => 39346,
    '99B3' => 39347,'99B4' => 39348,
    '99B5' => 39349,'99B9' => 39353,
    '99BA' => 39354,'99BB' => 39355,
    '99BD' => 39357,'99C1' => 39361,
    '99C2' => 39362,'99C3' => 39363,
    '99C7' => 39367,'99C9' => 39369,
    '99CB' => 39371,'99CC' => 39372,
    '99CD' => 39373,'99CE' => 39374,
    '99CF' => 39375,'99D0' => 39376,
    '99D1' => 39377,'99D2' => 39378,
    '99D3' => 39379,'99D4' => 39380,
    '99D5' => 39381,'99D6' => 39382,
    '99D7' => 39383,'99D8' => 39384,
    '99D9' => 39385,'99DB' => 39387,
    '99DC' => 39388,'99DD' => 39389,
    '99DF' => 39391,'99E2' => 39394,
    '99E3' => 39395,'99E4' => 39396,
    '99E5' => 39397,'99E7' => 39399,
    '99E9' => 39401,'99EA' => 39402,
    '99EC' => 39404,'99ED' => 39405,
    '99EE' => 39406,'99F0' => 39408,
    '99F1' => 39409,'99F4' => 39412,
    '99F6' => 39414,'99F7' => 39415,
    '99F8' => 39416,'99F9' => 39417,
    '99FA' => 39418,'99FB' => 39419,
    '99FC' => 39420,'99FD' => 39421,
    '99FE' => 39422,'99FF' => 39423,
    '9A01' => 39425,'9A02' => 39426,
    '9A03' => 39427,'9A04' => 39428,
    '9A05' => 39429,'9A06' => 39430,
    '9A07' => 39431,'9A09' => 39433,
    '9A0A' => 39434,'9A0B' => 39435,
    '9A0D' => 39437,'9A0E' => 39438,
    '9A0F' => 39439,'9A11' => 39441,
    '9A14' => 39444,'9A15' => 39445,
    '9A16' => 39446,'9A19' => 39449,
    '9A1A' => 39450,'9A1B' => 39451,
    '9A1C' => 39452,'9A1D' => 39453,
    '9A1E' => 39454,'9A20' => 39456,
    '9A22' => 39458,'9A23' => 39459,
    '9A24' => 39460,'9A25' => 39461,
    '9A27' => 39463,'9A29' => 39465,
    '9A2A' => 39466,'9A2B' => 39467,
    '9A2C' => 39468,'9A2D' => 39469,
    '9A2E' => 39470,'9A30' => 39472,
    '9A31' => 39473,'9A32' => 39474,
    '9A34' => 39476,'9A35' => 39477,
    '9A36' => 39478,'9A37' => 39479,
    '9A38' => 39480,'9A39' => 39481,
    '9A3A' => 39482,'9A3D' => 39485,
    '9A3E' => 39486,'9A3F' => 39487,
    '9A40' => 39488,'9A41' => 39489,
    '9A42' => 39490,'9A43' => 39491,
    '9A44' => 39492,'9A45' => 39493,
    '9A46' => 39494,'9A48' => 39496,
    '9A49' => 39497,'9A4A' => 39498,
    '9A4C' => 39500,'9A4D' => 39501,
    '9A4E' => 39502,'9A4F' => 39503,
    '9A50' => 39504,'9A52' => 39506,
    '9A53' => 39507,'9A54' => 39508,
    '9A55' => 39509,'9A56' => 39510,
    '9A57' => 39511,'9A59' => 39513,
    '9A5A' => 39514,'9A5B' => 39515,
    '9A5E' => 39518,'9A5F' => 39519,
    '9A60' => 39520,'9A62' => 39522,
    '9A64' => 39524,'9A65' => 39525,
    '9A66' => 39526,'9A67' => 39527,
    '9A68' => 39528,'9A69' => 39529,
    '9A6A' => 39530,'9A6B' => 39531,
    '9AAB' => 39595,'9AAD' => 39597,
    '9AAF' => 39599,'9AB3' => 39603,
    '9AB4' => 39604,'9AB9' => 39609,
    '9ABB' => 39611,'9ABE' => 39614,
    '9ABF' => 39615,'9AC6' => 39622,
    '9AC7' => 39623,'9ACA' => 39626,
    '9ACD' => 39629,'9ACF' => 39631,
    '9AD0' => 39632,'9AD2' => 39634,
    '9AD4' => 39636,'9AD5' => 39637,
    '9AD6' => 39638,'9ADC' => 39644,
    '9AE3' => 39651,'9AE7' => 39655,
    '9AEC' => 39660,'9AEE' => 39662,
    '9AF1' => 39665,'9AF2' => 39666,
    '9AF3' => 39667,'9AF6' => 39670,
    '9AF7' => 39671,'9AFA' => 39674,
    '9AFC' => 39676,'9AFD' => 39677,
    '9AFE' => 39678,'9B01' => 39681,
    '9B04' => 39684,'9B05' => 39685,
    '9B06' => 39686,'9B0A' => 39690,
    '9B0B' => 39691,'9B0C' => 39692,
    '9B0D' => 39693,'9B0E' => 39694,
    '9B10' => 39696,'9B11' => 39697,
    '9B12' => 39698,'9B15' => 39701,
    '9B16' => 39702,'9B17' => 39703,
    '9B18' => 39704,'9B19' => 39705,
    '9B1A' => 39706,'9B1E' => 39710,
    '9B20' => 39712,'9B22' => 39714,
    '9B24' => 39716,'9B25' => 39717,
    '9B27' => 39719,'9B28' => 39720,
    '9B29' => 39721,'9B2B' => 39723,
    '9B2E' => 39726,'9B31' => 39729,
    '9B33' => 39731,'9B35' => 39733,
    '9B37' => 39735,'9B3A' => 39738,
    '9B3E' => 39742,'9B3F' => 39743,
    '9B46' => 39750,'9B4A' => 39754,
    '9B4B' => 39755,'9B4C' => 39756,
    '9B4E' => 39758,'9B52' => 39762,
    '9B55' => 39765,'9B56' => 39766,
    '9B58' => 39768,'9B59' => 39769,
    '9B5A' => 39770,'9B5B' => 39771,
    '9B5F' => 39775,'9B60' => 39776,
    '9B61' => 39777,'9B64' => 39780,
    '9B66' => 39782,'9B67' => 39783,
    '9B68' => 39784,'9B6C' => 39788,
    '9B6F' => 39791,'9B70' => 39792,
    '9B71' => 39793,'9B74' => 39796,
    '9B75' => 39797,'9B76' => 39798,
    '9B77' => 39799,'9B7A' => 39802,
    '9B7B' => 39803,'9B7C' => 39804,
    '9B7D' => 39805,'9B7E' => 39806,
    '9B80' => 39808,'9B82' => 39810,
    '9B85' => 39813,'9B86' => 39814,
    '9B87' => 39815,'9B88' => 39816,
    '9B90' => 39824,'9B91' => 39825,
    '9B92' => 39826,'9B93' => 39827,
    '9B95' => 39829,'9B9A' => 39834,
    '9B9B' => 39835,'9B9E' => 39838,
    '9BA0' => 39840,'9BA1' => 39841,
    '9BA2' => 39842,'9BA4' => 39844,
    '9BA5' => 39845,'9BA6' => 39846,
    '9BA8' => 39848,'9BAA' => 39850,
    '9BAB' => 39851,'9BAD' => 39853,
    '9BAE' => 39854,'9BAF' => 39855,
    '9BB5' => 39861,'9BB6' => 39862,
    '9BB8' => 39864,'9BB9' => 39865,
    '9BBD' => 39869,'9BBF' => 39871,
    '9BC0' => 39872,'9BC1' => 39873,
    '9BC3' => 39875,'9BC4' => 39876,
    '9BC6' => 39878,'9BC7' => 39879,
    '9BC8' => 39880,'9BC9' => 39881,
    '9BCA' => 39882,'9BD3' => 39891,
    '9BD4' => 39892,'9BD5' => 39893,
    '9BD6' => 39894,'9BD7' => 39895,
    '9BD9' => 39897,'9BDA' => 39898,
    '9BDB' => 39899,'9BDC' => 39900,
    '9BDE' => 39902,'9BE0' => 39904,
    '9BE1' => 39905,'9BE2' => 39906,
    '9BE4' => 39908,'9BE5' => 39909,
    '9BE6' => 39910,'9BE7' => 39911,
    '9BE8' => 39912,'9BEA' => 39914,
    '9BEB' => 39915,'9BEC' => 39916,
    '9BF0' => 39920,'9BF7' => 39927,
    '9BF8' => 39928,'9BFD' => 39933,
    '9C05' => 39941,'9C06' => 39942,
    '9C07' => 39943,'9C08' => 39944,
    '9C09' => 39945,'9C0B' => 39947,
    '9C0D' => 39949,'9C0E' => 39950,
    '9C12' => 39954,'9C13' => 39955,
    '9C14' => 39956,'9C17' => 39959,
    '9C1C' => 39964,'9C1D' => 39965,
    '9C21' => 39969,'9C23' => 39971,
    '9C24' => 39972,'9C25' => 39973,
    '9C28' => 39976,'9C29' => 39977,
    '9C2B' => 39979,'9C2C' => 39980,
    '9C2D' => 39981,'9C31' => 39985,
    '9C32' => 39986,'9C33' => 39987,
    '9C34' => 39988,'9C36' => 39990,
    '9C37' => 39991,'9C39' => 39993,
    '9C3B' => 39995,'9C3C' => 39996,
    '9C3D' => 39997,'9C3E' => 39998,
    '9C3F' => 39999,'9C40' => 40000,
    '9C41' => 40001,'9C44' => 40004,
    '9C46' => 40006,'9C48' => 40008,
    '9C49' => 40009,'9C4A' => 40010,
    '9C4B' => 40011,'9C4C' => 40012,
    '9C4D' => 40013,'9C4E' => 40014,
    '9C50' => 40016,'9C52' => 40018,
    '9C54' => 40020,'9C55' => 40021,
    '9C56' => 40022,'9C57' => 40023,
    '9C58' => 40024,'9C59' => 40025,
    '9C5E' => 40030,'9C5F' => 40031,
    '9C60' => 40032,'9C62' => 40034,
    '9C63' => 40035,'9C66' => 40038,
    '9C67' => 40039,'9C68' => 40040,
    '9C6D' => 40045,'9C6E' => 40046,
    '9C71' => 40049,'9C73' => 40051,
    '9C74' => 40052,'9C75' => 40053,
    '9C77' => 40055,'9C78' => 40056,
    '9C79' => 40057,'9C7A' => 40058,
    '9CE5' => 40165,'9CE6' => 40166,
    '9CE7' => 40167,'9CE9' => 40169,
    '9CEA' => 40170,'9CED' => 40173,
    '9CF1' => 40177,'9CF2' => 40178,
    '9CF3' => 40179,'9CF4' => 40180,
    '9CF5' => 40181,'9CF6' => 40182,
    '9CF7' => 40183,'9CF9' => 40185,
    '9CFA' => 40186,'9CFB' => 40187,
    '9CFC' => 40188,'9CFD' => 40189,
    '9CFF' => 40191,'9D00' => 40192,
    '9D03' => 40195,'9D04' => 40196,
    '9D05' => 40197,'9D06' => 40198,
    '9D07' => 40199,'9D08' => 40200,
    '9D09' => 40201,'9D10' => 40208,
    '9D12' => 40210,'9D14' => 40212,
    '9D15' => 40213,'9D17' => 40215,
    '9D18' => 40216,'9D19' => 40217,
    '9D1B' => 40219,'9D1D' => 40221,
    '9D1E' => 40222,'9D1F' => 40223,
    '9D20' => 40224,'9D22' => 40226,
    '9D23' => 40227,'9D25' => 40229,
    '9D26' => 40230,'9D28' => 40232,
    '9D29' => 40233,'9D2D' => 40237,
    '9D2E' => 40238,'9D2F' => 40239,
    '9D30' => 40240,'9D31' => 40241,
    '9D33' => 40243,'9D36' => 40246,
    '9D37' => 40247,'9D38' => 40248,
    '9D3B' => 40251,'9D3D' => 40253,
    '9D3E' => 40254,'9D3F' => 40255,
    '9D40' => 40256,'9D41' => 40257,
    '9D42' => 40258,'9D43' => 40259,
    '9D45' => 40261,'9D4A' => 40266,
    '9D4B' => 40267,'9D4C' => 40268,
    '9D4F' => 40271,'9D51' => 40273,
    '9D52' => 40274,'9D53' => 40275,
    '9D54' => 40276,'9D56' => 40278,
    '9D57' => 40279,'9D58' => 40280,
    '9D59' => 40281,'9D5A' => 40282,
    '9D5B' => 40283,'9D5C' => 40284,
    '9D5D' => 40285,'9D5F' => 40287,
    '9D60' => 40288,'9D61' => 40289,
    '9D67' => 40295,'9D68' => 40296,
    '9D69' => 40297,'9D6A' => 40298,
    '9D6B' => 40299,'9D6C' => 40300,
    '9D6F' => 40303,'9D70' => 40304,
    '9D71' => 40305,'9D72' => 40306,
    '9D73' => 40307,'9D74' => 40308,
    '9D75' => 40309,'9D77' => 40311,
    '9D78' => 40312,'9D79' => 40313,
    '9D7B' => 40315,'9D7D' => 40317,
    '9D7F' => 40319,'9D80' => 40320,
    '9D81' => 40321,'9D82' => 40322,
    '9D84' => 40324,'9D85' => 40325,
    '9D86' => 40326,'9D87' => 40327,
    '9D88' => 40328,'9D89' => 40329,
    '9D8A' => 40330,'9D8B' => 40331,
    '9D8C' => 40332,'9D90' => 40336,
    '9D92' => 40338,'9D94' => 40340,
    '9D96' => 40342,'9D97' => 40343,
    '9D98' => 40344,'9D99' => 40345,
    '9D9A' => 40346,'9D9B' => 40347,
    '9D9C' => 40348,'9D9D' => 40349,
    '9D9E' => 40350,'9D9F' => 40351,
    '9DA0' => 40352,'9DA1' => 40353,
    '9DA2' => 40354,'9DA3' => 40355,
    '9DA4' => 40356,'9DA6' => 40358,
    '9DA7' => 40359,'9DA8' => 40360,
    '9DA9' => 40361,'9DAA' => 40362,
    '9DAC' => 40364,'9DAD' => 40365,
    '9DAF' => 40367,'9DB1' => 40369,
    '9DB2' => 40370,'9DB3' => 40371,
    '9DB4' => 40372,'9DB5' => 40373,
    '9DB6' => 40374,'9DB7' => 40375,
    '9DB8' => 40376,'9DB9' => 40377,
    '9DBA' => 40378,'9DBB' => 40379,
    '9DBC' => 40380,'9DBE' => 40382,
    '9DBF' => 40383,'9DC1' => 40385,
    '9DC2' => 40386,'9DC3' => 40387,
    '9DC5' => 40389,'9DC7' => 40391,
    '9DC8' => 40392,'9DCA' => 40394,
    '9DCB' => 40395,'9DCC' => 40396,
    '9DCD' => 40397,'9DCE' => 40398,
    '9DCF' => 40399,'9DD0' => 40400,
    '9DD1' => 40401,'9DD2' => 40402,
    '9DD3' => 40403,'9DD5' => 40405,
    '9DD6' => 40406,'9DD7' => 40407,
    '9DD8' => 40408,'9DD9' => 40409,
    '9DDA' => 40410,'9DDB' => 40411,
    '9DDC' => 40412,'9DDD' => 40413,
    '9DDE' => 40414,'9DDF' => 40415,
    '9DE1' => 40417,'9DE2' => 40418,
    '9DE3' => 40419,'9DE4' => 40420,
    '9DE5' => 40421,'9DE6' => 40422,
    '9DE8' => 40424,'9DE9' => 40425,
    '9DEB' => 40427,'9DEC' => 40428,
    '9DED' => 40429,'9DEE' => 40430,
    '9DEF' => 40431,'9DF0' => 40432,
    '9DF2' => 40434,'9DF3' => 40435,
    '9DF4' => 40436,'9DF5' => 40437,
    '9DF6' => 40438,'9DF7' => 40439,
    '9DF8' => 40440,'9DF9' => 40441,
    '9DFA' => 40442,'9DFB' => 40443,
    '9DFD' => 40445,'9DFE' => 40446,
    '9DFF' => 40447,'9E00' => 40448,
    '9E01' => 40449,'9E02' => 40450,
    '9E03' => 40451,'9E04' => 40452,
    '9E05' => 40453,'9E06' => 40454,
    '9E07' => 40455,'9E09' => 40457,
    '9E0B' => 40459,'9E0D' => 40461,
    '9E0F' => 40463,'9E10' => 40464,
    '9E11' => 40465,'9E12' => 40466,
    '9E13' => 40467,'9E14' => 40468,
    '9E15' => 40469,'9E17' => 40471,
    '9E19' => 40473,'9E1A' => 40474,
    '9E1B' => 40475,'9E1D' => 40477,
    '9E1E' => 40478,'9E75' => 40565,
    '9E79' => 40569,'9E7A' => 40570,
    '9E7C' => 40572,'9E7D' => 40573,
    '9E80' => 40576,'9E83' => 40579,
    '9E86' => 40582,'9E89' => 40585,
    '9E8A' => 40586,'9E8C' => 40588,
    '9E8D' => 40589,'9E8E' => 40590,
    '9E91' => 40593,'9E94' => 40596,
    '9E97' => 40599,'9E99' => 40601,
    '9E9A' => 40602,'9E9B' => 40603,
    '9E9C' => 40604,'9EA0' => 40608,
    '9EA1' => 40609,'9EA4' => 40612,
    '9EA5' => 40613,'9EA7' => 40615,
    '9EA9' => 40617,'9EAD' => 40621,
    '9EAE' => 40622,'9EB0' => 40624,
    '9EB5' => 40629,'9EB6' => 40630,
    '9EB7' => 40631,'9EBC' => 40636,
    '9EC0' => 40640,'9EC2' => 40642,
    '9EC3' => 40643,'9EC8' => 40648,
    '9ECC' => 40652,'9ED0' => 40656,
    '9ED3' => 40659,'9ED5' => 40661,
    '9ED6' => 40662,'9EDA' => 40666,
    '9EDE' => 40670,'9EE4' => 40676,
    '9EE6' => 40678,'9EE8' => 40680,
    '9EEB' => 40683,'9EED' => 40685,
    '9EEE' => 40686,'9EF0' => 40688,
    '9EF2' => 40690,'9EF3' => 40691,
    '9EF4' => 40692,'9EF5' => 40693,
    '9EF6' => 40694,'9EF7' => 40695,
    '9EFA' => 40698,'9EFD' => 40701,
    '9EFF' => 40703,'9F00' => 40704,
    '9F01' => 40705,'9F06' => 40710,
    '9F07' => 40711,'9F09' => 40713,
    '9F0A' => 40714,'9F0F' => 40719,
    '9F12' => 40722,'9F15' => 40725,
    '9F16' => 40726,'9F18' => 40728,
    '9F1A' => 40730,'9F1B' => 40731,
    '9F1C' => 40732,'9F1E' => 40734,
    '9F23' => 40739,'9F24' => 40740,
    '9F25' => 40741,'9F28' => 40744,
    '9F29' => 40745,'9F2A' => 40746,
    '9F2B' => 40747,'9F2D' => 40749,
    '9F2E' => 40750,'9F30' => 40752,
    '9F31' => 40753,'9F32' => 40754,
    '9F33' => 40755,'9F34' => 40756,
    '9F35' => 40757,'9F36' => 40758,
    '9F38' => 40760,'9F40' => 40768,
    '9F41' => 40769,'9F42' => 40770,
    '9F43' => 40771,'9F46' => 40774,
    '9F47' => 40775,'9F48' => 40776,
    '9F49' => 40777,'9F4A' => 40778,
    '9F4B' => 40779,'9F4C' => 40780,
    '9F4D' => 40781,'9F4E' => 40782,
    '9F4F' => 40783,'9F52' => 40786,
    '9F54' => 40788,'9F55' => 40789,
    '9F56' => 40790,'9F57' => 40791,
    '9F58' => 40792,'9F59' => 40793,
    '9F5B' => 40795,'9F5C' => 40796,
    '9F5D' => 40797,'9F5E' => 40798,
    '9F5F' => 40799,'9F60' => 40800,
    '9F61' => 40801,'9F63' => 40803,
    '9F64' => 40804,'9F65' => 40805,
    '9F66' => 40806,'9F67' => 40807,
    '9F6A' => 40810,'9F6B' => 40811,
    '9F6C' => 40812,'9F6E' => 40814,
    '9F6F' => 40815,'9F70' => 40816,
    '9F71' => 40817,'9F72' => 40818,
    '9F74' => 40820,'9F75' => 40821,
    '9F76' => 40822,'9F77' => 40823,
    '9F78' => 40824,'9F79' => 40825,
    '9F7A' => 40826,'9F7B' => 40827,
    '9F7E' => 40830,'9F8D' => 40845,
    '9F90' => 40848,'9F91' => 40849,
    '9F92' => 40850,'9F94' => 40852,
    '9F95' => 40853,'9F98' => 40856,
    '9F9C' => 40860,'9FA2' => 40866,
    '9FA4' => 40868,'F6EE' => 63214,
    'F6EF' => 63215,'F7E5' => 63461,
    'F7EA' => 63466,'F7EB' => 63467,
    'F7ED' => 63469,'F7EF' => 63471,
    'F7F0' => 63472,'F7F1' => 63473,
    'F7F2' => 63474,'F7F3' => 63475,
    'F7F4' => 63476,'F7F5' => 63477,
    'F7F6' => 63478,'F7F7' => 63479,
    'F7F8' => 63480,'F7F9' => 63481,
    'F7FA' => 63482,'F7FB' => 63483,
    'F7FC' => 63484,'F7FD' => 63485,
    'F7FE' => 63486,'F7FF' => 63487,
    'F800' => 63488,'F801' => 63489,
    'F802' => 63490,'F803' => 63491,
    'F804' => 63492,'F805' => 63493,
    'F806' => 63494,'F807' => 63495,
    'F808' => 63496,'F809' => 63497,
    'F80A' => 63498,'F80B' => 63499,
    'F80C' => 63500,'F80D' => 63501,
    'F80E' => 63502,'F80F' => 63503,
    'F810' => 63504,'F811' => 63505,
    'F812' => 63506,'F813' => 63507,
    'F814' => 63508,'F815' => 63509,
    'F816' => 63510,'FA0C' => 64012,
    'FA0D' => 64013,
};

# simplified chinese characteristic code page
$UTF8_SCH = {
    '2015' => 8213,'2016' => 8214,
    '2030' => 8240,'2033' => 8243,
    '216A' => 8554,'216B' => 8555,
    '2208' => 8712,'220F' => 8719,
    '2211' => 8721,'221D' => 8733,
    '2227' => 8743,'2228' => 8744,
    '2236' => 8758,'2237' => 8759,
    '223D' => 8765,'2248' => 8776,
    '224C' => 8780,'2264' => 8804,
    '2265' => 8805,'226E' => 8814,
    '226F' => 8815,'2312' => 8978,
    '2505' => 9477,'2506' => 9478,
    '2507' => 9479,'2508' => 9480,
    '2509' => 9481,'250A' => 9482,
    '250B' => 9483,'250D' => 9485,
    '250E' => 9486,'250F' => 9487,
    '2511' => 9489,'2512' => 9490,
    '2513' => 9491,'2515' => 9493,
    '2516' => 9494,'2517' => 9495,
    '2519' => 9497,'251A' => 9498,
    '251B' => 9499,'251D' => 9501,
    '251E' => 9502,'251F' => 9503,
    '2520' => 9504,'2521' => 9505,
    '2522' => 9506,'2523' => 9507,
    '2525' => 9509,'2526' => 9510,
    '2527' => 9511,'2528' => 9512,
    '2529' => 9513,'252A' => 9514,
    '252B' => 9515,'252D' => 9517,
    '252E' => 9518,'252F' => 9519,
    '2530' => 9520,'2531' => 9521,
    '2532' => 9522,'2533' => 9523,
    '2535' => 9525,'2536' => 9526,
    '2537' => 9527,'2538' => 9528,
    '2539' => 9529,'253A' => 9530,
    '253B' => 9531,'253D' => 9533,
    '253E' => 9534,'253F' => 9535,
    '2540' => 9536,'2541' => 9537,
    '2542' => 9538,'2543' => 9539,
    '2544' => 9540,'2545' => 9541,
    '2546' => 9542,'2547' => 9543,
    '2548' => 9544,'2549' => 9545,
    '254A' => 9546,'254B' => 9547,
    '3013' => 12307,'3016' => 12310,
    '3017' => 12311,'30FB' => 12539,
    '4E13' => 19987,'4E1A' => 19994,
    '4E1B' => 19995,'4E1C' => 19996,
    '4E1D' => 19997,'4E22' => 20002,
    '4E24' => 20004,'4E25' => 20005,
    '4E27' => 20007,'4E28' => 20008,
    '4E2A' => 20010,'4E2C' => 20012,
    '4E34' => 20020,'4E36' => 20022,
    '4E3A' => 20026,'4E3D' => 20029,
    '4E3E' => 20030,'4E3F' => 20031,
    '4E49' => 20041,'4E4C' => 20044,
    '4E50' => 20048,'4E54' => 20052,
    '4E60' => 20064,'4E61' => 20065,
    '4E66' => 20070,'4E70' => 20080,
    '4E71' => 20081,'4E89' => 20105,
    '4E8F' => 20111,'4E98' => 20120,
    '4E9A' => 20122,'4EA0' => 20128,
    '4EA7' => 20135,'4EA9' => 20137,
    '4EB2' => 20146,'4EB5' => 20149,
    '4EBB' => 20155,'4EBF' => 20159,
    '4EC5' => 20165,'4ECE' => 20174,
    '4ED1' => 20177,'4ED3' => 20179,
    '4EEA' => 20202,'4EEB' => 20203,
    '4EEC' => 20204,'4F17' => 20247,
    '4F1A' => 20250,'4F1B' => 20251,
    '4F1E' => 20254,'4F1F' => 20255,
    '4F20' => 20256,'4F24' => 20260,
    '4F25' => 20261,'4F26' => 20262,
    '4F27' => 20263,'4F2A' => 20266,
    '4F2B' => 20267,'4F32' => 20274,
    '4F65' => 20325,'4FA0' => 20384,
    '4FA3' => 20387,'4FA5' => 20389,
    '4FA6' => 20390,'4FA7' => 20391,
    '4FA8' => 20392,'4FA9' => 20393,
    '4FAA' => 20394,'4FAC' => 20396,
    '4FE3' => 20451,'4FE6' => 20454,
    '4FE8' => 20456,'4FE9' => 20457,
    '4FEA' => 20458,'4FED' => 20461,
    '502E' => 20526,'503A' => 20538,
    '503E' => 20542,'506C' => 20588,
    '507B' => 20603,'507E' => 20606,
    '507F' => 20607,'5088' => 20616,
    '50A5' => 20645,'50A7' => 20647,
    '50A8' => 20648,'50A9' => 20649,
    '5151' => 20817,'5156' => 20822,
    '5170' => 20848,'5173' => 20851,
    '5174' => 20852,'5179' => 20857,
    '517B' => 20859,'517D' => 20861,
    '5181' => 20865,'5182' => 20866,
    '5185' => 20869,'518C' => 20876,
    '5196' => 20886,'5199' => 20889,
    '519B' => 20891,'519C' => 20892,
    '51AB' => 20907,'51AF' => 20911,
    '51B2' => 20914,'51B3' => 20915,
    '51B5' => 20917,'51BB' => 20923,
    '51C0' => 20928,'51C7' => 20935,
    '51C9' => 20937,'51CF' => 20943,
    '51D1' => 20945,'51DB' => 20955,
    '51E4' => 20964,'51EB' => 20971,
    '51ED' => 20973,'51EF' => 20975,
    '51FB' => 20987,'51FC' => 20988,
    '51FF' => 20991,'520D' => 21005,
    '5218' => 21016,'5219' => 21017,
    '521A' => 21018,'521B' => 21019,
    '5220' => 21024,'522B' => 21035,
    '522D' => 21037,'5239' => 21049,
    '523D' => 21053,'523F' => 21055,
    '5240' => 21056,'5242' => 21058,
    '5250' => 21072,'5251' => 21073,
    '5265' => 21093,'5267' => 21095,
    '5290' => 21136,'529D' => 21149,
    '529E' => 21150,'52A1' => 21153,
    '52A2' => 21154,'52A8' => 21160,
    '52B1' => 21169,'52B2' => 21170,
    '52B3' => 21171,'52BF' => 21183,
    '52CB' => 21195,'52D0' => 21200,
    '52F9' => 21241,'5300' => 21248,
    '5326' => 21286,'532E' => 21294,
    '533A' => 21306,'533B' => 21307,
    '534E' => 21326,'534F' => 21327,
    '5355' => 21333,'5356' => 21334,
    '535F' => 21343,'5362' => 21346,
    '5364' => 21348,'5367' => 21351,
    '5369' => 21353,'536B' => 21355,
    '5374' => 21364,'537A' => 21370,
    '5385' => 21381,'5386' => 21382,
    '5389' => 21385,'538B' => 21387,
    '538C' => 21388,'538D' => 21389,
    '5395' => 21397,'53A2' => 21410,
    '53A3' => 21411,'53A6' => 21414,
    '53A8' => 21416,'53A9' => 21417,
    '53AE' => 21422,'53B6' => 21430,
    '53BF' => 21439,'53C1' => 21441,
    '53C2' => 21442,'53CC' => 21452,
    '53D1' => 21457,'53D8' => 21464,
    '53D9' => 21465,'53E0' => 21472,
    '53F6' => 21494,'53F7' => 21495,
    '53F9' => 21497,'53FD' => 21501,
    '5413' => 21523,'5415' => 21525,
    '5416' => 21526,'5417' => 21527,
    '5421' => 21537,'5423' => 21539,
    '542F' => 21551,'5432' => 21554,
    '5434' => 21556,'544B' => 21579,
    '5450' => 21584,'5452' => 21586,
    '5453' => 21587,'5455' => 21589,
    '5456' => 21590,'5457' => 21591,
    '5458' => 21592,'5459' => 21593,
    '545B' => 21595,'545C' => 21596,
    '548F' => 21647,'5494' => 21652,
    '5499' => 21657,'549B' => 21659,
    '549D' => 21661,'54A3' => 21667,
    '54A4' => 21668,'54B4' => 21684,
    '54CC' => 21708,'54CD' => 21709,
    '54D0' => 21712,'54D1' => 21713,
    '54D2' => 21714,'54D3' => 21715,
    '54D4' => 21716,'54D5' => 21717,
    '54D7' => 21719,'54D9' => 21721,
    '54DA' => 21722,'54DC' => 21724,
    '54DD' => 21725,'54DF' => 21727,
    '551B' => 21787,'5520' => 21792,
    '5522' => 21794,'5523' => 21795,
    '5524' => 21796,'553F' => 21823,
    '5549' => 21833,'5567' => 21863,
    '556C' => 21868,'556D' => 21869,
    '556E' => 21870,'5578' => 21880,
    '55B7' => 21943,'55B9' => 21945,
    '55BD' => 21949,'55BE' => 21950,
    '55EA' => 21994,'55EB' => 21995,
    '55EC' => 21996,'55F3' => 22003,
    '55F5' => 22005,'5618' => 22040,
    '561E' => 22046,'5623' => 22051,
    '5624' => 22052,'562D' => 22061,
    '5631' => 22065,'564D' => 22093,
    '5654' => 22100,'565C' => 22108,
    '567B' => 22139,'567C' => 22140,
    '56A3' => 22179,'56AF' => 22191,
    '56E2' => 22242,'56ED' => 22253,
    '56F1' => 22257,'56F4' => 22260,
    '56F5' => 22261,'56FD' => 22269,
    '56FE' => 22270,'5706' => 22278,
    '5739' => 22329,'573A' => 22330,
    '5742' => 22338,'5757' => 22359,
    '575A' => 22362,'575B' => 22363,
    '575C' => 22364,'575D' => 22365,
    '575E' => 22366,'575F' => 22367,
    '5760' => 22368,'5784' => 22404,
    '5785' => 22405,'5786' => 22406,
    '5792' => 22418,'57A1' => 22433,
    '57A6' => 22438,'57A7' => 22439,
    '57A9' => 22441,'57AB' => 22443,
    '57AD' => 22445,'57B2' => 22450,
    '57B4' => 22452,'57D8' => 22488,
    '57D9' => 22489,'57DA' => 22490,
    '57DD' => 22493,'57EF' => 22511,
    '5811' => 22545,'5815' => 22549,
    '5844' => 22596,'586C' => 22636,
    '5892' => 22674,'5899' => 22681,
    '589A' => 22682,'58EE' => 22766,
    '58F0' => 22768,'58F3' => 22771,
    '58F6' => 22774,'5902' => 22786,
    '5904' => 22788,'5907' => 22791,
    '591F' => 22815,'5934' => 22836,
    '5939' => 22841,'593A' => 22842,
    '5941' => 22849,'5942' => 22850,
    '594B' => 22859,'5956' => 22870,
    '5965' => 22885,'5986' => 22918,
    '5987' => 22919,'5988' => 22920,
    '59A9' => 22953,'59AA' => 22954,
    '59AB' => 22955,'59D7' => 22999,
    '59F9' => 23033,'5A04' => 23044,
    '5A05' => 23045,'5A06' => 23046,
    '5A07' => 23047,'5A08' => 23048,
    '5A31' => 23089,'5A32' => 23090,
    '5A34' => 23092,'5A74' => 23156,
    '5A75' => 23157,'5A76' => 23158,
    '5AAA' => 23210,'5AD2' => 23250,
    '5AD4' => 23252,'5AF1' => 23281,
    '5B37' => 23351,'5B59' => 23385,
    '5B66' => 23398,'5B6A' => 23402,
    '5B80' => 23424,'5B9D' => 23453,
    '5B9E' => 23454,'5BA0' => 23456,
    '5BA1' => 23457,'5BAA' => 23466,
    '5BAB' => 23467,'5BBD' => 23485,
    '5BBE' => 23486,'5BDD' => 23517,
    '5BF9' => 23545,'5BFB' => 23547,
    '5BFC' => 23548,'5BFF' => 23551,
    '5C06' => 23558,'5C14' => 23572,
    '5C18' => 23576,'5C1C' => 23580,
    '5C1D' => 23581,'5C27' => 23591,
    '5C34' => 23604,'5C3D' => 23613,
    '5C42' => 23618,'5C49' => 23625,
    '5C4A' => 23626,'5C5E' => 23646,
    '5C61' => 23649,'5C66' => 23654,
    '5C7F' => 23679,'5C81' => 23681,
    '5C82' => 23682,'5C96' => 23702,
    '5C97' => 23703,'5C98' => 23704,
    '5C99' => 23705,'5C9A' => 23706,
    '5C9B' => 23707,'5C9C' => 23708,
    '5CBD' => 23741,'5CBF' => 23743,
    '5CC1' => 23745,'5CC4' => 23748,
    '5CE1' => 23777,'5CE4' => 23780,
    '5CE5' => 23781,'5CE6' => 23782,
    '5D02' => 23810,'5D03' => 23811,
    '5D2D' => 23853,'5D3E' => 23870,
    '5D58' => 23896,'5D5B' => 23899,
    '5D5D' => 23901,'5D74' => 23924,
    '5DC5' => 24005,'5DDB' => 24027,
    '5DE9' => 24041,'5DEF' => 24047,
    '5E01' => 24065,'5E05' => 24069,
    '5E08' => 24072,'5E0F' => 24079,
    '5E10' => 24080,'5E1C' => 24092,
    '5E26' => 24102,'5E27' => 24103,
    '5E2E' => 24110,'5E31' => 24113,
    '5E3B' => 24123,'5E3C' => 24124,
    '5E42' => 24130,'5E5E' => 24158,
    '5E7A' => 24186,'5E7F' => 24191,
    '5E86' => 24198,'5E90' => 24208,
    '5E91' => 24209,'5E93' => 24211,
    '5E94' => 24212,'5E99' => 24217,
    '5E9E' => 24222,'5E9F' => 24223,
    '5EEA' => 24298,'5EF4' => 24308,
    '5F00' => 24320,'5F03' => 24323,
    '5F11' => 24337,'5F20' => 24352,
    '5F25' => 24357,'5F2A' => 24362,
    '5F2F' => 24367,'5F39' => 24377,
    '5F3A' => 24378,'5F50' => 24400,
    '5F52' => 24402,'5F53' => 24403,
    '5F55' => 24405,'5F5D' => 24413,
    '5F61' => 24417,'5F66' => 24422,
    '5F7B' => 24443,'5F84' => 24452,
    '5F95' => 24469,'5FC4' => 24516,
    '5FC6' => 24518,'5FE7' => 24551,
    '5FFE' => 24574,'6001' => 24577,
    '6002' => 24578,'6003' => 24579,
    '6004' => 24580,'6005' => 24581,
    '6006' => 24582,'603B' => 24635,
    '603C' => 24636,'603F' => 24639,
    '604B' => 24651,'6073' => 24691,
    '6076' => 24694,'6078' => 24696,
    '6079' => 24697,'607A' => 24698,
    '607B' => 24699,'607C' => 24700,
    '607D' => 24701,'60A6' => 24742,
    '60AB' => 24747,'60AC' => 24748,
    '60AD' => 24749,'60AF' => 24751,
    '60E7' => 24807,'60E8' => 24808,
    '60E9' => 24809,'60EB' => 24811,
    '60EC' => 24812,'60ED' => 24813,
    '60EE' => 24814,'60EF' => 24815,
    '6120' => 24864,'6124' => 24868,
    '6126' => 24870,'6151' => 24913,
    '61B7' => 25015,'61D1' => 25041,
    '61D2' => 25042,'61D4' => 25044,
    '6206' => 25094,'620B' => 25099,
    '620F' => 25103,'6217' => 25111,
    '6218' => 25112,'622C' => 25132,
    '6237' => 25143,'624C' => 25164,
    '6267' => 25191,'6269' => 25193,
    '626A' => 25194,'626B' => 25195,
    '626C' => 25196,'629A' => 25242,
    '629B' => 25243,'629F' => 25247,
    '62A0' => 25248,'62A1' => 25249,
    '62A2' => 25250,'62A4' => 25252,
    '62A5' => 25253,'62C5' => 25285,
    '62DF' => 25311,'62E2' => 25314,
    '62E3' => 25315,'62E5' => 25317,
    '62E6' => 25318,'62E7' => 25319,
    '62E8' => 25320,'62E9' => 25321,
    '631A' => 25370,'631B' => 25371,
    '631D' => 25373,'631E' => 25374,
    '631F' => 25375,'6320' => 25376,
    '6321' => 25377,'6322' => 25378,
    '6323' => 25379,'6324' => 25380,
    '6325' => 25381,'635E' => 25438,
    '635F' => 25439,'6361' => 25441,
    '6362' => 25442,'6363' => 25443,
    '63B3' => 25523,'63B4' => 25524,
    '63B7' => 25527,'63B8' => 25528,
    '63BA' => 25530,'63BC' => 25532,
    '63DE' => 25566,'63F8' => 25592,
    '63FD' => 25597,'63FF' => 25599,
    '6400' => 25600,'6401' => 25601,
    '6402' => 25602,'6405' => 25605,
    '643A' => 25658,'6444' => 25668,
    '6445' => 25669,'6446' => 25670,
    '6447' => 25671,'6448' => 25672,
    '644A' => 25674,'6484' => 25732,
    '6491' => 25745,'64B5' => 25781,
    '64B7' => 25783,'64B8' => 25784,
    '64BA' => 25786,'64C0' => 25792,
    '64DE' => 25822,'6512' => 25874,
    '6534' => 25908,'6535' => 25909,
    '654C' => 25932,'655B' => 25947,
    '656B' => 25963,'6570' => 25968,
    '658B' => 25995,'6593' => 26003,
    '65A9' => 26025,'65AD' => 26029,
    '65E0' => 26080,'65E7' => 26087,
    '65F6' => 26102,'65F7' => 26103,
    '6619' => 26137,'663C' => 26172,
    '663E' => 26174,'664B' => 26187,
    '6653' => 26195,'6654' => 26196,
    '6655' => 26197,'6656' => 26198,
    '6657' => 26199,'6682' => 26242,
    '66A7' => 26279,'672F' => 26415,
    '6740' => 26432,'6742' => 26434,
    '6743' => 26435,'6761' => 26465,
    '6765' => 26469,'6768' => 26472,
    '6769' => 26473,'679E' => 26526,
    '67A2' => 26530,'67A3' => 26531,
    '67A5' => 26533,'67A7' => 26535,
    '67A8' => 26536,'67AA' => 26538,
    '67AB' => 26539,'67AD' => 26541,
    '67E0' => 26592,'67FD' => 26621,
    '6800' => 26624,'6805' => 26629,
    '6807' => 26631,'6808' => 26632,
    '6809' => 26633,'680A' => 26634,
    '680B' => 26635,'680C' => 26636,
    '680E' => 26638,'680F' => 26639,
    '6811' => 26641,'6837' => 26679,
    '683E' => 26686,'684A' => 26698,
    '6855' => 26709,'6860' => 26720,
    '6861' => 26721,'6862' => 26722,
    '6863' => 26723,'6864' => 26724,
    '6865' => 26725,'6866' => 26726,
    '6867' => 26727,'6868' => 26728,
    '6869' => 26729,'68A6' => 26790,
    '68C0' => 26816,'68C2' => 26818,
    '6901' => 26881,'691F' => 26911,
    '6920' => 26912,'6924' => 26916,
    '692D' => 26925,'697C' => 27004,
    '6980' => 27008,'6984' => 27012,
    '6987' => 27015,'6988' => 27016,
    '6989' => 27017,'6998' => 27032,
    '69DB' => 27099,'69DF' => 27103,
    '69E0' => 27104,'6A2A' => 27178,
    '6A2F' => 27183,'6A31' => 27185,
    '6A65' => 27237,'6A71' => 27249,
    '6A79' => 27257,'6A7C' => 27260,
    '6AA9' => 27305,'6AAB' => 27307,
    '6B22' => 27426,'6B24' => 27428,
    '6B27' => 27431,'6B7C' => 27516,
    '6B81' => 27521,'6B87' => 27527,
    '6B8B' => 27531,'6B92' => 27538,
    '6B93' => 27539,'6B9A' => 27546,
    '6BA1' => 27553,'6BB4' => 27572,
    '6BC1' => 27585,'6BC2' => 27586,
    '6BD5' => 27605,'6BD9' => 27609,
    '6BE1' => 27617,'6BEA' => 27626,
    '6BF5' => 27637,'6C07' => 27655,
    '6C22' => 27682,'6C29' => 27689,
    '6C32' => 27698,'6C35' => 27701,
    '6C3D' => 27709,'6C47' => 27719,
    '6C49' => 27721,'6C64' => 27748,
    '6C79' => 27769,'6C9F' => 27807,
    '6CA1' => 27809,'6CA3' => 27811,
    '6CA4' => 27812,'6CA5' => 27813,
    '6CA6' => 27814,'6CA7' => 27815,
    '6CA9' => 27817,'6CAA' => 27818,
    '6CB2' => 27826,'6CEA' => 27882,
    '6CF6' => 27894,'6CF7' => 27895,
    '6CF8' => 27896,'6CFA' => 27898,
    '6CFB' => 27899,'6CFC' => 27900,
    '6CFD' => 27901,'6CFE' => 27902,
    '6D43' => 27971,'6D45' => 27973,
    '6D46' => 27974,'6D47' => 27975,
    '6D48' => 27976,'6D4A' => 27978,
    '6D4B' => 27979,'6D4D' => 27981,
    '6D4E' => 27982,'6D4F' => 27983,
    '6D51' => 27985,'6D52' => 27986,
    '6D53' => 27987,'6D54' => 27988,
    '6D5C' => 27996,'6D9B' => 28059,
    '6D9D' => 28061,'6D9E' => 28062,
    '6D9F' => 28063,'6DA0' => 28064,
    '6DA1' => 28065,'6DA3' => 28067,
    '6DA4' => 28068,'6DA6' => 28070,
    '6DA7' => 28071,'6DA8' => 28072,
    '6DA9' => 28073,'6E0A' => 28170,
    '6E0C' => 28172,'6E0D' => 28173,
    '6E0E' => 28174,'6E10' => 28176,
    '6E11' => 28177,'6E14' => 28180,
    '6E16' => 28182,'6E17' => 28183,
    '6E29' => 28201,'6E7E' => 28286,
    '6E7F' => 28287,'6E83' => 28291,
    '6E85' => 28293,'6E86' => 28294,
    '6EBB' => 28347,'6ED7' => 28375,
    '6EDA' => 28378,'6EDE' => 28382,
    '6EDF' => 28383,'6EE0' => 28384,
    '6EE1' => 28385,'6EE2' => 28386,
    '6EE4' => 28388,'6EE5' => 28389,
    '6EE6' => 28390,'6EE8' => 28392,
    '6EE9' => 28393,'6F24' => 28452,
    '6F46' => 28486,'6F47' => 28487,
    '6F4B' => 28491,'6F4D' => 28493,
    '6F5C' => 28508,'6F74' => 28532,
    '6F9C' => 28572,'6FD1' => 28625,
    '6FD2' => 28626,'704F' => 28751,
    '706C' => 28780,'706D' => 28781,
    '706F' => 28783,'7075' => 28789,
    '707E' => 28798,'707F' => 28799,
    '7080' => 28800,'7089' => 28809,
    '709C' => 28828,'709D' => 28829,
    '70B9' => 28857,'70BB' => 28859,
    '70BC' => 28860,'70BD' => 28861,
    '70C0' => 28864,'70C1' => 28865,
    '70C2' => 28866,'70C3' => 28867,
    '70DB' => 28891,'70DF' => 28895,
    '70E6' => 28902,'70E7' => 28903,
    '70E8' => 28904,'70E9' => 28905,
    '70EB' => 28907,'70EC' => 28908,
    '70ED' => 28909,'7115' => 28949,
    '7116' => 28950,'7118' => 28952,
    '7145' => 28997,'714A' => 29002,
    '7173' => 29043,'717A' => 29050,
    '7198' => 29080,'7231' => 29233,
    '7237' => 29239,'724D' => 29261,
    '7266' => 29286,'7275' => 29301,
    '727A' => 29306,'728A' => 29322,
    '728F' => 29327,'729F' => 29343,
    '72AD' => 29357,'72B6' => 29366,
    '72B7' => 29367,'72B8' => 29368,
    '72B9' => 29369,'72C8' => 29384,
    '72CD' => 29389,'72DE' => 29406,
    '72EC' => 29420,'72ED' => 29421,
    '72EE' => 29422,'72EF' => 29423,
    '72F0' => 29424,'72F1' => 29425,
    '72F2' => 29426,'7303' => 29443,
    '730E' => 29454,'7315' => 29461,
    '7321' => 29473,'732A' => 29482,
    '732B' => 29483,'732C' => 29484,
    '732E' => 29486,'7338' => 29496,
    '7339' => 29497,'736D' => 29549,
    '7391' => 29585,'739B' => 29595,
    '73AE' => 29614,'73AF' => 29615,
    '73B0' => 29616,'73BA' => 29626,
    '73C9' => 29641,'73CF' => 29647,
    '73D0' => 29648,'73D1' => 29649,
    '73F2' => 29682,'740F' => 29711,
    '7410' => 29712,'743C' => 29756,
    '7476' => 29814,'7477' => 29815,
    '748E' => 29838,'74D2' => 29906,
    '74EF' => 29935,'7519' => 29977,
    '7535' => 30005,'753B' => 30011,
    '7545' => 30021,'7572' => 30066,
    '7574' => 30068,'7583' => 30083,
    '7592' => 30098,'7596' => 30102,
    '7597' => 30103,'759F' => 30111,
    '75A0' => 30112,'75A1' => 30113,
    '75AC' => 30124,'75AE' => 30126,
    '75AF' => 30127,'75B1' => 30129,
    '75B4' => 30132,'75C3' => 30147,
    '75C8' => 30152,'75C9' => 30153,
    '75D6' => 30166,'75E8' => 30184,
    '75EA' => 30186,'75EB' => 30187,
    '7605' => 30213,'7617' => 30231,
    '7618' => 30232,'762A' => 30250,
    '762B' => 30251,'763E' => 30270,
    '763F' => 30271,'7640' => 30272,
    '764D' => 30285,'7654' => 30292,
    '765E' => 30302,'7663' => 30307,
    '766B' => 30315,'766F' => 30319,
    '7691' => 30353,'76B1' => 30385,
    '76B2' => 30386,'76CF' => 30415,
    '76D0' => 30416,'76D1' => 30417,
    '76D6' => 30422,'76D7' => 30423,
    '76D8' => 30424,'770D' => 30477,
    '7726' => 30502,'7740' => 30528,
    '7741' => 30529,'7743' => 30531,
    '7750' => 30544,'7751' => 30545,
    '7792' => 30610,'77A9' => 30633,
    '77EB' => 30699,'77F6' => 30710,
    '77FE' => 30718,'77FF' => 30719,
    '7800' => 30720,'7801' => 30721,
    '7816' => 30742,'7817' => 30743,
    '7818' => 30744,'781A' => 30746,
    '781C' => 30748,'7839' => 30777,
    '783A' => 30778,'783B' => 30779,
    '783C' => 30780,'783E' => 30782,
    '7840' => 30784,'7847' => 30791,
    '7855' => 30805,'7856' => 30806,
    '7857' => 30807,'7877' => 30839,
    '788D' => 30861,'789B' => 30875,
    '789C' => 30876,'78B1' => 30897,
    '78B9' => 30905,'78D9' => 30937,
    '7934' => 31028,'793B' => 31035,
    '793C' => 31036,'7962' => 31074,
    '796F' => 31087,'7977' => 31095,
    '7978' => 31096,'7980' => 31104,
    '7984' => 31108,'7985' => 31109,
    '79C3' => 31171,'79C6' => 31174,
    '79EF' => 31215,'79F0' => 31216,
    '79FD' => 31229,'7A06' => 31238,
    '7A0E' => 31246,'7A23' => 31267,
    '7A33' => 31283,'7A51' => 31313,
    '7A77' => 31351,'7A83' => 31363,
    '7A8D' => 31373,'7A91' => 31377,
    '7A9C' => 31388,'7A9D' => 31389,
    '7AA5' => 31397,'7AA6' => 31398,
    '7AAD' => 31405,'7AD6' => 31446,
    '7ADE' => 31454,'7B03' => 31491,
    '7B0B' => 31499,'7B14' => 31508,
    '7B15' => 31509,'7B3A' => 31546,
    '7B3C' => 31548,'7B3E' => 31550,
    '7B5A' => 31578,'7B5B' => 31579,
    '7B5D' => 31581,'7B62' => 31586,
    '7B79' => 31609,'7B7B' => 31611,
    '7B7E' => 31614,'7B80' => 31616,
    '7BA2' => 31650,'7BA6' => 31654,
    '7BA7' => 31655,'7BA8' => 31656,
    '7BA9' => 31657,'7BAA' => 31658,
    '7BAB' => 31659,'7BD1' => 31697,
    '7BD3' => 31699,'7BEE' => 31726,
    '7BFC' => 31740,'7C16' => 31766,
    '7C41' => 31809,'7C74' => 31860,
    '7C7B' => 31867,'7C7C' => 31868,
    '7C9C' => 31900,'7C9D' => 31901,
    '7CA4' => 31908,'7CAA' => 31914,
    '7CAE' => 31918,'7CC1' => 31937,
    '7CC7' => 31943,'7CCD' => 31949,
    '7D27' => 32039,'7D77' => 32119,
    '7E9F' => 32415,'7EA0' => 32416,
    '7EA1' => 32417,'7EA2' => 32418,
    '7EA3' => 32419,'7EA4' => 32420,
    '7EA5' => 32421,'7EA6' => 32422,
    '7EA7' => 32423,'7EA8' => 32424,
    '7EA9' => 32425,'7EAA' => 32426,
    '7EAB' => 32427,'7EAC' => 32428,
    '7EAD' => 32429,'7EAF' => 32431,
    '7EB0' => 32432,'7EB1' => 32433,
    '7EB2' => 32434,'7EB3' => 32435,
    '7EB5' => 32437,'7EB6' => 32438,
    '7EB7' => 32439,'7EB8' => 32440,
    '7EB9' => 32441,'7EBA' => 32442,
    '7EBD' => 32445,'7EBE' => 32446,
    '7EBF' => 32447,'7EC0' => 32448,
    '7EC1' => 32449,'7EC2' => 32450,
    '7EC3' => 32451,'7EC4' => 32452,
    '7EC5' => 32453,'7EC6' => 32454,
    '7EC7' => 32455,'7EC8' => 32456,
    '7EC9' => 32457,'7ECA' => 32458,
    '7ECB' => 32459,'7ECC' => 32460,
    '7ECD' => 32461,'7ECE' => 32462,
    '7ECF' => 32463,'7ED0' => 32464,
    '7ED1' => 32465,'7ED2' => 32466,
    '7ED3' => 32467,'7ED4' => 32468,
    '7ED5' => 32469,'7ED7' => 32471,
    '7ED8' => 32472,'7ED9' => 32473,
    '7EDA' => 32474,'7EDB' => 32475,
    '7EDC' => 32476,'7EDD' => 32477,
    '7EDE' => 32478,'7EDF' => 32479,
    '7EE0' => 32480,'7EE1' => 32481,
    '7EE2' => 32482,'7EE3' => 32483,
    '7EE5' => 32485,'7EE6' => 32486,
    '7EE7' => 32487,'7EE8' => 32488,
    '7EE9' => 32489,'7EEA' => 32490,
    '7EEB' => 32491,'7EED' => 32493,
    '7EEE' => 32494,'7EEF' => 32495,
    '7EF0' => 32496,'7EF1' => 32497,
    '7EF2' => 32498,'7EF3' => 32499,
    '7EF4' => 32500,'7EF5' => 32501,
    '7EF6' => 32502,'7EF7' => 32503,
    '7EF8' => 32504,'7EFA' => 32506,
    '7EFB' => 32507,'7EFC' => 32508,
    '7EFD' => 32509,'7EFE' => 32510,
    '7EFF' => 32511,'7F00' => 32512,
    '7F01' => 32513,'7F02' => 32514,
    '7F03' => 32515,'7F04' => 32516,
    '7F05' => 32517,'7F06' => 32518,
    '7F07' => 32519,'7F08' => 32520,
    '7F09' => 32521,'7F0B' => 32523,
    '7F0C' => 32524,'7F0D' => 32525,
    '7F0E' => 32526,'7F0F' => 32527,
    '7F11' => 32529,'7F12' => 32530,
    '7F13' => 32531,'7F14' => 32532,
    '7F15' => 32533,'7F16' => 32534,
    '7F17' => 32535,'7F18' => 32536,
    '7F19' => 32537,'7F1A' => 32538,
    '7F1B' => 32539,'7F1C' => 32540,
    '7F1D' => 32541,'7F1F' => 32543,
    '7F20' => 32544,'7F21' => 32545,
    '7F22' => 32546,'7F23' => 32547,
    '7F24' => 32548,'7F25' => 32549,
    '7F26' => 32550,'7F27' => 32551,
    '7F28' => 32552,'7F29' => 32553,
    '7F2A' => 32554,'7F2B' => 32555,
    '7F2C' => 32556,'7F2D' => 32557,
    '7F2E' => 32558,'7F2F' => 32559,
    '7F30' => 32560,'7F31' => 32561,
    '7F32' => 32562,'7F33' => 32563,
    '7F34' => 32564,'7F35' => 32565,
    '7F42' => 32578,'7F57' => 32599,
    '7F5A' => 32602,'7F62' => 32610,
    '7F71' => 32625,'7F74' => 32628,
    '7F81' => 32641,'7F9F' => 32671,
    '7FD8' => 32728,'8020' => 32800,
    '8022' => 32802,'8025' => 32805,
    '8027' => 32807,'8031' => 32817,
    '8038' => 32824,'803B' => 32827,
    '8042' => 32834,'804B' => 32843,
    '804C' => 32844,'804D' => 32845,
    '8054' => 32852,'8069' => 32873,
    '806A' => 32874,'8080' => 32896,
    '8083' => 32899,'809F' => 32927,
    '80A0' => 32928,'80A4' => 32932,
    '80B7' => 32951,'80BC' => 32956,
    '80BD' => 32957,'80BE' => 32958,
    '80BF' => 32959,'80C0' => 32960,
    '80C1' => 32961,'80C6' => 32966,
    '80E7' => 32999,'80E8' => 33000,
    '80E9' => 33001,'80EA' => 33002,
    '80EB' => 33003,'80EC' => 33004,
    '80F6' => 33014,'8109' => 33033,
    '810D' => 33037,'810E' => 33038,
    '810F' => 33039,'8110' => 33040,
    '8111' => 33041,'8112' => 33042,
    '8113' => 33043,'8114' => 33044,
    '811A' => 33050,'8131' => 33073,
    '8132' => 33074,'8136' => 33078,
    '8138' => 33080,'8148' => 33096,
    '8159' => 33113,'815A' => 33114,
    '816D' => 33133,'817B' => 33147,
    '817C' => 33148,'817D' => 33149,
    '817E' => 33150,'8191' => 33169,
    '81AA' => 33194,'81C1' => 33217,
    '8206' => 33286,'8223' => 33315,
    '822D' => 33325,'8230' => 33328,
    '8231' => 33329,'823B' => 33339,
    '823E' => 33342,'8270' => 33392,
    '8273' => 33395,'8279' => 33401,
    '827A' => 33402,'8282' => 33410,
    '8288' => 33416,'8297' => 33431,
    '829C' => 33436,'82A6' => 33446,
    '82AA' => 33450,'82C1' => 33473,
    '82C4' => 33476,'82C7' => 33479,
    '82C8' => 33480,'82CA' => 33482,
    '82CB' => 33483,'82CC' => 33484,
    '82CD' => 33485,'82CE' => 33486,
    '82CF' => 33487,'82D8' => 33496,
    '82F7' => 33527,'830E' => 33550,
    '830F' => 33551,'8311' => 33553,
    '8314' => 33556,'8315' => 33557,
    '831A' => 33562,'8346' => 33606,
    '835A' => 33626,'835B' => 33627,
    '835C' => 33628,'835E' => 33630,
    '835F' => 33631,'8360' => 33632,
    '8361' => 33633,'8363' => 33635,
    '8364' => 33636,'8365' => 33637,
    '8366' => 33638,'8367' => 33639,
    '8368' => 33640,'8369' => 33641,
    '836A' => 33642,'836B' => 33643,
    '836C' => 33644,'836D' => 33645,
    '836E' => 33646,'836F' => 33647,
    '8385' => 33669,'839C' => 33692,
    '83B1' => 33713,'83B2' => 33714,
    '83B3' => 33715,'83B4' => 33716,
    '83B6' => 33718,'83B7' => 33719,
    '83B8' => 33720,'83B9' => 33721,
    '83BA' => 33722,'83BC' => 33724,
    '8418' => 33816,'841C' => 33820,
    '841D' => 33821,'8424' => 33828,
    '8425' => 33829,'8426' => 33830,
    '8427' => 33831,'8428' => 33832,
    '845C' => 33884,'8471' => 33905,
    '8487' => 33927,'8488' => 33928,
    '8489' => 33929,'848B' => 33931,
    '848C' => 33932,'84BD' => 33981,
    '84DD' => 34013,'84DF' => 34015,
    '84E0' => 34016,'84E3' => 34019,
    '84E5' => 34021,'84E6' => 34022,
    '8537' => 34103,'8538' => 34104,
    '8539' => 34105,'853A' => 34106,
    '853C' => 34108,'8572' => 34162,
    '8574' => 34164,'85AE' => 34222,
    '85C1' => 34241,'85D3' => 34259,
    '8616' => 34326,'864F' => 34383,
    '8651' => 34385,'865A' => 34394,
    '866C' => 34412,'867D' => 34429,
    '867E' => 34430,'867F' => 34431,
    '8680' => 34432,'8681' => 34433,
    '8682' => 34434,'86AC' => 34476,
    '86CA' => 34506,'86CE' => 34510,
    '86CF' => 34511,'86EE' => 34542,
    '86F0' => 34544,'86F1' => 34545,
    '86F2' => 34546,'86F3' => 34547,
    '86F4' => 34548,'8715' => 34581,
    '8717' => 34583,'8747' => 34631,
    '8748' => 34632,'8749' => 34633,
    '8770' => 34672,'877C' => 34684,
    '877D' => 34685,'877E' => 34686,
    '878B' => 34699,'87A8' => 34728,
    '87EE' => 34798,'8845' => 34885,
    '8854' => 34900,'8864' => 34916,
    '8865' => 34917,'886C' => 34924,
    '886E' => 34926,'8884' => 34948,
    '8885' => 34949,'889C' => 34972,
    '88AD' => 34989,'88C5' => 35013,
    '88C6' => 35014,'88E2' => 35042,
    '88E3' => 35043,'88E4' => 35044,
    '88E5' => 35045,'891B' => 35099,
    '8934' => 35124,'89C1' => 35265,
    '89C2' => 35266,'89C4' => 35268,
    '89C5' => 35269,'89C6' => 35270,
    '89C7' => 35271,'89C8' => 35272,
    '89C9' => 35273,'89CA' => 35274,
    '89CB' => 35275,'89CC' => 35276,
    '89CE' => 35278,'89CF' => 35279,
    '89D0' => 35280,'89D1' => 35281,
    '89DE' => 35294,'89EF' => 35311,
    '8A89' => 35465,'8A8A' => 35466,
    '8BA0' => 35744,'8BA1' => 35745,
    '8BA2' => 35746,'8BA3' => 35747,
    '8BA4' => 35748,'8BA5' => 35749,
    '8BA6' => 35750,'8BA7' => 35751,
    '8BA8' => 35752,'8BA9' => 35753,
    '8BAA' => 35754,'8BAB' => 35755,
    '8BAD' => 35757,'8BAE' => 35758,
    '8BAF' => 35759,'8BB0' => 35760,
    '8BB2' => 35762,'8BB3' => 35763,
    '8BB4' => 35764,'8BB5' => 35765,
    '8BB6' => 35766,'8BB7' => 35767,
    '8BB8' => 35768,'8BB9' => 35769,
    '8BBA' => 35770,'8BBC' => 35772,
    '8BBD' => 35773,'8BBE' => 35774,
    '8BBF' => 35775,'8BC0' => 35776,
    '8BC1' => 35777,'8BC2' => 35778,
    '8BC3' => 35779,'8BC4' => 35780,
    '8BC5' => 35781,'8BC6' => 35782,
    '8BC8' => 35784,'8BC9' => 35785,
    '8BCA' => 35786,'8BCB' => 35787,
    '8BCC' => 35788,'8BCD' => 35789,
    '8BCE' => 35790,'8BCF' => 35791,
    '8BD1' => 35793,'8BD2' => 35794,
    '8BD3' => 35795,'8BD4' => 35796,
    '8BD5' => 35797,'8BD6' => 35798,
    '8BD7' => 35799,'8BD8' => 35800,
    '8BD9' => 35801,'8BDA' => 35802,
    '8BDB' => 35803,'8BDC' => 35804,
    '8BDD' => 35805,'8BDE' => 35806,
    '8BDF' => 35807,'8BE0' => 35808,
    '8BE1' => 35809,'8BE2' => 35810,
    '8BE3' => 35811,'8BE4' => 35812,
    '8BE5' => 35813,'8BE6' => 35814,
    '8BE7' => 35815,'8BE8' => 35816,
    '8BE9' => 35817,'8BEB' => 35819,
    '8BEC' => 35820,'8BED' => 35821,
    '8BEE' => 35822,'8BEF' => 35823,
    '8BF0' => 35824,'8BF1' => 35825,
    '8BF2' => 35826,'8BF3' => 35827,
    '8BF4' => 35828,'8BF5' => 35829,
    '8BF6' => 35830,'8BF7' => 35831,
    '8BF8' => 35832,'8BF9' => 35833,
    '8BFA' => 35834,'8BFB' => 35835,
    '8BFC' => 35836,'8BFD' => 35837,
    '8BFE' => 35838,'8BFF' => 35839,
    '8C00' => 35840,'8C01' => 35841,
    '8C02' => 35842,'8C03' => 35843,
    '8C04' => 35844,'8C05' => 35845,
    '8C06' => 35846,'8C07' => 35847,
    '8C08' => 35848,'8C0A' => 35850,
    '8C0B' => 35851,'8C0C' => 35852,
    '8C0D' => 35853,'8C0E' => 35854,
    '8C0F' => 35855,'8C10' => 35856,
    '8C11' => 35857,'8C12' => 35858,
    '8C13' => 35859,'8C14' => 35860,
    '8C15' => 35861,'8C16' => 35862,
    '8C17' => 35863,'8C18' => 35864,
    '8C19' => 35865,'8C1A' => 35866,
    '8C1B' => 35867,'8C1C' => 35868,
    '8C1D' => 35869,'8C1F' => 35871,
    '8C20' => 35872,'8C21' => 35873,
    '8C22' => 35874,'8C23' => 35875,
    '8C24' => 35876,'8C25' => 35877,
    '8C26' => 35878,'8C27' => 35879,
    '8C28' => 35880,'8C29' => 35881,
    '8C2A' => 35882,'8C2B' => 35883,
    '8C2C' => 35884,'8C2D' => 35885,
    '8C2E' => 35886,'8C2F' => 35887,
    '8C30' => 35888,'8C31' => 35889,
    '8C32' => 35890,'8C33' => 35891,
    '8C34' => 35892,'8C35' => 35893,
    '8C36' => 35894,'8D1D' => 36125,
    '8D1E' => 36126,'8D1F' => 36127,
    '8D21' => 36129,'8D22' => 36130,
    '8D23' => 36131,'8D24' => 36132,
    '8D25' => 36133,'8D26' => 36134,
    '8D27' => 36135,'8D28' => 36136,
    '8D29' => 36137,'8D2A' => 36138,
    '8D2B' => 36139,'8D2C' => 36140,
    '8D2D' => 36141,'8D2E' => 36142,
    '8D2F' => 36143,'8D30' => 36144,
    '8D31' => 36145,'8D32' => 36146,
    '8D33' => 36147,'8D34' => 36148,
    '8D35' => 36149,'8D36' => 36150,
    '8D37' => 36151,'8D38' => 36152,
    '8D39' => 36153,'8D3A' => 36154,
    '8D3B' => 36155,'8D3C' => 36156,
    '8D3D' => 36157,'8D3E' => 36158,
    '8D3F' => 36159,'8D40' => 36160,
    '8D41' => 36161,'8D42' => 36162,
    '8D43' => 36163,'8D44' => 36164,
    '8D45' => 36165,'8D46' => 36166,
    '8D47' => 36167,'8D48' => 36168,
    '8D49' => 36169,'8D4A' => 36170,
    '8D4B' => 36171,'8D4C' => 36172,
    '8D4D' => 36173,'8D4E' => 36174,
    '8D4F' => 36175,'8D50' => 36176,
    '8D53' => 36179,'8D54' => 36180,
    '8D55' => 36181,'8D56' => 36182,
    '8D58' => 36184,'8D59' => 36185,
    '8D5A' => 36186,'8D5B' => 36187,
    '8D5C' => 36188,'8D5D' => 36189,
    '8D5E' => 36190,'8D60' => 36192,
    '8D61' => 36193,'8D62' => 36194,
    '8D63' => 36195,'8D75' => 36213,
    '8D8B' => 36235,'8DB1' => 36273,
    '8DB8' => 36280,'8DC3' => 36291,
    '8DC4' => 36292,'8DDE' => 36318,
    '8DF5' => 36341,'8DF7' => 36343,
    '8DF8' => 36344,'8DF9' => 36345,
    '8DFB' => 36347,'8E0C' => 36364,
    '8E2A' => 36394,'8E2C' => 36396,
    '8E2F' => 36399,'8E3A' => 36410,
    '8E51' => 36433,'8E52' => 36434,
    '8E70' => 36464,'8E7F' => 36479,
    '8E8F' => 36495,'8E9C' => 36508,
    '8EAF' => 36527,'8ECE' => 36558,
    '8F66' => 36710,'8F67' => 36711,
    '8F68' => 36712,'8F69' => 36713,
    '8F6B' => 36715,'8F6C' => 36716,
    '8F6D' => 36717,'8F6E' => 36718,
    '8F6F' => 36719,'8F70' => 36720,
    '8F71' => 36721,'8F72' => 36722,
    '8F73' => 36723,'8F74' => 36724,
    '8F75' => 36725,'8F76' => 36726,
    '8F77' => 36727,'8F78' => 36728,
    '8F79' => 36729,'8F7A' => 36730,
    '8F7B' => 36731,'8F7C' => 36732,
    '8F7D' => 36733,'8F7E' => 36734,
    '8F7F' => 36735,'8F81' => 36737,
    '8F82' => 36738,'8F83' => 36739,
    '8F84' => 36740,'8F85' => 36741,
    '8F86' => 36742,'8F87' => 36743,
    '8F88' => 36744,'8F89' => 36745,
    '8F8A' => 36746,'8F8B' => 36747,
    '8F8D' => 36749,'8F8E' => 36750,
    '8F8F' => 36751,'8F90' => 36752,
    '8F91' => 36753,'8F93' => 36755,
    '8F94' => 36756,'8F95' => 36757,
    '8F96' => 36758,'8F97' => 36759,
    '8F98' => 36760,'8F99' => 36761,
    '8F9A' => 36762,'8F9E' => 36766,
    '8FA9' => 36777,'8FAB' => 36779,
    '8FB6' => 36790,'8FB9' => 36793,
    '8FBD' => 36797,'8FBE' => 36798,
    '8FC1' => 36801,'8FC7' => 36807,
    '8FC8' => 36808,'8FD0' => 36816,
    '8FD8' => 36824,'8FD9' => 36825,
    '8FDB' => 36827,'8FDC' => 36828,
    '8FDD' => 36829,'8FDE' => 36830,
    '8FDF' => 36831,'8FE9' => 36841,
    '8FF3' => 36851,'8FF9' => 36857,
    '9009' => 36873,'900A' => 36874,
    '9012' => 36882,'9026' => 36902,
    '903B' => 36923,'9057' => 36951,
    '9065' => 36965,'9093' => 37011,
    '909D' => 37021,'90AC' => 37036,
    '90AE' => 37038,'90B9' => 37049,
    '90BA' => 37050,'90BB' => 37051,
    '90C4' => 37060,'90CF' => 37071,
    '90D0' => 37072,'90D1' => 37073,
    '90D3' => 37075,'90E6' => 37094,
    '90E7' => 37095,'90F8' => 37112,
    '915D' => 37213,'915E' => 37214,
    '9170' => 37232,'9171' => 37233,
    '9176' => 37238,'917D' => 37245,
    '917E' => 37246,'917F' => 37247,
    '918C' => 37260,'91CA' => 37322,
    '9274' => 37492,'92AE' => 37550,
    '933E' => 37694,'943E' => 37950,
    '9485' => 38021,'9486' => 38022,
    '9487' => 38023,'9488' => 38024,
    '9489' => 38025,'948A' => 38026,
    '948B' => 38027,'948C' => 38028,
    '948D' => 38029,'948E' => 38030,
    '948F' => 38031,'9490' => 38032,
    '9492' => 38034,'9493' => 38035,
    '9494' => 38036,'9495' => 38037,
    '9497' => 38039,'9499' => 38041,
    '949A' => 38042,'949B' => 38043,
    '949C' => 38044,'949D' => 38045,
    '949E' => 38046,'949F' => 38047,
    '94A0' => 38048,'94A1' => 38049,
    '94A2' => 38050,'94A3' => 38051,
    '94A4' => 38052,'94A5' => 38053,
    '94A6' => 38054,'94A7' => 38055,
    '94A8' => 38056,'94A9' => 38057,
    '94AA' => 38058,'94AB' => 38059,
    '94AC' => 38060,'94AD' => 38061,
    '94AE' => 38062,'94AF' => 38063,
    '94B0' => 38064,'94B1' => 38065,
    '94B2' => 38066,'94B3' => 38067,
    '94B4' => 38068,'94B5' => 38069,
    '94B6' => 38070,'94B7' => 38071,
    '94B8' => 38072,'94B9' => 38073,
    '94BA' => 38074,'94BB' => 38075,
    '94BC' => 38076,'94BD' => 38077,
    '94BE' => 38078,'94BF' => 38079,
    '94C0' => 38080,'94C1' => 38081,
    '94C2' => 38082,'94C3' => 38083,
    '94C4' => 38084,'94C5' => 38085,
    '94C6' => 38086,'94C8' => 38088,
    '94C9' => 38089,'94CA' => 38090,
    '94CB' => 38091,'94CC' => 38092,
    '94CD' => 38093,'94CE' => 38094,
    '94D0' => 38096,'94D1' => 38097,
    '94D2' => 38098,'94D5' => 38101,
    '94D6' => 38102,'94D7' => 38103,
    '94D8' => 38104,'94D9' => 38105,
    '94DB' => 38107,'94DC' => 38108,
    '94DD' => 38109,'94DE' => 38110,
    '94DF' => 38111,'94E0' => 38112,
    '94E1' => 38113,'94E2' => 38114,
    '94E3' => 38115,'94E4' => 38116,
    '94E5' => 38117,'94E7' => 38119,
    '94E8' => 38120,'94E9' => 38121,
    '94EA' => 38122,'94EB' => 38123,
    '94EC' => 38124,'94ED' => 38125,
    '94EE' => 38126,'94EF' => 38127,
    '94F0' => 38128,'94F1' => 38129,
    '94F2' => 38130,'94F3' => 38131,
    '94F4' => 38132,'94F5' => 38133,
    '94F6' => 38134,'94F7' => 38135,
    '94F8' => 38136,'94F9' => 38137,
    '94FA' => 38138,'94FC' => 38140,
    '94FD' => 38141,'94FE' => 38142,
    '94FF' => 38143,'9500' => 38144,
    '9501' => 38145,'9502' => 38146,
    '9503' => 38147,'9504' => 38148,
    '9505' => 38149,'9506' => 38150,
    '9507' => 38151,'9508' => 38152,
    '9509' => 38153,'950A' => 38154,
    '950B' => 38155,'950C' => 38156,
    '950D' => 38157,'950E' => 38158,
    '950F' => 38159,'9510' => 38160,
    '9511' => 38161,'9512' => 38162,
    '9513' => 38163,'9514' => 38164,
    '9515' => 38165,'9516' => 38166,
    '9517' => 38167,'9518' => 38168,
    '9519' => 38169,'951A' => 38170,
    '951B' => 38171,'951D' => 38173,
    '951E' => 38174,'951F' => 38175,
    '9521' => 38177,'9522' => 38178,
    '9523' => 38179,'9524' => 38180,
    '9525' => 38181,'9526' => 38182,
    '9528' => 38184,'9529' => 38185,
    '952A' => 38186,'952B' => 38187,
    '952C' => 38188,'952D' => 38189,
    '952E' => 38190,'952F' => 38191,
    '9530' => 38192,'9531' => 38193,
    '9532' => 38194,'9534' => 38196,
    '9535' => 38197,'9536' => 38198,
    '9537' => 38199,'9538' => 38200,
    '9539' => 38201,'953A' => 38202,
    '953B' => 38203,'953C' => 38204,
    '953E' => 38206,'953F' => 38207,
    '9540' => 38208,'9541' => 38209,
    '9542' => 38210,'9544' => 38212,
    '9545' => 38213,'9546' => 38214,
    '9547' => 38215,'9549' => 38217,
    '954A' => 38218,'954C' => 38220,
    '954D' => 38221,'954E' => 38222,
    '954F' => 38223,'9550' => 38224,
    '9551' => 38225,'9552' => 38226,
    '9553' => 38227,'9554' => 38228,
    '9556' => 38230,'9557' => 38231,
    '9558' => 38232,'9559' => 38233,
    '955B' => 38235,'955C' => 38236,
    '955D' => 38237,'955E' => 38238,
    '955F' => 38239,'9561' => 38241,
    '9562' => 38242,'9563' => 38243,
    '9564' => 38244,'9565' => 38245,
    '9566' => 38246,'9567' => 38247,
    '9568' => 38248,'9569' => 38249,
    '956A' => 38250,'956B' => 38251,
    '956C' => 38252,'956D' => 38253,
    '956F' => 38255,'9570' => 38256,
    '9571' => 38257,'9572' => 38258,
    '9573' => 38259,'9576' => 38262,
    '957F' => 38271,'95E8' => 38376,
    '95E9' => 38377,'95EA' => 38378,
    '95EB' => 38379,'95ED' => 38381,
    '95EE' => 38382,'95EF' => 38383,
    '95F0' => 38384,'95F1' => 38385,
    '95F2' => 38386,'95F3' => 38387,
    '95F4' => 38388,'95F5' => 38389,
    '95F6' => 38390,'95F7' => 38391,
    '95F8' => 38392,'95F9' => 38393,
    '95FA' => 38394,'95FB' => 38395,
    '95FC' => 38396,'95FD' => 38397,
    '95FE' => 38398,'9600' => 38400,
    '9601' => 38401,'9602' => 38402,
    '9603' => 38403,'9604' => 38404,
    '9605' => 38405,'9606' => 38406,
    '9608' => 38408,'9609' => 38409,
    '960A' => 38410,'960B' => 38411,
    '960C' => 38412,'960D' => 38413,
    '960E' => 38414,'960F' => 38415,
    '9610' => 38416,'9611' => 38417,
    '9612' => 38418,'9614' => 38420,
    '9615' => 38421,'9616' => 38422,
    '9617' => 38423,'9619' => 38425,
    '961A' => 38426,'961D' => 38429,
    '961F' => 38431,'9633' => 38451,
    '9634' => 38452,'9635' => 38453,
    '9636' => 38454,'9645' => 38469,
    '9646' => 38470,'9647' => 38471,
    '9648' => 38472,'9649' => 38473,
    '9655' => 38485,'9667' => 38503,
    '9668' => 38504,'9669' => 38505,
    '968F' => 38543,'9690' => 38544,
    '96B6' => 38582,'96BD' => 38589,
    '96BE' => 38590,'96CF' => 38607,
    '96E0' => 38624,'96F3' => 38643,
    '96FE' => 38654,'9701' => 38657,
    '972D' => 38701,'9753' => 38739,
    '9759' => 38745,'9765' => 38757,
    '9791' => 38801,'9792' => 38802,
    '97AF' => 38831,'97B2' => 38834,
    '97B4' => 38836,'97E6' => 38886,
    '97E7' => 38887,'97E9' => 38889,
    '97EA' => 38890,'97EB' => 38891,
    '97EC' => 38892,'97F5' => 38901,
    '9875' => 39029,'9876' => 39030,
    '9877' => 39031,'9878' => 39032,
    '9879' => 39033,'987A' => 39034,
    '987B' => 39035,'987C' => 39036,
    '987D' => 39037,'987E' => 39038,
    '987F' => 39039,'9880' => 39040,
    '9881' => 39041,'9882' => 39042,
    '9883' => 39043,'9884' => 39044,
    '9885' => 39045,'9886' => 39046,
    '9887' => 39047,'9888' => 39048,
    '9889' => 39049,'988A' => 39050,
    '988C' => 39052,'988D' => 39053,
    '988F' => 39055,'9890' => 39056,
    '9891' => 39057,'9893' => 39059,
    '9894' => 39060,'9896' => 39062,
    '9897' => 39063,'9898' => 39064,
    '989A' => 39066,'989B' => 39067,
    '989C' => 39068,'989D' => 39069,
    '989E' => 39070,'989F' => 39071,
    '98A0' => 39072,'98A1' => 39073,
    '98A2' => 39074,'98A4' => 39076,
    '98A5' => 39077,'98A6' => 39078,
    '98A7' => 39079,'98CE' => 39118,
    '98D1' => 39121,'98D2' => 39122,
    '98D3' => 39123,'98D5' => 39125,
    '98D8' => 39128,'98D9' => 39129,
    '98DA' => 39130,'98DE' => 39134,
    '98E8' => 39144,'990D' => 39181,
    '9963' => 39267,'9965' => 39269,
    '9967' => 39271,'9968' => 39272,
    '9969' => 39273,'996A' => 39274,
    '996B' => 39275,'996C' => 39276,
    '996D' => 39277,'996E' => 39278,
    '996F' => 39279,'9970' => 39280,
    '9971' => 39281,'9972' => 39282,
    '9974' => 39284,'9975' => 39285,
    '9976' => 39286,'9977' => 39287,
    '997A' => 39290,'997C' => 39292,
    '997D' => 39293,'997F' => 39295,
    '9980' => 39296,'9981' => 39297,
    '9984' => 39300,'9985' => 39301,
    '9986' => 39302,'9987' => 39303,
    '9988' => 39304,'998A' => 39306,
    '998B' => 39307,'998D' => 39309,
    '998F' => 39311,'9990' => 39312,
    '9991' => 39313,'9992' => 39314,
    '9993' => 39315,'9994' => 39316,
    '9995' => 39317,'9A6C' => 39532,
    '9A6D' => 39533,'9A6E' => 39534,
    '9A6F' => 39535,'9A70' => 39536,
    '9A71' => 39537,'9A73' => 39539,
    '9A74' => 39540,'9A75' => 39541,
    '9A76' => 39542,'9A77' => 39543,
    '9A78' => 39544,'9A79' => 39545,
    '9A7A' => 39546,'9A7B' => 39547,
    '9A7C' => 39548,'9A7D' => 39549,
    '9A7E' => 39550,'9A7F' => 39551,
    '9A80' => 39552,'9A81' => 39553,
    '9A82' => 39554,'9A84' => 39556,
    '9A85' => 39557,'9A86' => 39558,
    '9A87' => 39559,'9A88' => 39560,
    '9A8A' => 39562,'9A8B' => 39563,
    '9A8C' => 39564,'9A8F' => 39567,
    '9A90' => 39568,'9A91' => 39569,
    '9A92' => 39570,'9A93' => 39571,
    '9A96' => 39574,'9A97' => 39575,
    '9A98' => 39576,'9A9A' => 39578,
    '9A9B' => 39579,'9A9C' => 39580,
    '9A9D' => 39581,'9A9E' => 39582,
    '9A9F' => 39583,'9AA0' => 39584,
    '9AA1' => 39585,'9AA2' => 39586,
    '9AA3' => 39587,'9AA4' => 39588,
    '9AA5' => 39589,'9AA7' => 39591,
    '9AB6' => 39606,'9ABA' => 39610,
    '9AC5' => 39621,'9ACB' => 39627,
    '9ACC' => 39628,'9B0F' => 39695,
    '9B13' => 39699,'9B47' => 39751,
    '9B49' => 39753,'9C7C' => 40060,
    '9C7F' => 40063,'9C81' => 40065,
    '9C82' => 40066,'9C85' => 40069,
    '9C86' => 40070,'9C87' => 40071,
    '9C88' => 40072,'9C8B' => 40075,
    '9C8D' => 40077,'9C8E' => 40078,
    '9C90' => 40080,'9C91' => 40081,
    '9C92' => 40082,'9C94' => 40084,
    '9C95' => 40085,'9C9A' => 40090,
    '9C9B' => 40091,'9C9C' => 40092,
    '9C9E' => 40094,'9C9F' => 40095,
    '9CA0' => 40096,'9CA1' => 40097,
    '9CA2' => 40098,'9CA3' => 40099,
    '9CA4' => 40100,'9CA5' => 40101,
    '9CA6' => 40102,'9CA7' => 40103,
    '9CA8' => 40104,'9CA9' => 40105,
    '9CAB' => 40107,'9CAD' => 40109,
    '9CAE' => 40110,'9CB0' => 40112,
    '9CB1' => 40113,'9CB2' => 40114,
    '9CB3' => 40115,'9CB4' => 40116,
    '9CB5' => 40117,'9CB6' => 40118,
    '9CB7' => 40119,'9CB8' => 40120,
    '9CBA' => 40122,'9CBB' => 40123,
    '9CBC' => 40124,'9CBD' => 40125,
    '9CC3' => 40131,'9CC4' => 40132,
    '9CC5' => 40133,'9CC6' => 40134,
    '9CC7' => 40135,'9CCA' => 40138,
    '9CCB' => 40139,'9CCC' => 40140,
    '9CCD' => 40141,'9CCE' => 40142,
    '9CCF' => 40143,'9CD0' => 40144,
    '9CD3' => 40147,'9CD4' => 40148,
    '9CD5' => 40149,'9CD6' => 40150,
    '9CD7' => 40151,'9CD8' => 40152,
    '9CD9' => 40153,'9CDC' => 40156,
    '9CDD' => 40157,'9CDE' => 40158,
    '9CDF' => 40159,'9CE2' => 40162,
    '9E1F' => 40479,'9E20' => 40480,
    '9E21' => 40481,'9E22' => 40482,
    '9E23' => 40483,'9E25' => 40485,
    '9E26' => 40486,'9E28' => 40488,
    '9E29' => 40489,'9E2A' => 40490,
    '9E2B' => 40491,'9E2C' => 40492,
    '9E2D' => 40493,'9E2F' => 40495,
    '9E31' => 40497,'9E32' => 40498,
    '9E33' => 40499,'9E35' => 40501,
    '9E36' => 40502,'9E37' => 40503,
    '9E38' => 40504,'9E39' => 40505,
    '9E3A' => 40506,'9E3D' => 40509,
    '9E3E' => 40510,'9E3F' => 40511,
    '9E41' => 40513,'9E42' => 40514,
    '9E43' => 40515,'9E44' => 40516,
    '9E45' => 40517,'9E46' => 40518,
    '9E47' => 40519,'9E48' => 40520,
    '9E49' => 40521,'9E4A' => 40522,
    '9E4B' => 40523,'9E4C' => 40524,
    '9E4E' => 40526,'9E4F' => 40527,
    '9E51' => 40529,'9E55' => 40533,
    '9E57' => 40535,'9E58' => 40536,
    '9E5A' => 40538,'9E5B' => 40539,
    '9E5C' => 40540,'9E5E' => 40542,
    '9E63' => 40547,'9E64' => 40548,
    '9E66' => 40550,'9E67' => 40551,
    '9E68' => 40552,'9E69' => 40553,
    '9E6A' => 40554,'9E6B' => 40555,
    '9E6C' => 40556,'9E6D' => 40557,
    '9E70' => 40560,'9E71' => 40561,
    '9E73' => 40563,'9E7E' => 40574,
    '9EA6' => 40614,'9EB8' => 40632,
    '9EBD' => 40637,'9EC4' => 40644,
    '9EC9' => 40649,'9EE2' => 40674,
    '9EE9' => 40681,'9EEA' => 40682,
    '9EFE' => 40702,'9F0B' => 40715,
    '9F0D' => 40717,'9F17' => 40727,
    '9F39' => 40761,'9F44' => 40772,
    '9F50' => 40784,'9F51' => 40785,
    '9F7F' => 40831,'9F80' => 40832,
    '9F83' => 40835,'9F84' => 40836,
    '9F85' => 40837,'9F86' => 40838,
    '9F87' => 40839,'9F88' => 40840,
    '9F89' => 40841,'9F8A' => 40842,
    '9F8B' => 40843,'9F8C' => 40844,
    '9F99' => 40857,'9F9A' => 40858,
    '9F9B' => 40859,'9F9F' => 40863,
};

sub detect {
    my $str = shift;
    my $len = $_[0];
    return unless $str;

    $str = substr($str, 0, $len) if ($len && $len>0 && length $len);

    my $jp_regexp =  join('|', @JP);
    my $kr_regexp =  join('|', @KR);
    my $han_regexp = join('|', @HAN);
    my %RES = ();

    # setup unicode flag
    Encode::_utf8_on($str);

    if ($str =~ /$jp_regexp/) {
        $RES{'JP'} = VAR_JP;
    }

    if ($str =~ /$kr_regexp/) {
        $RES{'KR'} = VAR_KR;
    }

    $str =~ s/([$han_regexp]+)/cn_detect(\%RES, $1)/egsx;

    # not match any CJK characters
    $RES{'7BIT'} = VAR_7BIT if (!keys %RES);

    \%RES;
}

sub detect1 {
    my $ref = detect(@_);
    join ('+', sort { $ref->{$a} <=> $ref->{$b} } keys %$ref);
}

sub cn_detect {
    my $ref = shift;
    my $str = shift; # utf8 on

    # myprint("call me again by $str</br>\n");
    for my $char (split(//, $str)) {
        my $hex = uc sprintf("%x", ord($char));
        if (!$ref->{SCH}) {
            $ref->{SCH} = VAR_SCH if ($UTF8_SCH->{$hex});
        }
        if (!$ref->{TCH}) {
            $ref->{TCH} = VAR_TCH if ($UTF8_TCH->{$hex});
        }
        if (!$ref->{HAN}) {
            $ref->{HAN} = VAR_HAN;
        }
    }

    $str;
}

sub trylocal {
    my $pattern = shift;        # must exist
    my $str = shift;            # must exist
    my $prefer_charset = uc shift; # optional param

    if (my $r = $DETECT_MAPS->{"$pattern"}) {

        # a fix to a Unihan string that map to our prefer BIG5/GB2312
        # local encoding if we specify it, but only valid for Chinese
        # now, sorry for Japan and Korea users. will improve soon :)
        if ($prefer_charset && $pattern eq 'HAN'
            && grep (/^$prefer_charset$/i, @$r)) {
            return $prefer_charset;
        }

        return $r unless ref $r;
        return $r->[scalar @$r-1] # the last fallback euc
            unless $str;          # is there any string to detect?
        # array

        for my $enc (@$r) {
            my $buf = $str;
            eval { $buf = _iconv($buf, 'utf-8', $enc, 1) };
            next if ($@);
            return $enc;
        }
    } else {
        return 'UTF-8';
    }
    'UTF-8';
}

sub trylocal1 {
    my $str = shift;
    trylocal(detect1($str), $str);
}

sub trylocal2 {
    my $str = shift;
    my $prefer_charset = shift;

    trylocal(detect1($str), $str, $prefer_charset);
}

sub _iconv {
    my ($str, $fchar, $tchar, $_CHK) = @_;

    if ($fchar =~/^gb2312$/i) {
        $fchar = 'GBK';
    }
    if ($tchar =~/^gb2312$/i) {
        $tchar = 'GBK';
    }

    from_to($str, $fchar, $tchar, $_CHK);
    $str;
}

1;

__END__

=head1 NAME

Encode::PPUniDetector - Detect language from unicode (utf8) stream

=head1 SYNOPSIS

  use Encode::PPUniDetector;

  # arbitrary utf8 string, caution: UTF8 string only!
  my $utf8_string = "...";

  # language pattern join by '+', valid language is:
  # HAN : Unihan charactor
  # SCH : Simplified chinese
  # TCH : Traditional chinese
  # JP  : Japanese
  # KR  : Korea
  my $pattern = Encode::PPUniDetector::detect1($utf8_string);

  # the closest local encoding to transfer $utf8_string
  my $euc = Encode::PPUniDetector::trylocal1($utf8_string);

=head1 COPYRIGHT

The Encode::PPDetector module is Copyright (c) 1998-2007 He zhiqiang
All rights reserved.

You may distribute under the terms of either the GNU General Public
License or the Artistic License, as specified in the Perl README file.
